﻿// Fraqcel 2.6.0 DLL Source Code
// Originally compiled in Microsoft Visual C# 2015
//
// Copyright(c) 2016, Kevin Orloske
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without modification, 
// are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice, 
// this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright notice, 
// this list of conditions and the following disclaimer in the documentation 
// and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
// OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
// OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
// EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

using RGiesecke.DllExport;  //from the Visual Studio Nuget Package Manager Console... Install-Package UnmanagedExports
// DllExport requires a platform target of either x86, ia64 or x64. AnyCPU assemblies cannot export functions. 
using System;
using System.Runtime.InteropServices;
using System.Threading;
using System.Threading.Tasks;
using Excel = Microsoft.Office.Interop.Excel;  //need to add reference to the Microsoft Excel 12.0 Object Library

namespace FraqcelLib
{
    [ComVisible(true), ClassInterface(ClassInterfaceType.AutoDual)]
    public class MainClass
    {
        public void fraqcelDbl([MarshalAs(UnmanagedType.SafeArray)]ref double[,] imageData,  //base 1 array
                               [MarshalAs(UnmanagedType.BStr)] string typeOption,
                               [MarshalAs(UnmanagedType.R8)] double eqnParameter,
                               [MarshalAs(UnmanagedType.Bool)] bool flagSmooth,
                               [MarshalAs(UnmanagedType.I2)] int flagDllOption,
                               [MarshalAs(UnmanagedType.R8)] double zr,
                               [MarshalAs(UnmanagedType.R8)] double zi,
                               [MarshalAs(UnmanagedType.R8)] double scaleTerm,
                               [MarshalAs(UnmanagedType.I2)] int res,
                               [MarshalAs(UnmanagedType.I4)] int maxIter,
                               [MarshalAs(UnmanagedType.I2)] int refreshType,
                               [MarshalAs(UnmanagedType.I4)] int prevMaxIter,
                               [MarshalAs(UnmanagedType.I2)] int r1,
                               [MarshalAs(UnmanagedType.I2)] int c1,
                               [MarshalAs(UnmanagedType.I2)] int navSize,
                               [MarshalAs(UnmanagedType.I2)] int rSize,
                               [MarshalAs(UnmanagedType.I2)] int cSize)
        {
            if (typeOption == "b") {  //Burning Ship fractal
                CLoop.bDbl(ref imageData, flagSmooth, flagDllOption, zr, zi, scaleTerm, res, maxIter,
                    refreshType, prevMaxIter, r1, c1, navSize, rSize, cSize);
            }
            else if (typeOption == "c") {  //Celtic fractal
                CLoop.cDbl(ref imageData, flagSmooth, flagDllOption, zr, zi, scaleTerm, res, maxIter,
                    refreshType, prevMaxIter, r1, c1, navSize, rSize, cSize);
            }
            else if (typeOption == "f") {  //Buffalo fractal
                CLoop.fDbl(ref imageData, flagSmooth, flagDllOption, zr, zi, scaleTerm, res, maxIter,
                    refreshType, prevMaxIter, r1, c1, navSize, rSize, cSize);
            }
            else if (typeOption == "r") {  //Rotation Family
                CLoop.rDbl(ref imageData, eqnParameter, flagSmooth, flagDllOption, zr, zi, scaleTerm, res, maxIter,
                    refreshType, prevMaxIter, r1, c1, navSize, rSize, cSize);
            }
            else if (typeOption == "t") {  //Beetle fractal
                CLoop.tDbl(ref imageData, flagSmooth, flagDllOption, zr, zi, scaleTerm, res, maxIter,
                    refreshType, prevMaxIter, r1, c1, navSize, rSize, cSize);
            }
            else if (typeOption == "p") {  //Perpendicular Burning Ship fractal
                CLoop.pDbl(ref imageData, flagSmooth, flagDllOption, zr, zi, scaleTerm, res, maxIter,
                    refreshType, prevMaxIter, r1, c1, navSize, rSize, cSize);
            }
            else {  //Mandelbrot set
                CLoop.mDbl(ref imageData, flagSmooth, flagDllOption, zr, zi, scaleTerm, res, maxIter,
                    refreshType, prevMaxIter, r1, c1, navSize, rSize, cSize);
            }
        }

        public void fraqcelPert([MarshalAs(UnmanagedType.SafeArray)]ref double[,] imageData,   //base 1 array
                                [MarshalAs(UnmanagedType.SafeArray)]ref int[,] glitchFlag,  //base 1 array
                                [MarshalAs(UnmanagedType.Bool)]ref bool flagGlitch,
                                [MarshalAs(UnmanagedType.BStr)] string typeOption,
                                [MarshalAs(UnmanagedType.R8)] double eqnParameter,
                                [MarshalAs(UnmanagedType.Bool)] bool flagSmooth,
                                [MarshalAs(UnmanagedType.I2)] int flagDllOption,
                                [MarshalAs(UnmanagedType.R8)] double x0,
                                [MarshalAs(UnmanagedType.R8)] double y0,
                                [MarshalAs(UnmanagedType.R8)] double scaleTerm,
                                [MarshalAs(UnmanagedType.I2)] int res,
                                [MarshalAs(UnmanagedType.I4)] int maxIter,
                                [MarshalAs(UnmanagedType.I2)] int refreshType,
                                [MarshalAs(UnmanagedType.I4)] int prevMaxIter,
                                [MarshalAs(UnmanagedType.I2)] int dr,
                                [MarshalAs(UnmanagedType.I2)] int dc,
                                [MarshalAs(UnmanagedType.I2)] int navSize,
                                [MarshalAs(UnmanagedType.I2)] int rr1,
                                [MarshalAs(UnmanagedType.I2)] int rr2,
                                [MarshalAs(UnmanagedType.I2)] int cc1,
                                [MarshalAs(UnmanagedType.I2)] int cc2,
                                [MarshalAs(UnmanagedType.I2)] int rRef,
                                [MarshalAs(UnmanagedType.I2)] int cRef,
                                [MarshalAs(UnmanagedType.I4)] int numPointsMP,
                                [MarshalAs(UnmanagedType.SafeArray)] double[,] xMP,  //base 1 array
                                [MarshalAs(UnmanagedType.SafeArray)] double[,] yMP,  //base 1 array
                                [MarshalAs(UnmanagedType.SafeArray)] double[,] glitchDetect,  //base varies
                                [MarshalAs(UnmanagedType.R8)] double divThreshold,
                                [MarshalAs(UnmanagedType.Bool)] bool flagDebugGlitch,
                                [MarshalAs(UnmanagedType.I4)] int iteration,
                                [MarshalAs(UnmanagedType.I4)] int maxPastRef,
                                [MarshalAs(UnmanagedType.SafeArray)] int[,] prevGlitchFlag = null)
        {
            if (typeOption == "b") {  //Burning Ship fractal
                CLoop.bPert(ref imageData, ref glitchFlag, ref flagGlitch, flagSmooth, flagDllOption, x0, y0, scaleTerm, res, maxIter,
                    refreshType, prevMaxIter, dr, dc, navSize, rr1, rr2, cc1, cc2, rRef, cRef, numPointsMP,
                    xMP, yMP, glitchDetect, divThreshold, flagDebugGlitch, iteration, maxPastRef, prevGlitchFlag);
            }
            else if (typeOption == "c") {  //Celtic fractal
                CLoop.cPert(ref imageData, ref glitchFlag, ref flagGlitch, flagSmooth, flagDllOption, x0, y0, scaleTerm, res, maxIter,
                    refreshType, prevMaxIter, dr, dc, navSize, rr1, rr2, cc1, cc2, rRef, cRef, numPointsMP,
                    xMP, yMP, glitchDetect, divThreshold, flagDebugGlitch, iteration, maxPastRef, prevGlitchFlag);
            }
            else if (typeOption == "f") {  //Buffalo fractal
                CLoop.fPert(ref imageData, ref glitchFlag, ref flagGlitch, flagSmooth, flagDllOption, x0, y0, scaleTerm, res, maxIter,
                    refreshType, prevMaxIter, dr, dc, navSize, rr1, rr2, cc1, cc2, rRef, cRef, numPointsMP,
                    xMP, yMP, glitchDetect, divThreshold, flagDebugGlitch, iteration, maxPastRef, prevGlitchFlag);
            }
            else if (typeOption == "r") {  //Rotation Family
                CLoop.rPert(ref imageData, ref glitchFlag, ref flagGlitch, eqnParameter, flagSmooth, flagDllOption, x0, y0, scaleTerm, res, maxIter,
                    refreshType, prevMaxIter, dr, dc, navSize, rr1, rr2, cc1, cc2, rRef, cRef, numPointsMP,
                    xMP, yMP, glitchDetect, divThreshold, flagDebugGlitch, iteration, maxPastRef, prevGlitchFlag);
            }
            else if (typeOption == "t") {  //Beetle fractal
                CLoop.tPert(ref imageData, ref glitchFlag, ref flagGlitch, flagSmooth, flagDllOption, x0, y0, scaleTerm, res, maxIter,
                    refreshType, prevMaxIter, dr, dc, navSize, rr1, rr2, cc1, cc2, rRef, cRef, numPointsMP,
                    xMP, yMP, glitchDetect, divThreshold, flagDebugGlitch, iteration, maxPastRef, prevGlitchFlag);
            }
            else if (typeOption == "p") {  //Perpendicular Burning Ship fractal
                CLoop.pPert(ref imageData, ref glitchFlag, ref flagGlitch, flagSmooth, flagDllOption, x0, y0, scaleTerm, res, maxIter,
                    refreshType, prevMaxIter, dr, dc, navSize, rr1, rr2, cc1, cc2, rRef, cRef, numPointsMP,
                    xMP, yMP, glitchDetect, divThreshold, flagDebugGlitch, iteration, maxPastRef, prevGlitchFlag);
            }
            else {  //Mandelbrot set
                CLoop.mPert(ref imageData, ref glitchFlag, ref flagGlitch, flagSmooth, flagDllOption, x0, y0, scaleTerm, res, maxIter,
                    refreshType, prevMaxIter, dr, dc, navSize, rr1, rr2, cc1, cc2, rRef, cRef, numPointsMP,
                    xMP, yMP, glitchDetect, divThreshold, flagDebugGlitch, iteration, maxPastRef, prevGlitchFlag);
            }
        }

        public void fraqcelSeries([MarshalAs(UnmanagedType.SafeArray)]ref double[,] imageData,   //base 1 array
                                  [MarshalAs(UnmanagedType.SafeArray)]ref int[,] glitchFlag,  //base 1 array
                                  [MarshalAs(UnmanagedType.Bool)]ref bool flagGlitch,
                                  [MarshalAs(UnmanagedType.BStr)] string typeOption,
                                  [MarshalAs(UnmanagedType.R8)] double eqnParameter,
                                  [MarshalAs(UnmanagedType.I4)]ref int numSkip,
                                  [MarshalAs(UnmanagedType.I2)] int orderOption,
                                  [MarshalAs(UnmanagedType.Bool)] bool flagSmooth,
                                  [MarshalAs(UnmanagedType.I2)] int flagDllOption,
                                  [MarshalAs(UnmanagedType.R8)] double x0,
                                  [MarshalAs(UnmanagedType.R8)] double y0,
                                  [MarshalAs(UnmanagedType.R8)] double scaleTerm,
                                  [MarshalAs(UnmanagedType.I2)] int res,
                                  [MarshalAs(UnmanagedType.I4)] int maxIter,
                                  [MarshalAs(UnmanagedType.I2)] int refreshType,
                                  [MarshalAs(UnmanagedType.I4)] int prevMaxIter,
                                  [MarshalAs(UnmanagedType.I2)] int dr,
                                  [MarshalAs(UnmanagedType.I2)] int dc,
                                  [MarshalAs(UnmanagedType.I2)] int navSize,
                                  [MarshalAs(UnmanagedType.I2)] int rr1,
                                  [MarshalAs(UnmanagedType.I2)] int rr2,
                                  [MarshalAs(UnmanagedType.I2)] int cc1,
                                  [MarshalAs(UnmanagedType.I2)] int cc2,
                                  [MarshalAs(UnmanagedType.I2)] int rRef,
                                  [MarshalAs(UnmanagedType.I2)] int cRef,
                                  [MarshalAs(UnmanagedType.I4)] int numPointsMP,
                                  [MarshalAs(UnmanagedType.R8)] double seriesTolerance,
                                  [MarshalAs(UnmanagedType.SafeArray)] double[,] xMP,  //base 1 array
                                  [MarshalAs(UnmanagedType.SafeArray)] double[,] yMP,  //base 1 array
                                  [MarshalAs(UnmanagedType.SafeArray)] double[,] glitchDetect,  //base varies
                                  [MarshalAs(UnmanagedType.R8)] double divThreshold,
                                  [MarshalAs(UnmanagedType.Bool)] bool flagDebugGlitch,
                                  [MarshalAs(UnmanagedType.I4)] int iteration,
                                  [MarshalAs(UnmanagedType.I4)] int maxPastRef,
                                  [MarshalAs(UnmanagedType.SafeArray)] int[,] prevGlitchFlag = null,
                                  [MarshalAs(UnmanagedType.I4)] int prevGlitchInfo2 = 0)
        {
            if (typeOption == "r") {  //Rotation Family
                if (orderOption == 3) {  //force 3rd order series approximation
                    CLoop.rSeries3(ref imageData, ref glitchFlag, ref flagGlitch, ref numSkip, eqnParameter, flagSmooth, flagDllOption, x0, y0, scaleTerm, res, maxIter,
                       refreshType, prevMaxIter, dr, dc, navSize, rr1, rr2, cc1, cc2, rRef, cRef, numPointsMP, seriesTolerance,
                       xMP, yMP, glitchDetect, divThreshold, flagDebugGlitch, iteration, maxPastRef, prevGlitchFlag, prevGlitchInfo2);
                }
                else if (orderOption == 5) {  //force 5th order series approximation
                    CLoop.rSeries5(ref imageData, ref glitchFlag, ref flagGlitch, ref numSkip, eqnParameter, flagSmooth, flagDllOption, x0, y0, scaleTerm, res, maxIter,
                       refreshType, prevMaxIter, dr, dc, navSize, rr1, rr2, cc1, cc2, rRef, cRef, numPointsMP, seriesTolerance,
                       xMP, yMP, glitchDetect, divThreshold, flagDebugGlitch, iteration, maxPastRef, prevGlitchFlag, prevGlitchInfo2);
                }
                else {  //adaptive 3rd or 5th order series approximation
                    CLoop.rSeries(ref imageData, ref glitchFlag, ref flagGlitch, ref numSkip, eqnParameter, flagSmooth, flagDllOption, x0, y0, scaleTerm, res, maxIter,
                       refreshType, prevMaxIter, dr, dc, navSize, rr1, rr2, cc1, cc2, rRef, cRef, numPointsMP, seriesTolerance,
                       xMP, yMP, glitchDetect, divThreshold, flagDebugGlitch, iteration, maxPastRef, prevGlitchFlag, prevGlitchInfo2);
                }
            }
            else if (typeOption == "t") {  //Beetle fractal
                if (orderOption == 3) {  //force 3rd order series approximation
                    CLoop.tSeries3(ref imageData, ref glitchFlag, ref flagGlitch, ref numSkip, flagSmooth, flagDllOption, x0, y0, scaleTerm, res, maxIter,
                       refreshType, prevMaxIter, dr, dc, navSize, rr1, rr2, cc1, cc2, rRef, cRef, numPointsMP, seriesTolerance,
                       xMP, yMP, glitchDetect, divThreshold, flagDebugGlitch, iteration, maxPastRef, prevGlitchFlag, prevGlitchInfo2);
                }
                else if (orderOption == 5) {  //force 5th order series approximation
                    CLoop.tSeries5(ref imageData, ref glitchFlag, ref flagGlitch, ref numSkip, flagSmooth, flagDllOption, x0, y0, scaleTerm, res, maxIter,
                       refreshType, prevMaxIter, dr, dc, navSize, rr1, rr2, cc1, cc2, rRef, cRef, numPointsMP, seriesTolerance,
                       xMP, yMP, glitchDetect, divThreshold, flagDebugGlitch, iteration, maxPastRef, prevGlitchFlag, prevGlitchInfo2);
                }
                else {  //adaptive 3rd or 5th order series approximation
                    CLoop.tSeries(ref imageData, ref glitchFlag, ref flagGlitch, ref numSkip, flagSmooth, flagDllOption, x0, y0, scaleTerm, res, maxIter,
                       refreshType, prevMaxIter, dr, dc, navSize, rr1, rr2, cc1, cc2, rRef, cRef, numPointsMP, seriesTolerance,
                       xMP, yMP, glitchDetect, divThreshold, flagDebugGlitch, iteration, maxPastRef, prevGlitchFlag, prevGlitchInfo2);
                }
            }
            else {  //Mandelbrot set
                if (orderOption == 3) {  //force 3rd order series approximation
                    CLoop.mSeries3(ref imageData, ref glitchFlag, ref flagGlitch, ref numSkip, flagSmooth, flagDllOption, x0, y0, scaleTerm, res, maxIter,
                       refreshType, prevMaxIter, dr, dc, navSize, rr1, rr2, cc1, cc2, rRef, cRef, numPointsMP, seriesTolerance,
                       xMP, yMP, glitchDetect, divThreshold, flagDebugGlitch, iteration, maxPastRef, prevGlitchFlag, prevGlitchInfo2);
                }
                else if (orderOption == 5) {  //force 5th order series approximation
                    CLoop.mSeries5(ref imageData, ref glitchFlag, ref flagGlitch, ref numSkip, flagSmooth, flagDllOption, x0, y0, scaleTerm, res, maxIter,
                       refreshType, prevMaxIter, dr, dc, navSize, rr1, rr2, cc1, cc2, rRef, cRef, numPointsMP, seriesTolerance,
                       xMP, yMP, glitchDetect, divThreshold, flagDebugGlitch, iteration, maxPastRef, prevGlitchFlag, prevGlitchInfo2);
                }
                else {  //adaptive 3rd or 5th order series approximation
                    CLoop.mSeries(ref imageData, ref glitchFlag, ref flagGlitch, ref numSkip, flagSmooth, flagDllOption, x0, y0, scaleTerm, res, maxIter,
                       refreshType, prevMaxIter, dr, dc, navSize, rr1, rr2, cc1, cc2, rRef, cRef, numPointsMP, seriesTolerance,
                       xMP, yMP, glitchDetect, divThreshold, flagDebugGlitch, iteration, maxPastRef, prevGlitchFlag, prevGlitchInfo2);
                }
            }
        }

    }

    static class UnmanagedExports
    {
        [DllExport]
        [return: MarshalAs(UnmanagedType.IDispatch)]
        static Object CreateMainClass()
        {
            return new MainClass();
        }
    }

    public class CLoop
    {
        private const int highBailout = 1048576;  //2^20

        private static Excel.Application excelApp = null;  //used to manipulate the Excel status bar

        private static void reduceResolution(ref double[,] imageDataTemp, int r, int c, double iteration, int res)
        {  // This subroutine writes the iteration result to the imageData array when the resolution is reduced
            int r1, c1, r2, c2;
            if (r == 2) r1 = 0;  //2-res = either 0 or -1 for res 2 and 3, respectively
            else r1 = 2 - res;
            if (c == 2) c1 = 0;
            else c1 = 2 - res;
            r2 = 1;  //2+2*399=800 , 2+2*297=596
            c2 = 1;  //2+3*266=800 , 2+3*198=596
            for (int i = r1; i <= r2; i++) {
                for (int j = c1; j <= c2; j++) {
                    imageDataTemp[r - 1 + i, c - 1 + j] = iteration;
                }
            }
        }

        private static bool quickPan(int r1, int c1, int r, int c, int navSize)
        {  // Quick pan code
            if (r1 >= 0 && c1 >= 0) {
                if (r - 1 < navSize - r1 && c - 1 < navSize - c1) return true;
            }
            else if (r1 < 0 && c1 >= 0) {
                if (r - 1 > -r1 && c - 1 < navSize - c1) return true;
            }
            else if (r1 >= 0 && c1 < 0) {
                if (r - 1 < navSize - r1 && c - 1 > -c1) return true;
            }
            else {  //r1 < 0 && c1 < 0
                if (r - 1 > -r1 && c - 1 > -c1) return true;
            }
            return false;
        }

        private static void copyArray(ref int[,] A, int[,] B)
        {  // This subroutine assigns the values in array B to array A
            for (int i = A.GetLowerBound(0); i <= A.GetUpperBound(0); i++) {
                for (int j = A.GetLowerBound(1); j <= A.GetUpperBound(1); j++) {
                    A[i, j] = B[i, j];
                }  //next j
            }  //next i
        }

        private static void copyArray(ref double[,] A, double[,] B)
        {  // This subroutine assigns the values in array B to array A
            for (int i = A.GetLowerBound(0); i <= A.GetUpperBound(0); i++) {
                for (int j = A.GetLowerBound(1); j <= A.GetUpperBound(1); j++) {
                    A[i, j] = B[i, j];
                }  //next j
            }  //next i
        }

        private static int getOuterLoopNum(int rUpperIndex)
        {  // This function returns the number of times to run the Parallel.For loop for status reporting
            int i, j = 10;
            if (Environment.ProcessorCount < 1) return j;
            i = rUpperIndex / Environment.ProcessorCount;
            if (i < 1) return 1;
            if (i < j) return i;
            return j;
        }

        public static void mDbl(ref double[,] imageData, bool flagSmooth, int flagDllOption, double zr, double zi, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int r1, int c1, int navSize, int rSize, int cSize)
        {  // Mandelbrot set, standard double precision
            double pixelStep = scaleTerm * res;
            int rUpperIndex = ((rSize - 1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double[,] imageDataTemp = imageData.Clone() as double[,];  //base 1 array, imageData is not allowed in Parallel.For loop because it is passed by reference

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int r = res * ri + 2;          //mapping the parallel loop index to the row number
                    double y0 = zi - pixelStep * ri;
                    double x0 = zr - pixelStep;
                    double x, xtemp, y;
                    int iteration;
                    double iterationD;
                    bool skipCalc;
                    double log_zn, nu;

                    for (int c = 2; c <= cSize + 1; c = c + res) {
                        skipCalc = false;
                        x0 = x0 + pixelStep;
                        if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[r - 1, c - 1] == prevMaxIter)) {

                            // Quick pan code
                            if (refreshType == -1) {
                                skipCalc = quickPan(r1, c1, r, c, navSize);
                            }

                            // Iteration loop
                            if (skipCalc == false) {
                                iteration = 1;
                                x = x0;
                                y = y0;

                                // Discrete coloring
                                if (flagSmooth == false) {
                                    while (x * x + y * y < 4 && iteration < maxIter) {
                                        xtemp = x * x - y * y + x0;
                                        y = 2 * x * y + y0;
                                        x = xtemp;
                                        iteration++;
                                    }
                                    iterationD = iteration;
                                }

                                // Continuous coloring
                                else {
                                    while (x * x + y * y < highBailout && iteration < maxIter) {
                                        xtemp = x * x - y * y + x0;
                                        y = 2 * x * y + y0;
                                        x = xtemp;
                                        iteration++;
                                    }

                                    // Additional calculations for continuous coloring
                                    if (iteration < maxIter) {
                                        log_zn = Math.Log(x * x + y * y) / 2;
                                        nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                        iterationD = iteration + 1 - nu;
                                        iterationD = Math.Round(iterationD, 3);
                                    }
                                    else {
                                        iterationD = iteration;
                                    }
                                }

                                // Store result to array
                                if (res == 1) imageDataTemp[r - 1, c - 1] = iterationD;
                                else reduceResolution(ref imageDataTemp, r, c, iterationD, res);

                            }
                        }
                    }  //next c
                });  //next r
                excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
        }

        public static void mPert(ref double[,] imageData, ref int[,] glitchFlag, ref bool flagGlitch, bool flagSmooth, int flagDllOption, double x0, double y0, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int dr, int dc, int navSize, int rr1, int rr2, int cc1, int cc2, int rRef, int cRef, int numPointsMP,
            double[,] xMP, double[,] yMP, double[,] glitchDetect, double divThreshold, bool flagDebugGlitch, int iteration, int maxPastRef, int[,] prevGlitchFlag)
        {  // Mandelbrot set, perturbation methods
            double pixelStepP = scaleTerm * res;
            int rUpperIndex = ((rr2 - rr1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double zrP = x0 + scaleTerm * cc1;  //min real position
            double ziP = y0 - scaleTerm * rr1;  //max imaginary position
            int bailout = 4;
            if (flagSmooth == true) bailout = highBailout;

            // The following are used to avoid passing variables by reference in the Parallel.For loop, array bases are unchanged when using the Clone method
            double[,] imageDataTemp = imageData.Clone() as double[,];
            int[,] glitchFlagTemp = glitchFlag.Clone() as int[,];
            flagGlitch = false;  //this is set in the Parallel.For loop using the Interlocked.Increment Method on cntFlagGlitch and then checking for non-zero values
            int cntFlagGlitch = 0;

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);
            else excelApp.StatusBar = string.Format("Correcting initial results ({1})... 0%      DLL in use on {0} logical processors.{2}", numProcessors, numPointsMP, dllOptionString);

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int rP = res * ri + rr1;       //mapping the parallel loop index to the row number relative to the reference point
                    double y0P = ziP - pixelStepP * ri;
                    double x0P = zrP - pixelStepP;
                    double x, y;
                    double xP = 0, xtempP, yP = 0;
                    double Dx, Dxtemp, Dy, Dx0, Dy0;
                    int iterationP = 0;
                    double iterationPD = 0;
                    bool skipCalc, goToGlitch;
                    double log_zn, nu;

                    for (int cP = cc1; cP <= cc2; cP = cP + res) {
                        skipCalc = false;
                        x0P = x0P + pixelStepP;
                        if ((cP == 0 && rP == 0) == false) {
                            if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] == prevMaxIter) || (refreshType == 2 && numPointsMP > 1)) {

                                // Quick pan code
                                if (refreshType == -1 && numPointsMP == 1) {
                                    skipCalc = quickPan(dr, dc, rRef + rP, cRef + cP, navSize);
                                }

                                // If correcting a flagged glitch, only correct location that is flagged
                                if (numPointsMP > 1) {
                                    if (prevGlitchFlag[rRef + rP, cRef + cP] == 0) skipCalc = true;
                                }

                                // Iteration loop
                                if (skipCalc == false) {
                                    Dx0 = cP * scaleTerm;
                                    Dy0 = -rP * scaleTerm;
                                    Dx = Dx0;
                                    Dy = Dy0;
                                    goToGlitch = false;

                                    for (int j = 1; j <= iteration; j++) {
                                        iterationP = j;
                                        x = xMP[j, 1];
                                        y = yMP[j, 1];
                                        xP = Dx + x;
                                        yP = Dy + y;
                                        if (xP * xP + yP * yP < glitchDetect[j, 1]) {
                                            glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                            Interlocked.Increment(ref cntFlagGlitch);
                                            if (flagDebugGlitch == true) iterationPD = maxIter;
                                            goToGlitch = true;
                                            break;
                                        }
                                        if (xP * xP + yP * yP > bailout) break;
                                        Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0;
                                        Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dy0;
                                        Dx = Dxtemp;
                                        if (Dx * Dx + Dy * Dy > divThreshold) break;
                                    }  //next j

                                    if (goToGlitch == false) {

                                        // Discrete coloring
                                        if (flagSmooth == false) {
                                            while (xP * xP + yP * yP < 4 && iterationP < maxIter) {
                                                xtempP = xP * xP - yP * yP + x0P;
                                                yP = 2 * xP * yP + y0P;
                                                xP = xtempP;
                                                iterationP++;
                                            }
                                            iterationPD = iterationP;
                                        }

                                        // Continuous coloring
                                        else {
                                            while (xP * xP + yP * yP < highBailout && iterationP < maxIter) {
                                                xtempP = xP * xP - yP * yP + x0P;
                                                yP = 2 * xP * yP + y0P;
                                                xP = xtempP;
                                                iterationP++;
                                            }

                                            // Additional calculations for continuous coloring
                                            if (iterationP < maxIter) {
                                                log_zn = Math.Log(xP * xP + yP * yP) / 2;
                                                nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                                iterationPD = iterationP + 1 - nu;
                                                iterationPD = Math.Round(iterationPD, 3);
                                            }
                                            else {
                                                iterationPD = iterationP;
                                            }
                                        }

                                    }

                                    // Check if too many additional iterations past the reference were performed
                                    if (maxPastRef != 0 && iterationP > maxPastRef) {
                                        glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                        Interlocked.Increment(ref cntFlagGlitch);
                                        if (flagDebugGlitch == true) iterationPD = maxIter;
                                    }

                                    // Store result to array
                                    if (res == 1) imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] = iterationPD;
                                    else reduceResolution(ref imageDataTemp, rRef + rP, cRef + cP, iterationPD, res);
                                }
                            }
                        }

                    }  //next c
                });  //next r
                if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
                else excelApp.StatusBar = string.Format("Correcting initial results ({2})... {0:0%}      DLL in use on {1} logical processors.{3}", (i + 1) / (double)n, numProcessors, numPointsMP, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
            copyArray(ref glitchFlag, glitchFlagTemp);
            if (cntFlagGlitch > 0) flagGlitch = true;
        }

        public static void mSeries3(ref double[,] imageData, ref int[,] glitchFlag, ref bool flagGlitch, ref int numSkip, bool flagSmooth, int flagDllOption, double x0, double y0, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int dr, int dc, int navSize, int rr1, int rr2, int cc1, int cc2, int rRef, int cRef, int numPointsMP, double seriesTolerance,
            double[,] xMP, double[,] yMP, double[,] glitchDetect, double divThreshold, bool flagDebugGlitch, int iteration, int maxPastRef, int[,] prevGlitchFlag, int prevGlitchInfo2)
        {  // Mandelbrot set, perturbation with forced 3rd order series approximation
            double pixelStepP = scaleTerm * res;
            int rUpperIndex = ((rr2 - rr1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double zrP = x0 + scaleTerm * cc1;  //min real position
            double ziP = y0 - scaleTerm * rr1;  //max imaginary position
            bool skipSeriesFlag = false;
            bool flagOverflow = false;
            int maxRefSeries = 0;
            int skipNumber = 0;
            double[] A1r = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
            double[] A2r = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
            double[] A1i = new double[iteration];
            double[] A2i = new double[iteration];
            double[] B1r = new double[iteration];
            double[] B2r = new double[iteration];
            double[] B3r = new double[iteration];
            double[] B1i = new double[iteration];
            double[] B2i = new double[iteration];
            double[] B3i = new double[iteration];
            double[] C1r = new double[iteration];
            double[] C2r = new double[iteration];
            double[] C3r = new double[iteration];
            double[] C4r = new double[iteration];
            double[] C1i = new double[iteration];
            double[] C2i = new double[iteration];
            double[] C3i = new double[iteration];
            double[] C4i = new double[iteration];
            int bailout = 4;
            if (flagSmooth == true) bailout = highBailout;

            // The following are used to avoid passing variables by reference in the Parallel.For loop, array bases are unchanged when using the Clone method
            double[,] imageDataTemp = imageData.Clone() as double[,];
            int[,] glitchFlagTemp = glitchFlag.Clone() as int[,];
            flagGlitch = false;  //this is set in the Parallel.For loop using the Interlocked.Increment Method on cntFlagGlitch and then checking for non-zero values
            int cntFlagGlitch = 0;

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);
            else excelApp.StatusBar = string.Format("Correcting initial results ({1})... 0%      DLL in use on {0} logical processors.{2}", numProcessors, numPointsMP, dllOptionString);

            // Calculate coefficients of series approximation up to a 3rd order approximation
            // When correcting glitches, series approximation is skipped if the largest glitch is small
            if (numPointsMP == 1) {
                if (iteration < 10) skipSeriesFlag = true;
            }
            else {
                if (iteration < 10 || prevGlitchInfo2 < 10) skipSeriesFlag = true;
            }
            if (skipSeriesFlag == false) {
                A1r[1] = 2 * xMP[1, 1] + 1;
                A2r[1] = -2 * yMP[1, 1];
                B1r[1] = 1;
                B2r[1] = -1;
                A1i[1] = -A2r[1];
                A2i[1] = A1r[1];
                B3i[1] = 2;
                for (int i = 2; i <= iteration - 1; i++) {
                    maxRefSeries = i;
                    A1r[i] = 2 * (-A1i[i - 1] * yMP[i, 1] + A1r[i - 1] * xMP[i, 1]) + 1;
                    A2r[i] = 2 * (-A2i[i - 1] * yMP[i, 1] + A2r[i - 1] * xMP[i, 1]);
                    B1r[i] = -A1i[i - 1] * A1i[i - 1] + A1r[i - 1] * A1r[i - 1] + 2 * (-B1i[i - 1] * yMP[i, 1] + B1r[i - 1] * xMP[i, 1]);
                    B2r[i] = -A2i[i - 1] * A2i[i - 1] + A2r[i - 1] * A2r[i - 1] + 2 * (-B2i[i - 1] * yMP[i, 1] + B2r[i - 1] * xMP[i, 1]);
                    B3r[i] = 2 * (-A1i[i - 1] * A2i[i - 1] + A1r[i - 1] * A2r[i - 1] - B3i[i - 1] * yMP[i, 1] + B3r[i - 1] * xMP[i, 1]);
                    C1r[i] = 2 * (-A1i[i - 1] * B1i[i - 1] + A1r[i - 1] * B1r[i - 1] - C1i[i - 1] * yMP[i, 1] + C1r[i - 1] * xMP[i, 1]);
                    C2r[i] = 2 * (-A2i[i - 1] * B2i[i - 1] + A2r[i - 1] * B2r[i - 1] - C2i[i - 1] * yMP[i, 1] + C2r[i - 1] * xMP[i, 1]);
                    C3r[i] = 2 * (-A1i[i - 1] * B3i[i - 1] + A1r[i - 1] * B3r[i - 1] - A2i[i - 1] * B1i[i - 1] + A2r[i - 1] * B1r[i - 1] - C3i[i - 1] * yMP[i, 1] + C3r[i - 1] * xMP[i, 1]);
                    C4r[i] = 2 * (-A1i[i - 1] * B2i[i - 1] + A1r[i - 1] * B2r[i - 1] - A2i[i - 1] * B3i[i - 1] + A2r[i - 1] * B3r[i - 1] - C4i[i - 1] * yMP[i, 1] + C4r[i - 1] * xMP[i, 1]);
                    A1i[i] = 2 * (A1i[i - 1] * xMP[i, 1] + A1r[i - 1] * yMP[i, 1]);
                    A2i[i] = 2 * (A2i[i - 1] * xMP[i, 1] + A2r[i - 1] * yMP[i, 1]) + 1;
                    B1i[i] = 2 * (A1i[i - 1] * A1r[i - 1] + B1i[i - 1] * xMP[i, 1] + B1r[i - 1] * yMP[i, 1]);
                    B2i[i] = 2 * (A2i[i - 1] * A2r[i - 1] + B2i[i - 1] * xMP[i, 1] + B2r[i - 1] * yMP[i, 1]);
                    B3i[i] = 2 * (A1i[i - 1] * A2r[i - 1] + A1r[i - 1] * A2i[i - 1] + B3i[i - 1] * xMP[i, 1] + B3r[i - 1] * yMP[i, 1]);
                    C1i[i] = 2 * (A1i[i - 1] * B1r[i - 1] + A1r[i - 1] * B1i[i - 1] + C1i[i - 1] * xMP[i, 1] + C1r[i - 1] * yMP[i, 1]);
                    C2i[i] = 2 * (A2i[i - 1] * B2r[i - 1] + A2r[i - 1] * B2i[i - 1] + C2i[i - 1] * xMP[i, 1] + C2r[i - 1] * yMP[i, 1]);
                    C3i[i] = 2 * (A1i[i - 1] * B3r[i - 1] + A1r[i - 1] * B3i[i - 1] + A2i[i - 1] * B1r[i - 1] + A2r[i - 1] * B1i[i - 1] + C3i[i - 1] * xMP[i, 1] + C3r[i - 1] * yMP[i, 1]);
                    C4i[i] = 2 * (A1i[i - 1] * B2r[i - 1] + A1r[i - 1] * B2i[i - 1] + A2i[i - 1] * B3r[i - 1] + A2r[i - 1] * B3i[i - 1] + C4i[i - 1] * xMP[i, 1] + C4r[i - 1] * yMP[i, 1]);
                    // Check for overflow, unlike VBA an exception is not thrown when a double exceeds the maximum magnitude
                    if (Double.IsInfinity(C4i[i]) || Double.IsInfinity(C4r[i]) || Double.IsInfinity(C3i[i]) || Double.IsInfinity(C3r[i]) || Double.IsInfinity(C2i[i]) || Double.IsInfinity(C2r[i]) ||
                        Double.IsInfinity(C1i[i]) || Double.IsInfinity(C1r[i]) || Double.IsInfinity(B3i[i]) || Double.IsInfinity(B3r[i]) || Double.IsInfinity(B2i[i]) || Double.IsInfinity(B2r[i]) ||
                        Double.IsInfinity(B1i[i]) || Double.IsInfinity(B1r[i]) || Double.IsInfinity(A2i[i]) || Double.IsInfinity(A2r[i]) || Double.IsInfinity(A1i[i]) || Double.IsInfinity(A1r[i])) {
                        flagOverflow = true;
                        break;
                    }
                }  //next i

                // Catch overflow condition with flag
                if (flagOverflow == true) {
                    if (maxRefSeries < 10) skipSeriesFlag = true;
                    maxRefSeries = maxRefSeries - 1;
                }
            }

            // Determine how many iterations to skip
            if (skipSeriesFlag == false) {
                for (int j = 1; j <= 4; j++) {  //4 corner pixels of range
                    int rP = 0, cP = 0;
                    if (j == 1) {  //top left, min Re & max Im
                        rP = rr1;
                        cP = cc1;
                    }
                    else if (j == 2) {  //top right, max Re & max Im
                        rP = rr1;
                        cP = cc2;
                    }
                    else if (j == 3) {  //bottom left, min Re & min Im
                        rP = rr2;
                        cP = cc1;
                    }
                    else if (j == 4) {  //bottom right, max Re & min Im
                        rP = rr2;
                        cP = cc2;
                    }
                    // Note rr1 and cc1 are negative while rr2 and cc2 are positive
                    double x0P = x0 + scaleTerm * cP;  //Re position
                    double y0P = y0 - scaleTerm * rP;  //Im position

                    // Iteration loop using perturbation method, save delta values
                    double[] DxP = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
                    double[] DyP = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
                    double x, y, Dxtemp;
                    int i2 = 1;
                    double xP = x0P;
                    double yP = y0P;
                    double Dx0 = cP * scaleTerm;
                    double Dy0 = -rP * scaleTerm;
                    double Dx = Dx0;
                    double Dy = Dy0;
                    bool flagBreak = false;
                    DxP[0] = Dx;
                    DyP[0] = Dy;
                    for (int i = 1; i <= maxRefSeries - 1; i++) {
                        i2 = i;
                        x = xMP[i, 1];
                        y = yMP[i, 1];
                        xP = Dx + x;
                        yP = Dy + y;
                        if (xP * xP + yP * yP < glitchDetect[i, 1]) {
                            flagBreak = true;
                            break;
                        }
                        if (xP * xP + yP * yP > bailout) {
                            flagBreak = true;
                            break;
                        }
                        Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0;
                        Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dy0;
                        Dx = Dxtemp;
                        if (Dx * Dx + Dy * Dy > divThreshold) {
                            flagBreak = true;
                            break;
                        }
                        DxP[i] = Dx;
                        DyP[i] = Dy;
                    }  //next i
                    if (flagBreak == false) i2++;

                    // Compare delta values calculated using series approximation (DxS, DyS) to delta values calculated using
                    // perturbation theory (DxP, DyP) to determine how long the series approximation is valid
                    if (i2 < (10 + 1)) {
                        skipSeriesFlag = true;
                        break;
                    }
                    int k2 = 1;
                    double DxS, DyS;
                    flagBreak = false;
                    for (int k = 1; k <= i2 - 2; k++) {
                        k2 = k + 1;
                        DxS = A1r[k] * Dx0 + A2r[k] * Dy0 + B1r[k] * Dx0 * Dx0 + B2r[k] * Dy0 * Dy0 + B3r[k] * Dx0 * Dy0 + C1r[k] * Dx0 * Dx0 * Dx0 + C2r[k] * Dy0 * Dy0 * Dy0 + C3r[k] * Dx0 * Dx0 * Dy0 + C4r[k] * Dx0 * Dy0 * Dy0;
                        DyS = A1i[k] * Dx0 + A2i[k] * Dy0 + B1i[k] * Dx0 * Dx0 + B2i[k] * Dy0 * Dy0 + B3i[k] * Dx0 * Dy0 + C1i[k] * Dx0 * Dx0 * Dx0 + C2i[k] * Dy0 * Dy0 * Dy0 + C3i[k] * Dx0 * Dx0 * Dy0 + C4i[k] * Dx0 * Dy0 * Dy0;
                        if (DxP[k] == 0 || DyP[k] == 0) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DxP[k] - DxS) / DxP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DyP[k] - DyS) / DyP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                    }
                    if (flagBreak == true) k2 = k2 - 1;  //go back to when relative difference was below tolerance, this line is surprisingly important
                    if (k2 < 5) {
                        skipSeriesFlag = true;
                        skipNumber = 0;
                        break;
                    }
                    if (j == 1) {
                        skipNumber = k2;
                    }
                    else {
                        if (k2 < skipNumber) skipNumber = k2;
                    }
                }  //next j
            }
            if (numPointsMP == 1) numSkip = skipNumber;  //store number of iterations skipped in initial pass for calculation info output

            // Calculate points about reference point orbit in double precision

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int rP = res * ri + rr1;       //mapping the parallel loop index to the row number relative to the reference point
                    double y0P = ziP - pixelStepP * ri;
                    double x0P = zrP - pixelStepP;
                    double x, y;
                    double xP = 0, xtempP, yP = 0;
                    double Dx = 0, Dxtemp, Dy = 0, Dx0, Dy0;
                    int iterationP = 0;
                    bool skipCalc, goToGlitch;
                    double iterationPD = 0;
                    double log_zn, nu;

                    for (int cP = cc1; cP <= cc2; cP = cP + res) {
                        skipCalc = false;
                        x0P = x0P + pixelStepP;
                        if ((cP == 0 && rP == 0) == false) {
                            if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] == prevMaxIter) || (refreshType == 2 && numPointsMP > 1)) {

                                // Quick pan code
                                if (refreshType == -1 && numPointsMP == 1) {
                                    skipCalc = quickPan(dr, dc, rRef + rP, cRef + cP, navSize);
                                }

                                // If correcting a flagged glitch, only correct location that is flagged
                                if (numPointsMP > 1) {
                                    if (prevGlitchFlag[rRef + rP, cRef + cP] == 0) skipCalc = true;
                                }

                                // Series approximation to skip iterations
                                if (skipCalc == false) {
                                    int m = skipNumber;
                                    int k = 1;
                                    bool tryLessFlag = false;
                                    bool goToTryLess = true;
                                    Dx0 = cP * scaleTerm;
                                    Dy0 = -rP * scaleTerm;
                                    if (skipSeriesFlag == false) {
                                        while (goToTryLess == true) {
                                            goToTryLess = false;
                                            Dx = A1r[m - 1] * Dx0 + A2r[m - 1] * Dy0 + B1r[m - 1] * Dx0 * Dx0 + B2r[m - 1] * Dy0 * Dy0 + B3r[m - 1] * Dx0 * Dy0 + C1r[m - 1] * Dx0 * Dx0 * Dx0 + C2r[m - 1] * Dy0 * Dy0 * Dy0 + C3r[m - 1] * Dx0 * Dx0 * Dy0 + C4r[m - 1] * Dx0 * Dy0 * Dy0;
                                            Dy = A1i[m - 1] * Dx0 + A2i[m - 1] * Dy0 + B1i[m - 1] * Dx0 * Dx0 + B2i[m - 1] * Dy0 * Dy0 + B3i[m - 1] * Dx0 * Dy0 + C1i[m - 1] * Dx0 * Dx0 * Dx0 + C2i[m - 1] * Dy0 * Dy0 * Dy0 + C3i[m - 1] * Dx0 * Dx0 * Dy0 + C4i[m - 1] * Dx0 * Dy0 * Dy0;
                                            if (Dx * Dx + Dy * Dy > divThreshold) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            xP = Dx + xMP[m, 1];
                                            yP = Dy + yMP[m, 1];
                                            if (xP * xP + yP * yP > bailout) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            if (goToTryLess == false) {
                                                k = m;  //skip first m-1 iterations
                                            }
                                            else {
                                                m = m / 2;
                                            }
                                        }  //Wend
                                    }

                                    // Iteration loop
                                    if (k == 1) {  //no series approximation used
                                        Dx = Dx0;
                                        Dy = Dy0;
                                    }

                                    goToGlitch = false;
                                    for (int j = k; j <= iteration; j++) {
                                        iterationP = j;
                                        x = xMP[j, 1];
                                        y = yMP[j, 1];
                                        xP = Dx + x;
                                        yP = Dy + y;
                                        if (xP * xP + yP * yP < glitchDetect[j, 1]) {
                                            glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                            Interlocked.Increment(ref cntFlagGlitch);
                                            if (flagDebugGlitch == true) iterationPD = maxIter;
                                            goToGlitch = true;
                                            break;
                                        }
                                        if (xP * xP + yP * yP > bailout) break;
                                        Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0;
                                        Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dy0;
                                        Dx = Dxtemp;
                                        if (Dx * Dx + Dy * Dy > divThreshold) break;
                                    }  //next j

                                    if (goToGlitch == false) {

                                        // Discrete coloring
                                        if (flagSmooth == false) {
                                            while (xP * xP + yP * yP < 4 && iterationP < maxIter) {
                                                xtempP = xP * xP - yP * yP + x0P;
                                                yP = 2 * xP * yP + y0P;
                                                xP = xtempP;
                                                iterationP++;
                                            }
                                            iterationPD = iterationP;
                                        }

                                        // Continuous coloring
                                        else {
                                            while (xP * xP + yP * yP < highBailout && iterationP < maxIter) {
                                                xtempP = xP * xP - yP * yP + x0P;
                                                yP = 2 * xP * yP + y0P;
                                                xP = xtempP;
                                                iterationP++;
                                            }

                                            // Additional calculations for continuous coloring
                                            if (iterationP < maxIter) {
                                                log_zn = Math.Log(xP * xP + yP * yP) / 2;
                                                nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                                iterationPD = iterationP + 1 - nu;
                                                iterationPD = Math.Round(iterationPD, 3);
                                            }
                                            else {
                                                iterationPD = iterationP;
                                            }
                                        }

                                    }

                                    // Check if too many additional iterations past the reference were performed
                                    if (maxPastRef != 0 && iterationP > maxPastRef) {
                                        glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                        Interlocked.Increment(ref cntFlagGlitch);
                                        if (flagDebugGlitch == true) iterationPD = maxIter;
                                    }

                                    // Store result to array
                                    if (res == 1) imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] = iterationPD;
                                    else reduceResolution(ref imageDataTemp, rRef + rP, cRef + cP, iterationPD, res);
                                }
                            }
                        }

                    }  //next c
                });  //next r
                if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
                else excelApp.StatusBar = string.Format("Correcting initial results ({2})... {0:0%}      DLL in use on {1} logical processors.{3}", (i + 1) / (double)n, numProcessors, numPointsMP, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
            copyArray(ref glitchFlag, glitchFlagTemp);
            if (cntFlagGlitch > 0) flagGlitch = true;
        }

        public static void mSeries5(ref double[,] imageData, ref int[,] glitchFlag, ref bool flagGlitch, ref int numSkip, bool flagSmooth, int flagDllOption, double x0, double y0, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int dr, int dc, int navSize, int rr1, int rr2, int cc1, int cc2, int rRef, int cRef, int numPointsMP, double seriesTolerance,
            double[,] xMP, double[,] yMP, double[,] glitchDetect, double divThreshold, bool flagDebugGlitch, int iteration, int maxPastRef, int[,] prevGlitchFlag, int prevGlitchInfo2)
        {  // Mandelbrot set, perturbation with forced 5th order series approximation
            double pixelStepP = scaleTerm * res;
            int rUpperIndex = ((rr2 - rr1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double zrP = x0 + scaleTerm * cc1;  //min real position
            double ziP = y0 - scaleTerm * rr1;  //max imaginary position
            bool skipSeriesFlag = false;
            bool flagOverflow = false;
            int maxRefSeries = 0;
            int skipNumber = 0;
            double[] A1r = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
            double[] A2r = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
            double[] A1i = new double[iteration];
            double[] A2i = new double[iteration];
            double[] B1r = new double[iteration];
            double[] B2r = new double[iteration];
            double[] B3r = new double[iteration];
            double[] B1i = new double[iteration];
            double[] B2i = new double[iteration];
            double[] B3i = new double[iteration];
            double[] C1r = new double[iteration];
            double[] C2r = new double[iteration];
            double[] C3r = new double[iteration];
            double[] C4r = new double[iteration];
            double[] C1i = new double[iteration];
            double[] C2i = new double[iteration];
            double[] C3i = new double[iteration];
            double[] C4i = new double[iteration];
            double[] D1r = new double[iteration];
            double[] D2r = new double[iteration];
            double[] D3r = new double[iteration];
            double[] D4r = new double[iteration];
            double[] D5r = new double[iteration];
            double[] D1i = new double[iteration];
            double[] D2i = new double[iteration];
            double[] D3i = new double[iteration];
            double[] D4i = new double[iteration];
            double[] D5i = new double[iteration];
            double[] E1r = new double[iteration];
            double[] E2r = new double[iteration];
            double[] E3r = new double[iteration];
            double[] E4r = new double[iteration];
            double[] E5r = new double[iteration];
            double[] E6r = new double[iteration];
            double[] E1i = new double[iteration];
            double[] E2i = new double[iteration];
            double[] E3i = new double[iteration];
            double[] E4i = new double[iteration];
            double[] E5i = new double[iteration];
            double[] E6i = new double[iteration];
            int bailout = 4;
            if (flagSmooth == true) bailout = highBailout;

            // The following are used to avoid passing variables by reference in the Parallel.For loop, array bases are unchanged when using the Clone method
            double[,] imageDataTemp = imageData.Clone() as double[,];
            int[,] glitchFlagTemp = glitchFlag.Clone() as int[,];
            flagGlitch = false;  //this is set in the Parallel.For loop using the Interlocked.Increment Method on cntFlagGlitch and then checking for non-zero values
            int cntFlagGlitch = 0;

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);
            else excelApp.StatusBar = string.Format("Correcting initial results ({1})... 0%      DLL in use on {0} logical processors.{2}", numProcessors, numPointsMP, dllOptionString);

            // Calculate coefficients of series approximation up to a 5th order approximation
            // When correcting glitches, series approximation is skipped if the largest glitch is small
            if (numPointsMP == 1) {
                if (iteration < 10) skipSeriesFlag = true;
            }
            else {
                if (iteration < 10 || prevGlitchInfo2 < 10) skipSeriesFlag = true;
            }
            if (skipSeriesFlag == false) {
                A1r[1] = 2 * xMP[1, 1] + 1;
                A2r[1] = -2 * yMP[1, 1];
                B1r[1] = 1;
                B2r[1] = -1;
                A1i[1] = -A2r[1];
                A2i[1] = A1r[1];
                B3i[1] = 2;
                for (int i = 2; i <= iteration - 1; i++) {
                    maxRefSeries = i;
                    A1r[i] = 2 * (-A1i[i - 1] * yMP[i, 1] + A1r[i - 1] * xMP[i, 1]) + 1;
                    A2r[i] = 2 * (-A2i[i - 1] * yMP[i, 1] + A2r[i - 1] * xMP[i, 1]);
                    B1r[i] = -A1i[i - 1] * A1i[i - 1] + A1r[i - 1] * A1r[i - 1] + 2 * (-B1i[i - 1] * yMP[i, 1] + B1r[i - 1] * xMP[i, 1]);
                    B2r[i] = -A2i[i - 1] * A2i[i - 1] + A2r[i - 1] * A2r[i - 1] + 2 * (-B2i[i - 1] * yMP[i, 1] + B2r[i - 1] * xMP[i, 1]);
                    B3r[i] = 2 * (-A1i[i - 1] * A2i[i - 1] + A1r[i - 1] * A2r[i - 1] - B3i[i - 1] * yMP[i, 1] + B3r[i - 1] * xMP[i, 1]);
                    C1r[i] = 2 * (-A1i[i - 1] * B1i[i - 1] + A1r[i - 1] * B1r[i - 1] - C1i[i - 1] * yMP[i, 1] + C1r[i - 1] * xMP[i, 1]);
                    C2r[i] = 2 * (-A2i[i - 1] * B2i[i - 1] + A2r[i - 1] * B2r[i - 1] - C2i[i - 1] * yMP[i, 1] + C2r[i - 1] * xMP[i, 1]);
                    C3r[i] = 2 * (-A1i[i - 1] * B3i[i - 1] + A1r[i - 1] * B3r[i - 1] - A2i[i - 1] * B1i[i - 1] + A2r[i - 1] * B1r[i - 1] - C3i[i - 1] * yMP[i, 1] + C3r[i - 1] * xMP[i, 1]);
                    C4r[i] = 2 * (-A1i[i - 1] * B2i[i - 1] + A1r[i - 1] * B2r[i - 1] - A2i[i - 1] * B3i[i - 1] + A2r[i - 1] * B3r[i - 1] - C4i[i - 1] * yMP[i, 1] + C4r[i - 1] * xMP[i, 1]);
                    A1i[i] = 2 * (A1i[i - 1] * xMP[i, 1] + A1r[i - 1] * yMP[i, 1]);
                    A2i[i] = 2 * (A2i[i - 1] * xMP[i, 1] + A2r[i - 1] * yMP[i, 1]) + 1;
                    B1i[i] = 2 * (A1i[i - 1] * A1r[i - 1] + B1i[i - 1] * xMP[i, 1] + B1r[i - 1] * yMP[i, 1]);
                    B2i[i] = 2 * (A2i[i - 1] * A2r[i - 1] + B2i[i - 1] * xMP[i, 1] + B2r[i - 1] * yMP[i, 1]);
                    B3i[i] = 2 * (A1i[i - 1] * A2r[i - 1] + A1r[i - 1] * A2i[i - 1] + B3i[i - 1] * xMP[i, 1] + B3r[i - 1] * yMP[i, 1]);
                    C1i[i] = 2 * (A1i[i - 1] * B1r[i - 1] + A1r[i - 1] * B1i[i - 1] + C1i[i - 1] * xMP[i, 1] + C1r[i - 1] * yMP[i, 1]);
                    C2i[i] = 2 * (A2i[i - 1] * B2r[i - 1] + A2r[i - 1] * B2i[i - 1] + C2i[i - 1] * xMP[i, 1] + C2r[i - 1] * yMP[i, 1]);
                    C3i[i] = 2 * (A1i[i - 1] * B3r[i - 1] + A1r[i - 1] * B3i[i - 1] + A2i[i - 1] * B1r[i - 1] + A2r[i - 1] * B1i[i - 1] + C3i[i - 1] * xMP[i, 1] + C3r[i - 1] * yMP[i, 1]);
                    C4i[i] = 2 * (A1i[i - 1] * B2r[i - 1] + A1r[i - 1] * B2i[i - 1] + A2i[i - 1] * B3r[i - 1] + A2r[i - 1] * B3i[i - 1] + C4i[i - 1] * xMP[i, 1] + C4r[i - 1] * yMP[i, 1]);
                    D1r[i] = -B1i[i - 1] * B1i[i - 1] + B1r[i - 1] * B1r[i - 1] + 2 * (-A1i[i - 1] * C1i[i - 1] + A1r[i - 1] * C1r[i - 1] - D1i[i - 1] * yMP[i, 1] + D1r[i - 1] * xMP[i, 1]);
                    D2r[i] = -B2i[i - 1] * B2i[i - 1] + B2r[i - 1] * B2r[i - 1] + 2 * (-A2i[i - 1] * C2i[i - 1] + A2r[i - 1] * C2r[i - 1] - D2i[i - 1] * yMP[i, 1] + D2r[i - 1] * xMP[i, 1]);
                    D3r[i] = 2 * (-A1i[i - 1] * C3i[i - 1] + A1r[i - 1] * C3r[i - 1] - A2i[i - 1] * C1i[i - 1] + A2r[i - 1] * C1r[i - 1] - B1i[i - 1] * B3i[i - 1] + B1r[i - 1] * B3r[i - 1] - D3i[i - 1] * yMP[i, 1] + D3r[i - 1] * xMP[i, 1]);
                    D4r[i] = 2 * (-A1i[i - 1] * C2i[i - 1] + A1r[i - 1] * C2r[i - 1] - A2i[i - 1] * C4i[i - 1] + A2r[i - 1] * C4r[i - 1] - B2i[i - 1] * B3i[i - 1] + B2r[i - 1] * B3r[i - 1] - D4i[i - 1] * yMP[i, 1] + D4r[i - 1] * xMP[i, 1]);
                    D5r[i] = -B3i[i - 1] * B3i[i - 1] + B3r[i - 1] * B3r[i - 1] + 2 * (-A1i[i - 1] * C4i[i - 1] + A1r[i - 1] * C4r[i - 1] - A2i[i - 1] * C3i[i - 1] + A2r[i - 1] * C3r[i - 1] - B1i[i - 1] * B2i[i - 1] + B1r[i - 1] * B2r[i - 1] - D5i[i - 1] * yMP[i, 1] + D5r[i - 1] * xMP[i, 1]);
                    E1r[i] = 2 * (-A1i[i - 1] * D1i[i - 1] + A1r[i - 1] * D1r[i - 1] - B1i[i - 1] * C1i[i - 1] + B1r[i - 1] * C1r[i - 1] - E1i[i - 1] * yMP[i, 1] + E1r[i - 1] * xMP[i, 1]);
                    E2r[i] = 2 * (-A2i[i - 1] * D2i[i - 1] + A2r[i - 1] * D2r[i - 1] - B2i[i - 1] * C2i[i - 1] + B2r[i - 1] * C2r[i - 1] - E2i[i - 1] * yMP[i, 1] + E2r[i - 1] * xMP[i, 1]);
                    E3r[i] = 2 * (-A1i[i - 1] * D3i[i - 1] + A1r[i - 1] * D3r[i - 1] - A2i[i - 1] * D1i[i - 1] + A2r[i - 1] * D1r[i - 1] - B1i[i - 1] * C3i[i - 1] + B1r[i - 1] * C3r[i - 1] - B3i[i - 1] * C1i[i - 1] + B3r[i - 1] * C1r[i - 1] - E3i[i - 1] * yMP[i, 1] + E3r[i - 1] * xMP[i, 1]);
                    E4r[i] = 2 * (-A1i[i - 1] * D2i[i - 1] + A1r[i - 1] * D2r[i - 1] - A2i[i - 1] * D4i[i - 1] + A2r[i - 1] * D4r[i - 1] - B2i[i - 1] * C4i[i - 1] + B2r[i - 1] * C4r[i - 1] - B3i[i - 1] * C2i[i - 1] + B3r[i - 1] * C2r[i - 1] - E4i[i - 1] * yMP[i, 1] + E4r[i - 1] * xMP[i, 1]);
                    E5r[i] = 2 * (-A1i[i - 1] * D5i[i - 1] + A1r[i - 1] * D5r[i - 1] - A2i[i - 1] * D3i[i - 1] + A2r[i - 1] * D3r[i - 1] - B1i[i - 1] * C4i[i - 1] + B1r[i - 1] * C4r[i - 1] - B2i[i - 1] * C1i[i - 1] + B2r[i - 1] * C1r[i - 1] - B3i[i - 1] * C3i[i - 1] + B3r[i - 1] * C3r[i - 1] - E5i[i - 1] * yMP[i, 1] + E5r[i - 1] * xMP[i, 1]);
                    E6r[i] = 2 * (-A1i[i - 1] * D4i[i - 1] + A1r[i - 1] * D4r[i - 1] - A2i[i - 1] * D5i[i - 1] + A2r[i - 1] * D5r[i - 1] - B1i[i - 1] * C2i[i - 1] + B1r[i - 1] * C2r[i - 1] - B2i[i - 1] * C3i[i - 1] + B2r[i - 1] * C3r[i - 1] - B3i[i - 1] * C4i[i - 1] + B3r[i - 1] * C4r[i - 1] - E6i[i - 1] * yMP[i, 1] + E6r[i - 1] * xMP[i, 1]);
                    D1i[i] = 2 * (A1i[i - 1] * C1r[i - 1] + A1r[i - 1] * C1i[i - 1] + B1i[i - 1] * B1r[i - 1] + D1i[i - 1] * xMP[i, 1] + D1r[i - 1] * yMP[i, 1]);
                    D2i[i] = 2 * (A2i[i - 1] * C2r[i - 1] + A2r[i - 1] * C2i[i - 1] + B2i[i - 1] * B2r[i - 1] + D2i[i - 1] * xMP[i, 1] + D2r[i - 1] * yMP[i, 1]);
                    D3i[i] = 2 * (A1i[i - 1] * C3r[i - 1] + A1r[i - 1] * C3i[i - 1] + A2i[i - 1] * C1r[i - 1] + A2r[i - 1] * C1i[i - 1] + B1i[i - 1] * B3r[i - 1] + B1r[i - 1] * B3i[i - 1] + D3i[i - 1] * xMP[i, 1] + D3r[i - 1] * yMP[i, 1]);
                    D4i[i] = 2 * (A1i[i - 1] * C2r[i - 1] + A1r[i - 1] * C2i[i - 1] + A2i[i - 1] * C4r[i - 1] + A2r[i - 1] * C4i[i - 1] + B2i[i - 1] * B3r[i - 1] + B2r[i - 1] * B3i[i - 1] + D4i[i - 1] * xMP[i, 1] + D4r[i - 1] * yMP[i, 1]);
                    D5i[i] = 2 * (A1i[i - 1] * C4r[i - 1] + A1r[i - 1] * C4i[i - 1] + A2i[i - 1] * C3r[i - 1] + A2r[i - 1] * C3i[i - 1] + B1i[i - 1] * B2r[i - 1] + B1r[i - 1] * B2i[i - 1] + B3i[i - 1] * B3r[i - 1] + D5i[i - 1] * xMP[i, 1] + D5r[i - 1] * yMP[i, 1]);
                    E1i[i] = 2 * (A1i[i - 1] * D1r[i - 1] + A1r[i - 1] * D1i[i - 1] + B1i[i - 1] * C1r[i - 1] + B1r[i - 1] * C1i[i - 1] + E1i[i - 1] * xMP[i, 1] + E1r[i - 1] * yMP[i, 1]);
                    E2i[i] = 2 * (A2i[i - 1] * D2r[i - 1] + A2r[i - 1] * D2i[i - 1] + B2i[i - 1] * C2r[i - 1] + B2r[i - 1] * C2i[i - 1] + E2i[i - 1] * xMP[i, 1] + E2r[i - 1] * yMP[i, 1]);
                    E3i[i] = 2 * (A1i[i - 1] * D3r[i - 1] + A1r[i - 1] * D3i[i - 1] + A2i[i - 1] * D1r[i - 1] + A2r[i - 1] * D1i[i - 1] + B1i[i - 1] * C3r[i - 1] + B1r[i - 1] * C3i[i - 1] + B3i[i - 1] * C1r[i - 1] + B3r[i - 1] * C1i[i - 1] + E3i[i - 1] * xMP[i, 1] + E3r[i - 1] * yMP[i, 1]);
                    E4i[i] = 2 * (A1i[i - 1] * D2r[i - 1] + A1r[i - 1] * D2i[i - 1] + A2i[i - 1] * D4r[i - 1] + A2r[i - 1] * D4i[i - 1] + B2i[i - 1] * C4r[i - 1] + B2r[i - 1] * C4i[i - 1] + B3i[i - 1] * C2r[i - 1] + B3r[i - 1] * C2i[i - 1] + E4i[i - 1] * xMP[i, 1] + E4r[i - 1] * yMP[i, 1]);
                    E5i[i] = 2 * (A1i[i - 1] * D5r[i - 1] + A1r[i - 1] * D5i[i - 1] + A2i[i - 1] * D3r[i - 1] + A2r[i - 1] * D3i[i - 1] + B1i[i - 1] * C4r[i - 1] + B1r[i - 1] * C4i[i - 1] + B2i[i - 1] * C1r[i - 1] + B2r[i - 1] * C1i[i - 1] + B3i[i - 1] * C3r[i - 1] + B3r[i - 1] * C3i[i - 1] + E5i[i - 1] * xMP[i, 1] + E5r[i - 1] * yMP[i, 1]);
                    E6i[i] = 2 * (A1i[i - 1] * D4r[i - 1] + A1r[i - 1] * D4i[i - 1] + A2i[i - 1] * D5r[i - 1] + A2r[i - 1] * D5i[i - 1] + B1i[i - 1] * C2r[i - 1] + B1r[i - 1] * C2i[i - 1] + B2i[i - 1] * C3r[i - 1] + B2r[i - 1] * C3i[i - 1] + B3i[i - 1] * C4r[i - 1] + B3r[i - 1] * C4i[i - 1] + E6i[i - 1] * xMP[i, 1] + E6r[i - 1] * yMP[i, 1]);
                    // Check for overflow, unlike VBA an exception is not thrown when a double exceeds the maximum magnitude
                    if (Double.IsInfinity(E6i[i]) || Double.IsInfinity(E6r[i]) || Double.IsInfinity(E5i[i]) || Double.IsInfinity(E5r[i]) || Double.IsInfinity(E4i[i]) || Double.IsInfinity(E4r[i]) ||
                        Double.IsInfinity(E3i[i]) || Double.IsInfinity(E3r[i]) || Double.IsInfinity(E2i[i]) || Double.IsInfinity(E2r[i]) || Double.IsInfinity(E1i[i]) || Double.IsInfinity(E1r[i]) ||
                        Double.IsInfinity(D5i[i]) || Double.IsInfinity(D5r[i]) || Double.IsInfinity(D4i[i]) || Double.IsInfinity(D4r[i]) || Double.IsInfinity(D3i[i]) || Double.IsInfinity(D3r[i]) ||
                        Double.IsInfinity(D2i[i]) || Double.IsInfinity(D2r[i]) || Double.IsInfinity(D1i[i]) || Double.IsInfinity(D1r[i]) ||
                        Double.IsInfinity(C4i[i]) || Double.IsInfinity(C4r[i]) || Double.IsInfinity(C3i[i]) || Double.IsInfinity(C3r[i]) || Double.IsInfinity(C2i[i]) || Double.IsInfinity(C2r[i]) ||
                        Double.IsInfinity(C1i[i]) || Double.IsInfinity(C1r[i]) || Double.IsInfinity(B3i[i]) || Double.IsInfinity(B3r[i]) || Double.IsInfinity(B2i[i]) || Double.IsInfinity(B2r[i]) ||
                        Double.IsInfinity(B1i[i]) || Double.IsInfinity(B1r[i]) || Double.IsInfinity(A2i[i]) || Double.IsInfinity(A2r[i]) || Double.IsInfinity(A1i[i]) || Double.IsInfinity(A1r[i])) {
                        flagOverflow = true;
                        break;
                    }
                }  //next i

                // Catch overflow condition with flag
                if (flagOverflow == true) {
                    if (maxRefSeries < 10) skipSeriesFlag = true;
                    maxRefSeries = maxRefSeries - 1;
                }
            }

            // Determine how many iterations to skip
            if (skipSeriesFlag == false) {
                for (int j = 1; j <= 4; j++) {  //4 corner pixels of range
                    int rP = 0, cP = 0;
                    if (j == 1) {  //top left, min Re & max Im
                        rP = rr1;
                        cP = cc1;
                    }
                    else if (j == 2) {  //top right, max Re & max Im
                        rP = rr1;
                        cP = cc2;
                    }
                    else if (j == 3) {  //bottom left, min Re & min Im
                        rP = rr2;
                        cP = cc1;
                    }
                    else if (j == 4) {  //bottom right, max Re & min Im
                        rP = rr2;
                        cP = cc2;
                    }
                    // Note rr1 and cc1 are negative while rr2 and cc2 are positive
                    double x0P = x0 + scaleTerm * cP;  //Re position
                    double y0P = y0 - scaleTerm * rP;  //Im position

                    // Iteration loop using perturbation method, save delta values
                    double[] DxP = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
                    double[] DyP = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
                    double x, y, Dxtemp;
                    int i2 = 1;
                    double xP = x0P;
                    double yP = y0P;
                    double Dx0 = cP * scaleTerm;
                    double Dy0 = -rP * scaleTerm;
                    double Dx = Dx0;
                    double Dy = Dy0;
                    bool flagBreak = false;
                    DxP[0] = Dx;
                    DyP[0] = Dy;
                    for (int i = 1; i <= maxRefSeries - 1; i++) {
                        i2 = i;
                        x = xMP[i, 1];
                        y = yMP[i, 1];
                        xP = Dx + x;
                        yP = Dy + y;
                        if (xP * xP + yP * yP < glitchDetect[i, 1]) {
                            flagBreak = true;
                            break;
                        }
                        if (xP * xP + yP * yP > bailout) {
                            flagBreak = true;
                            break;
                        }
                        Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0;
                        Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dy0;
                        Dx = Dxtemp;
                        if (Dx * Dx + Dy * Dy > divThreshold) {
                            flagBreak = true;
                            break;
                        }
                        DxP[i] = Dx;
                        DyP[i] = Dy;
                    }  //next i
                    if (flagBreak == false) i2++;

                    // Compare delta values calculated using series approximation (DxS, DyS) to delta values calculated using
                    // perturbation theory (DxP, DyP) to determine how long the series approximation is valid
                    if (i2 < (10 + 1)) {
                        skipSeriesFlag = true;
                        break;
                    }
                    int k2 = 1;
                    double DxS, DyS;
                    flagBreak = false;
                    for (int k = 1; k <= i2 - 2; k++) {
                        k2 = k + 1;
                        DxS = A1r[k] * Dx0 + A2r[k] * Dy0 + B1r[k] * Dx0 * Dx0 + B2r[k] * Dy0 * Dy0 + B3r[k] * Dx0 * Dy0 + C1r[k] * Dx0 * Dx0 * Dx0 + C2r[k] * Dy0 * Dy0 * Dy0 + C3r[k] * Dx0 * Dx0 * Dy0 + C4r[k] * Dx0 * Dy0 * Dy0 + D1r[k] * Dx0 * Dx0 * Dx0 * Dx0 + D2r[k] * Dy0 * Dy0 * Dy0 * Dy0 + D3r[k] * Dx0 * Dx0 * Dx0 * Dy0 + D4r[k] * Dx0 * Dy0 * Dy0 * Dy0 + D5r[k] * Dx0 * Dx0 * Dy0 * Dy0 + E1r[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2r[k] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3r[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4r[k] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5r[k] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6r[k] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                        DyS = A1i[k] * Dx0 + A2i[k] * Dy0 + B1i[k] * Dx0 * Dx0 + B2i[k] * Dy0 * Dy0 + B3i[k] * Dx0 * Dy0 + C1i[k] * Dx0 * Dx0 * Dx0 + C2i[k] * Dy0 * Dy0 * Dy0 + C3i[k] * Dx0 * Dx0 * Dy0 + C4i[k] * Dx0 * Dy0 * Dy0 + D1i[k] * Dx0 * Dx0 * Dx0 * Dx0 + D2i[k] * Dy0 * Dy0 * Dy0 * Dy0 + D3i[k] * Dx0 * Dx0 * Dx0 * Dy0 + D4i[k] * Dx0 * Dy0 * Dy0 * Dy0 + D5i[k] * Dx0 * Dx0 * Dy0 * Dy0 + E1i[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2i[k] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3i[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4i[k] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5i[k] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6i[k] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                        if (DxP[k] == 0 || DyP[k] == 0) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DxP[k] - DxS) / DxP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DyP[k] - DyS) / DyP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                    }
                    if (flagBreak == true) k2 = k2 - 1;  //go back to when relative difference was below tolerance, this line is surprisingly important
                    if (k2 < 5) {
                        skipSeriesFlag = true;
                        skipNumber = 0;
                        break;
                    }
                    if (j == 1) {
                        skipNumber = k2;
                    }
                    else {
                        if (k2 < skipNumber) skipNumber = k2;
                    }
                }  //next j
            }
            if (numPointsMP == 1) numSkip = skipNumber;  //store number of iterations skipped in initial pass for calculation info output

            // Calculate points about reference point orbit in double precision

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int rP = res * ri + rr1;       //mapping the parallel loop index to the row number relative to the reference point
                    double y0P = ziP - pixelStepP * ri;
                    double x0P = zrP - pixelStepP;
                    double x, y;
                    double xP = 0, xtempP, yP = 0;
                    double Dx = 0, Dxtemp, Dy = 0, Dx0, Dy0;
                    int iterationP = 0;
                    bool skipCalc, goToGlitch;
                    double iterationPD = 0;
                    double log_zn, nu;

                    for (int cP = cc1; cP <= cc2; cP = cP + res) {
                        skipCalc = false;
                        x0P = x0P + pixelStepP;
                        if ((cP == 0 && rP == 0) == false) {
                            if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] == prevMaxIter) || (refreshType == 2 && numPointsMP > 1)) {

                                // Quick pan code
                                if (refreshType == -1 && numPointsMP == 1) {
                                    skipCalc = quickPan(dr, dc, rRef + rP, cRef + cP, navSize);
                                }

                                // If correcting a flagged glitch, only correct location that is flagged
                                if (numPointsMP > 1) {
                                    if (prevGlitchFlag[rRef + rP, cRef + cP] == 0) skipCalc = true;
                                }

                                // Series approximation to skip iterations
                                if (skipCalc == false) {
                                    int m = skipNumber;
                                    int k = 1;
                                    bool tryLessFlag = false;
                                    bool goToTryLess = true;
                                    Dx0 = cP * scaleTerm;
                                    Dy0 = -rP * scaleTerm;
                                    if (skipSeriesFlag == false) {
                                        while (goToTryLess == true) {
                                            goToTryLess = false;
                                            Dx = A1r[m - 1] * Dx0 + A2r[m - 1] * Dy0 + B1r[m - 1] * Dx0 * Dx0 + B2r[m - 1] * Dy0 * Dy0 + B3r[m - 1] * Dx0 * Dy0 + C1r[m - 1] * Dx0 * Dx0 * Dx0 + C2r[m - 1] * Dy0 * Dy0 * Dy0 + C3r[m - 1] * Dx0 * Dx0 * Dy0 + C4r[m - 1] * Dx0 * Dy0 * Dy0 + D1r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 + D2r[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 + D3r[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 + D4r[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 + D5r[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 + E1r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2r[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4r[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5r[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6r[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                                            Dy = A1i[m - 1] * Dx0 + A2i[m - 1] * Dy0 + B1i[m - 1] * Dx0 * Dx0 + B2i[m - 1] * Dy0 * Dy0 + B3i[m - 1] * Dx0 * Dy0 + C1i[m - 1] * Dx0 * Dx0 * Dx0 + C2i[m - 1] * Dy0 * Dy0 * Dy0 + C3i[m - 1] * Dx0 * Dx0 * Dy0 + C4i[m - 1] * Dx0 * Dy0 * Dy0 + D1i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 + D2i[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 + D3i[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 + D4i[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 + D5i[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 + E1i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2i[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4i[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5i[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6i[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                                            if (Dx * Dx + Dy * Dy > divThreshold) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            xP = Dx + xMP[m, 1];
                                            yP = Dy + yMP[m, 1];
                                            if (xP * xP + yP * yP > bailout) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            if (goToTryLess == false) {
                                                k = m;  //skip first m-1 iterations
                                            }
                                            else {
                                                m = m / 2;
                                            }
                                        }  //Wend
                                    }

                                    // Iteration loop
                                    if (k == 1) {  //no series approximation used
                                        Dx = Dx0;
                                        Dy = Dy0;
                                    }

                                    goToGlitch = false;
                                    for (int j = k; j <= iteration; j++) {
                                        iterationP = j;
                                        x = xMP[j, 1];
                                        y = yMP[j, 1];
                                        xP = Dx + x;
                                        yP = Dy + y;
                                        if (xP * xP + yP * yP < glitchDetect[j, 1]) {
                                            glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                            Interlocked.Increment(ref cntFlagGlitch);
                                            if (flagDebugGlitch == true) iterationPD = maxIter;
                                            goToGlitch = true;
                                            break;
                                        }
                                        if (xP * xP + yP * yP > bailout) break;
                                        Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0;
                                        Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dy0;
                                        Dx = Dxtemp;
                                        if (Dx * Dx + Dy * Dy > divThreshold) break;
                                    }  //next j

                                    if (goToGlitch == false) {

                                        // Discrete coloring
                                        if (flagSmooth == false) {
                                            while (xP * xP + yP * yP < 4 && iterationP < maxIter) {
                                                xtempP = xP * xP - yP * yP + x0P;
                                                yP = 2 * xP * yP + y0P;
                                                xP = xtempP;
                                                iterationP++;
                                            }
                                            iterationPD = iterationP;
                                        }

                                        // Continuous coloring
                                        else {
                                            while (xP * xP + yP * yP < highBailout && iterationP < maxIter) {
                                                xtempP = xP * xP - yP * yP + x0P;
                                                yP = 2 * xP * yP + y0P;
                                                xP = xtempP;
                                                iterationP++;
                                            }

                                            // Additional calculations for continuous coloring
                                            if (iterationP < maxIter) {
                                                log_zn = Math.Log(xP * xP + yP * yP) / 2;
                                                nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                                iterationPD = iterationP + 1 - nu;
                                                iterationPD = Math.Round(iterationPD, 3);
                                            }
                                            else {
                                                iterationPD = iterationP;
                                            }
                                        }

                                    }

                                    // Check if too many additional iterations past the reference were performed
                                    if (maxPastRef != 0 && iterationP > maxPastRef) {
                                        glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                        Interlocked.Increment(ref cntFlagGlitch);
                                        if (flagDebugGlitch == true) iterationPD = maxIter;
                                    }

                                    // Store result to array
                                    if (res == 1) imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] = iterationPD;
                                    else reduceResolution(ref imageDataTemp, rRef + rP, cRef + cP, iterationPD, res);
                                }
                            }
                        }

                    }  //next c
                });  //next r
                if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
                else excelApp.StatusBar = string.Format("Correcting initial results ({2})... {0:0%}      DLL in use on {1} logical processors.{3}", (i + 1) / (double)n, numProcessors, numPointsMP, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
            copyArray(ref glitchFlag, glitchFlagTemp);
            if (cntFlagGlitch > 0) flagGlitch = true;
        }

        public static void mSeries(ref double[,] imageData, ref int[,] glitchFlag, ref bool flagGlitch, ref int numSkip, bool flagSmooth, int flagDllOption, double x0, double y0, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int dr, int dc, int navSize, int rr1, int rr2, int cc1, int cc2, int rRef, int cRef, int numPointsMP, double seriesTolerance,
            double[,] xMP, double[,] yMP, double[,] glitchDetect, double divThreshold, bool flagDebugGlitch, int iteration, int maxPastRef, int[,] prevGlitchFlag, int prevGlitchInfo2)
        {  // Mandelbrot set, perturbation with series approximation
            double pixelStepP = scaleTerm * res;
            int rUpperIndex = ((rr2 - rr1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double zrP = x0 + scaleTerm * cc1;  //min real position
            double ziP = y0 - scaleTerm * rr1;  //max imaginary position
            bool skipSeriesFlag = false;
            bool flagOverflow = false;
            int maxRefSeries = 0;
            int skipNumber = 0;
            double[] A1r = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
            double[] A2r = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
            double[] A1i = new double[iteration];
            double[] A2i = new double[iteration];
            double[] B1r = new double[iteration];
            double[] B2r = new double[iteration];
            double[] B3r = new double[iteration];
            double[] B1i = new double[iteration];
            double[] B2i = new double[iteration];
            double[] B3i = new double[iteration];
            double[] C1r = new double[iteration];
            double[] C2r = new double[iteration];
            double[] C3r = new double[iteration];
            double[] C4r = new double[iteration];
            double[] C1i = new double[iteration];
            double[] C2i = new double[iteration];
            double[] C3i = new double[iteration];
            double[] C4i = new double[iteration];
            double[] D1r = new double[iteration];
            double[] D2r = new double[iteration];
            double[] D3r = new double[iteration];
            double[] D4r = new double[iteration];
            double[] D5r = new double[iteration];
            double[] D1i = new double[iteration];
            double[] D2i = new double[iteration];
            double[] D3i = new double[iteration];
            double[] D4i = new double[iteration];
            double[] D5i = new double[iteration];
            double[] E1r = new double[iteration];
            double[] E2r = new double[iteration];
            double[] E3r = new double[iteration];
            double[] E4r = new double[iteration];
            double[] E5r = new double[iteration];
            double[] E6r = new double[iteration];
            double[] E1i = new double[iteration];
            double[] E2i = new double[iteration];
            double[] E3i = new double[iteration];
            double[] E4i = new double[iteration];
            double[] E5i = new double[iteration];
            double[] E6i = new double[iteration];
            bool calculatingO3Flag = false, calcO5flag = false, useOrder3 = false;
            int maxO3Series = 0, maxO5Series = 0, maxSeries = 0;
            int skipNumberO3 = 0, skipNumberO5 = 0;
            int i3 = 0;
            int bailout = 4;
            if (flagSmooth == true) bailout = highBailout;

            // The following are used to avoid passing variables by reference in the Parallel.For loop, array bases are unchanged when using the Clone method
            double[,] imageDataTemp = imageData.Clone() as double[,];
            int[,] glitchFlagTemp = glitchFlag.Clone() as int[,];
            flagGlitch = false;  //this is set in the Parallel.For loop using the Interlocked.Increment Method on cntFlagGlitch and then checking for non-zero values
            int cntFlagGlitch = 0;

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);
            else excelApp.StatusBar = string.Format("Correcting initial results ({1})... 0%      DLL in use on {0} logical processors.{2}", numProcessors, numPointsMP, dllOptionString);

            // Calculate coefficients of series approximation up to a 5th order approximation
            // When correcting glitches, series approximation is skipped if the largest glitch is small
            if (numPointsMP == 1) {
                if (iteration < 10) skipSeriesFlag = true;
            }
            else {
                if (iteration < 10 || prevGlitchInfo2 < 10) skipSeriesFlag = true;
            }
            if (skipSeriesFlag == false) {
                A1r[1] = 2 * xMP[1, 1] + 1;
                A2r[1] = -2 * yMP[1, 1];
                B1r[1] = 1;
                B2r[1] = -1;
                A1i[1] = -A2r[1];
                A2i[1] = A1r[1];
                B3i[1] = 2;
                calcO5flag = true;
                for (int i = 2; i <= iteration - 1; i++) {
                    maxRefSeries = i;
                    calculatingO3Flag = true;
                    A1r[i] = 2 * (-A1i[i - 1] * yMP[i, 1] + A1r[i - 1] * xMP[i, 1]) + 1;
                    A2r[i] = 2 * (-A2i[i - 1] * yMP[i, 1] + A2r[i - 1] * xMP[i, 1]);
                    B1r[i] = -A1i[i - 1] * A1i[i - 1] + A1r[i - 1] * A1r[i - 1] + 2 * (-B1i[i - 1] * yMP[i, 1] + B1r[i - 1] * xMP[i, 1]);
                    B2r[i] = -A2i[i - 1] * A2i[i - 1] + A2r[i - 1] * A2r[i - 1] + 2 * (-B2i[i - 1] * yMP[i, 1] + B2r[i - 1] * xMP[i, 1]);
                    B3r[i] = 2 * (-A1i[i - 1] * A2i[i - 1] + A1r[i - 1] * A2r[i - 1] - B3i[i - 1] * yMP[i, 1] + B3r[i - 1] * xMP[i, 1]);
                    C1r[i] = 2 * (-A1i[i - 1] * B1i[i - 1] + A1r[i - 1] * B1r[i - 1] - C1i[i - 1] * yMP[i, 1] + C1r[i - 1] * xMP[i, 1]);
                    C2r[i] = 2 * (-A2i[i - 1] * B2i[i - 1] + A2r[i - 1] * B2r[i - 1] - C2i[i - 1] * yMP[i, 1] + C2r[i - 1] * xMP[i, 1]);
                    C3r[i] = 2 * (-A1i[i - 1] * B3i[i - 1] + A1r[i - 1] * B3r[i - 1] - A2i[i - 1] * B1i[i - 1] + A2r[i - 1] * B1r[i - 1] - C3i[i - 1] * yMP[i, 1] + C3r[i - 1] * xMP[i, 1]);
                    C4r[i] = 2 * (-A1i[i - 1] * B2i[i - 1] + A1r[i - 1] * B2r[i - 1] - A2i[i - 1] * B3i[i - 1] + A2r[i - 1] * B3r[i - 1] - C4i[i - 1] * yMP[i, 1] + C4r[i - 1] * xMP[i, 1]);
                    A1i[i] = 2 * (A1i[i - 1] * xMP[i, 1] + A1r[i - 1] * yMP[i, 1]);
                    A2i[i] = 2 * (A2i[i - 1] * xMP[i, 1] + A2r[i - 1] * yMP[i, 1]) + 1;
                    B1i[i] = 2 * (A1i[i - 1] * A1r[i - 1] + B1i[i - 1] * xMP[i, 1] + B1r[i - 1] * yMP[i, 1]);
                    B2i[i] = 2 * (A2i[i - 1] * A2r[i - 1] + B2i[i - 1] * xMP[i, 1] + B2r[i - 1] * yMP[i, 1]);
                    B3i[i] = 2 * (A1i[i - 1] * A2r[i - 1] + A1r[i - 1] * A2i[i - 1] + B3i[i - 1] * xMP[i, 1] + B3r[i - 1] * yMP[i, 1]);
                    C1i[i] = 2 * (A1i[i - 1] * B1r[i - 1] + A1r[i - 1] * B1i[i - 1] + C1i[i - 1] * xMP[i, 1] + C1r[i - 1] * yMP[i, 1]);
                    C2i[i] = 2 * (A2i[i - 1] * B2r[i - 1] + A2r[i - 1] * B2i[i - 1] + C2i[i - 1] * xMP[i, 1] + C2r[i - 1] * yMP[i, 1]);
                    C3i[i] = 2 * (A1i[i - 1] * B3r[i - 1] + A1r[i - 1] * B3i[i - 1] + A2i[i - 1] * B1r[i - 1] + A2r[i - 1] * B1i[i - 1] + C3i[i - 1] * xMP[i, 1] + C3r[i - 1] * yMP[i, 1]);
                    C4i[i] = 2 * (A1i[i - 1] * B2r[i - 1] + A1r[i - 1] * B2i[i - 1] + A2i[i - 1] * B3r[i - 1] + A2r[i - 1] * B3i[i - 1] + C4i[i - 1] * xMP[i, 1] + C4r[i - 1] * yMP[i, 1]);
                    // Check for overflow, unlike VBA an exception is not thrown when a double exceeds the maximum magnitude
                    if (Double.IsInfinity(C4i[i]) || Double.IsInfinity(C4r[i]) || Double.IsInfinity(C3i[i]) || Double.IsInfinity(C3r[i]) || Double.IsInfinity(C2i[i]) || Double.IsInfinity(C2r[i]) ||
                        Double.IsInfinity(C1i[i]) || Double.IsInfinity(C1r[i]) || Double.IsInfinity(B3i[i]) || Double.IsInfinity(B3r[i]) || Double.IsInfinity(B2i[i]) || Double.IsInfinity(B2r[i]) ||
                        Double.IsInfinity(B1i[i]) || Double.IsInfinity(B1r[i]) || Double.IsInfinity(A2i[i]) || Double.IsInfinity(A2r[i]) || Double.IsInfinity(A1i[i]) || Double.IsInfinity(A1r[i])) {
                        flagOverflow = true;
                        break;
                    }
                    calculatingO3Flag = false;
                    D1r[i] = -B1i[i - 1] * B1i[i - 1] + B1r[i - 1] * B1r[i - 1] + 2 * (-A1i[i - 1] * C1i[i - 1] + A1r[i - 1] * C1r[i - 1] - D1i[i - 1] * yMP[i, 1] + D1r[i - 1] * xMP[i, 1]);
                    D2r[i] = -B2i[i - 1] * B2i[i - 1] + B2r[i - 1] * B2r[i - 1] + 2 * (-A2i[i - 1] * C2i[i - 1] + A2r[i - 1] * C2r[i - 1] - D2i[i - 1] * yMP[i, 1] + D2r[i - 1] * xMP[i, 1]);
                    D3r[i] = 2 * (-A1i[i - 1] * C3i[i - 1] + A1r[i - 1] * C3r[i - 1] - A2i[i - 1] * C1i[i - 1] + A2r[i - 1] * C1r[i - 1] - B1i[i - 1] * B3i[i - 1] + B1r[i - 1] * B3r[i - 1] - D3i[i - 1] * yMP[i, 1] + D3r[i - 1] * xMP[i, 1]);
                    D4r[i] = 2 * (-A1i[i - 1] * C2i[i - 1] + A1r[i - 1] * C2r[i - 1] - A2i[i - 1] * C4i[i - 1] + A2r[i - 1] * C4r[i - 1] - B2i[i - 1] * B3i[i - 1] + B2r[i - 1] * B3r[i - 1] - D4i[i - 1] * yMP[i, 1] + D4r[i - 1] * xMP[i, 1]);
                    D5r[i] = -B3i[i - 1] * B3i[i - 1] + B3r[i - 1] * B3r[i - 1] + 2 * (-A1i[i - 1] * C4i[i - 1] + A1r[i - 1] * C4r[i - 1] - A2i[i - 1] * C3i[i - 1] + A2r[i - 1] * C3r[i - 1] - B1i[i - 1] * B2i[i - 1] + B1r[i - 1] * B2r[i - 1] - D5i[i - 1] * yMP[i, 1] + D5r[i - 1] * xMP[i, 1]);
                    E1r[i] = 2 * (-A1i[i - 1] * D1i[i - 1] + A1r[i - 1] * D1r[i - 1] - B1i[i - 1] * C1i[i - 1] + B1r[i - 1] * C1r[i - 1] - E1i[i - 1] * yMP[i, 1] + E1r[i - 1] * xMP[i, 1]);
                    E2r[i] = 2 * (-A2i[i - 1] * D2i[i - 1] + A2r[i - 1] * D2r[i - 1] - B2i[i - 1] * C2i[i - 1] + B2r[i - 1] * C2r[i - 1] - E2i[i - 1] * yMP[i, 1] + E2r[i - 1] * xMP[i, 1]);
                    E3r[i] = 2 * (-A1i[i - 1] * D3i[i - 1] + A1r[i - 1] * D3r[i - 1] - A2i[i - 1] * D1i[i - 1] + A2r[i - 1] * D1r[i - 1] - B1i[i - 1] * C3i[i - 1] + B1r[i - 1] * C3r[i - 1] - B3i[i - 1] * C1i[i - 1] + B3r[i - 1] * C1r[i - 1] - E3i[i - 1] * yMP[i, 1] + E3r[i - 1] * xMP[i, 1]);
                    E4r[i] = 2 * (-A1i[i - 1] * D2i[i - 1] + A1r[i - 1] * D2r[i - 1] - A2i[i - 1] * D4i[i - 1] + A2r[i - 1] * D4r[i - 1] - B2i[i - 1] * C4i[i - 1] + B2r[i - 1] * C4r[i - 1] - B3i[i - 1] * C2i[i - 1] + B3r[i - 1] * C2r[i - 1] - E4i[i - 1] * yMP[i, 1] + E4r[i - 1] * xMP[i, 1]);
                    E5r[i] = 2 * (-A1i[i - 1] * D5i[i - 1] + A1r[i - 1] * D5r[i - 1] - A2i[i - 1] * D3i[i - 1] + A2r[i - 1] * D3r[i - 1] - B1i[i - 1] * C4i[i - 1] + B1r[i - 1] * C4r[i - 1] - B2i[i - 1] * C1i[i - 1] + B2r[i - 1] * C1r[i - 1] - B3i[i - 1] * C3i[i - 1] + B3r[i - 1] * C3r[i - 1] - E5i[i - 1] * yMP[i, 1] + E5r[i - 1] * xMP[i, 1]);
                    E6r[i] = 2 * (-A1i[i - 1] * D4i[i - 1] + A1r[i - 1] * D4r[i - 1] - A2i[i - 1] * D5i[i - 1] + A2r[i - 1] * D5r[i - 1] - B1i[i - 1] * C2i[i - 1] + B1r[i - 1] * C2r[i - 1] - B2i[i - 1] * C3i[i - 1] + B2r[i - 1] * C3r[i - 1] - B3i[i - 1] * C4i[i - 1] + B3r[i - 1] * C4r[i - 1] - E6i[i - 1] * yMP[i, 1] + E6r[i - 1] * xMP[i, 1]);
                    D1i[i] = 2 * (A1i[i - 1] * C1r[i - 1] + A1r[i - 1] * C1i[i - 1] + B1i[i - 1] * B1r[i - 1] + D1i[i - 1] * xMP[i, 1] + D1r[i - 1] * yMP[i, 1]);
                    D2i[i] = 2 * (A2i[i - 1] * C2r[i - 1] + A2r[i - 1] * C2i[i - 1] + B2i[i - 1] * B2r[i - 1] + D2i[i - 1] * xMP[i, 1] + D2r[i - 1] * yMP[i, 1]);
                    D3i[i] = 2 * (A1i[i - 1] * C3r[i - 1] + A1r[i - 1] * C3i[i - 1] + A2i[i - 1] * C1r[i - 1] + A2r[i - 1] * C1i[i - 1] + B1i[i - 1] * B3r[i - 1] + B1r[i - 1] * B3i[i - 1] + D3i[i - 1] * xMP[i, 1] + D3r[i - 1] * yMP[i, 1]);
                    D4i[i] = 2 * (A1i[i - 1] * C2r[i - 1] + A1r[i - 1] * C2i[i - 1] + A2i[i - 1] * C4r[i - 1] + A2r[i - 1] * C4i[i - 1] + B2i[i - 1] * B3r[i - 1] + B2r[i - 1] * B3i[i - 1] + D4i[i - 1] * xMP[i, 1] + D4r[i - 1] * yMP[i, 1]);
                    D5i[i] = 2 * (A1i[i - 1] * C4r[i - 1] + A1r[i - 1] * C4i[i - 1] + A2i[i - 1] * C3r[i - 1] + A2r[i - 1] * C3i[i - 1] + B1i[i - 1] * B2r[i - 1] + B1r[i - 1] * B2i[i - 1] + B3i[i - 1] * B3r[i - 1] + D5i[i - 1] * xMP[i, 1] + D5r[i - 1] * yMP[i, 1]);
                    E1i[i] = 2 * (A1i[i - 1] * D1r[i - 1] + A1r[i - 1] * D1i[i - 1] + B1i[i - 1] * C1r[i - 1] + B1r[i - 1] * C1i[i - 1] + E1i[i - 1] * xMP[i, 1] + E1r[i - 1] * yMP[i, 1]);
                    E2i[i] = 2 * (A2i[i - 1] * D2r[i - 1] + A2r[i - 1] * D2i[i - 1] + B2i[i - 1] * C2r[i - 1] + B2r[i - 1] * C2i[i - 1] + E2i[i - 1] * xMP[i, 1] + E2r[i - 1] * yMP[i, 1]);
                    E3i[i] = 2 * (A1i[i - 1] * D3r[i - 1] + A1r[i - 1] * D3i[i - 1] + A2i[i - 1] * D1r[i - 1] + A2r[i - 1] * D1i[i - 1] + B1i[i - 1] * C3r[i - 1] + B1r[i - 1] * C3i[i - 1] + B3i[i - 1] * C1r[i - 1] + B3r[i - 1] * C1i[i - 1] + E3i[i - 1] * xMP[i, 1] + E3r[i - 1] * yMP[i, 1]);
                    E4i[i] = 2 * (A1i[i - 1] * D2r[i - 1] + A1r[i - 1] * D2i[i - 1] + A2i[i - 1] * D4r[i - 1] + A2r[i - 1] * D4i[i - 1] + B2i[i - 1] * C4r[i - 1] + B2r[i - 1] * C4i[i - 1] + B3i[i - 1] * C2r[i - 1] + B3r[i - 1] * C2i[i - 1] + E4i[i - 1] * xMP[i, 1] + E4r[i - 1] * yMP[i, 1]);
                    E5i[i] = 2 * (A1i[i - 1] * D5r[i - 1] + A1r[i - 1] * D5i[i - 1] + A2i[i - 1] * D3r[i - 1] + A2r[i - 1] * D3i[i - 1] + B1i[i - 1] * C4r[i - 1] + B1r[i - 1] * C4i[i - 1] + B2i[i - 1] * C1r[i - 1] + B2r[i - 1] * C1i[i - 1] + B3i[i - 1] * C3r[i - 1] + B3r[i - 1] * C3i[i - 1] + E5i[i - 1] * xMP[i, 1] + E5r[i - 1] * yMP[i, 1]);
                    E6i[i] = 2 * (A1i[i - 1] * D4r[i - 1] + A1r[i - 1] * D4i[i - 1] + A2i[i - 1] * D5r[i - 1] + A2r[i - 1] * D5i[i - 1] + B1i[i - 1] * C2r[i - 1] + B1r[i - 1] * C2i[i - 1] + B2i[i - 1] * C3r[i - 1] + B2r[i - 1] * C3i[i - 1] + B3i[i - 1] * C4r[i - 1] + B3r[i - 1] * C4i[i - 1] + E6i[i - 1] * xMP[i, 1] + E6r[i - 1] * yMP[i, 1]);
                    if (Double.IsInfinity(E6i[i]) || Double.IsInfinity(E6r[i]) || Double.IsInfinity(E5i[i]) || Double.IsInfinity(E5r[i]) || Double.IsInfinity(E4i[i]) || Double.IsInfinity(E4r[i]) ||
                        Double.IsInfinity(E3i[i]) || Double.IsInfinity(E3r[i]) || Double.IsInfinity(E2i[i]) || Double.IsInfinity(E2r[i]) || Double.IsInfinity(E1i[i]) || Double.IsInfinity(E1r[i]) ||
                        Double.IsInfinity(D5i[i]) || Double.IsInfinity(D5r[i]) || Double.IsInfinity(D4i[i]) || Double.IsInfinity(D4r[i]) || Double.IsInfinity(D3i[i]) || Double.IsInfinity(D3r[i]) ||
                        Double.IsInfinity(D2i[i]) || Double.IsInfinity(D2r[i]) || Double.IsInfinity(D1i[i]) || Double.IsInfinity(D1r[i])) {
                        flagOverflow = true;
                        break;
                    }
                }  //next i

                // Catch overflow condition with flag
                if (flagOverflow == true) {
                    if (calculatingO3Flag == false) {  //overflow occurred in order 4 or 5 term
                        maxO5Series = maxRefSeries - 1;
                        flagOverflow = false;  //reset flag
                        calcO5flag = false;
                        i3 = maxRefSeries;
                    }
                    else {  //overflow occurred in order 3 term
                        maxO3Series = maxRefSeries - 1;
                        maxRefSeries = maxRefSeries - 1;
                    }
                }

                // If overflow occurred in order 4 or 5 term, calculate additional order 3 terms if possible
                if (calcO5flag == false) {  //OverflowReturnIn
                    for (int i = i3 + 1; i <= iteration - 1; i++) {
                        maxRefSeries = i;
                        A1r[i] = 2 * (-A1i[i - 1] * yMP[i, 1] + A1r[i - 1] * xMP[i, 1]) + 1;
                        A2r[i] = 2 * (-A2i[i - 1] * yMP[i, 1] + A2r[i - 1] * xMP[i, 1]);
                        B1r[i] = -A1i[i - 1] * A1i[i - 1] + A1r[i - 1] * A1r[i - 1] + 2 * (-B1i[i - 1] * yMP[i, 1] + B1r[i - 1] * xMP[i, 1]);
                        B2r[i] = -A2i[i - 1] * A2i[i - 1] + A2r[i - 1] * A2r[i - 1] + 2 * (-B2i[i - 1] * yMP[i, 1] + B2r[i - 1] * xMP[i, 1]);
                        B3r[i] = 2 * (-A1i[i - 1] * A2i[i - 1] + A1r[i - 1] * A2r[i - 1] - B3i[i - 1] * yMP[i, 1] + B3r[i - 1] * xMP[i, 1]);
                        C1r[i] = 2 * (-A1i[i - 1] * B1i[i - 1] + A1r[i - 1] * B1r[i - 1] - C1i[i - 1] * yMP[i, 1] + C1r[i - 1] * xMP[i, 1]);
                        C2r[i] = 2 * (-A2i[i - 1] * B2i[i - 1] + A2r[i - 1] * B2r[i - 1] - C2i[i - 1] * yMP[i, 1] + C2r[i - 1] * xMP[i, 1]);
                        C3r[i] = 2 * (-A1i[i - 1] * B3i[i - 1] + A1r[i - 1] * B3r[i - 1] - A2i[i - 1] * B1i[i - 1] + A2r[i - 1] * B1r[i - 1] - C3i[i - 1] * yMP[i, 1] + C3r[i - 1] * xMP[i, 1]);
                        C4r[i] = 2 * (-A1i[i - 1] * B2i[i - 1] + A1r[i - 1] * B2r[i - 1] - A2i[i - 1] * B3i[i - 1] + A2r[i - 1] * B3r[i - 1] - C4i[i - 1] * yMP[i, 1] + C4r[i - 1] * xMP[i, 1]);
                        A1i[i] = 2 * (A1i[i - 1] * xMP[i, 1] + A1r[i - 1] * yMP[i, 1]);
                        A2i[i] = 2 * (A2i[i - 1] * xMP[i, 1] + A2r[i - 1] * yMP[i, 1]) + 1;
                        B1i[i] = 2 * (A1i[i - 1] * A1r[i - 1] + B1i[i - 1] * xMP[i, 1] + B1r[i - 1] * yMP[i, 1]);
                        B2i[i] = 2 * (A2i[i - 1] * A2r[i - 1] + B2i[i - 1] * xMP[i, 1] + B2r[i - 1] * yMP[i, 1]);
                        B3i[i] = 2 * (A1i[i - 1] * A2r[i - 1] + A1r[i - 1] * A2i[i - 1] + B3i[i - 1] * xMP[i, 1] + B3r[i - 1] * yMP[i, 1]);
                        C1i[i] = 2 * (A1i[i - 1] * B1r[i - 1] + A1r[i - 1] * B1i[i - 1] + C1i[i - 1] * xMP[i, 1] + C1r[i - 1] * yMP[i, 1]);
                        C2i[i] = 2 * (A2i[i - 1] * B2r[i - 1] + A2r[i - 1] * B2i[i - 1] + C2i[i - 1] * xMP[i, 1] + C2r[i - 1] * yMP[i, 1]);
                        C3i[i] = 2 * (A1i[i - 1] * B3r[i - 1] + A1r[i - 1] * B3i[i - 1] + A2i[i - 1] * B1r[i - 1] + A2r[i - 1] * B1i[i - 1] + C3i[i - 1] * xMP[i, 1] + C3r[i - 1] * yMP[i, 1]);
                        C4i[i] = 2 * (A1i[i - 1] * B2r[i - 1] + A1r[i - 1] * B2i[i - 1] + A2i[i - 1] * B3r[i - 1] + A2r[i - 1] * B3i[i - 1] + C4i[i - 1] * xMP[i, 1] + C4r[i - 1] * yMP[i, 1]);
                        if (Double.IsInfinity(C4i[i]) || Double.IsInfinity(C4r[i]) || Double.IsInfinity(C3i[i]) || Double.IsInfinity(C3r[i]) || Double.IsInfinity(C2i[i]) || Double.IsInfinity(C2r[i]) ||
                            Double.IsInfinity(C1i[i]) || Double.IsInfinity(C1r[i]) || Double.IsInfinity(B3i[i]) || Double.IsInfinity(B3r[i]) || Double.IsInfinity(B2i[i]) || Double.IsInfinity(B2r[i]) ||
                            Double.IsInfinity(B1i[i]) || Double.IsInfinity(B1r[i]) || Double.IsInfinity(A2i[i]) || Double.IsInfinity(A2r[i]) || Double.IsInfinity(A1i[i]) || Double.IsInfinity(A1r[i])) {
                            flagOverflow = true;
                            break;
                        }
                    }  //next i
                    if (flagOverflow == true) {
                        maxO3Series = maxRefSeries - 1;
                        maxRefSeries = maxRefSeries - 1;
                    }
                }
            }
            if (maxO3Series == 0) maxO3Series = maxRefSeries;
            if (maxO5Series == 0) maxO5Series = maxRefSeries;
            if (maxRefSeries < 10) skipSeriesFlag = true;

            // Determine how many iterations to skip
            if (skipSeriesFlag == false) {
                for (int j = 1; j <= 4; j++) {  //4 corner pixels of range
                    int rP = 0, cP = 0;
                    if (j == 1) {  //top left, min Re & max Im
                        rP = rr1;
                        cP = cc1;
                    }
                    else if (j == 2) {  //top right, max Re & max Im
                        rP = rr1;
                        cP = cc2;
                    }
                    else if (j == 3) {  //bottom left, min Re & min Im
                        rP = rr2;
                        cP = cc1;
                    }
                    else if (j == 4) {  //bottom right, max Re & min Im
                        rP = rr2;
                        cP = cc2;
                    }
                    // Note rr1 and cc1 are negative while rr2 and cc2 are positive
                    double x0P = x0 + scaleTerm * cP;  //Re position
                    double y0P = y0 - scaleTerm * rP;  //Im position

                    // Iteration loop using perturbation method, save delta values
                    double[] DxP = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
                    double[] DyP = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
                    double x, y, Dxtemp;
                    int i2 = 1;
                    double xP = x0P;
                    double yP = y0P;
                    double Dx0 = cP * scaleTerm;
                    double Dy0 = -rP * scaleTerm;
                    double Dx = Dx0;
                    double Dy = Dy0;
                    bool flagBreak = false;
                    DxP[0] = Dx;
                    DyP[0] = Dy;
                    for (int i = 1; i <= maxRefSeries - 1; i++) {
                        i2 = i;
                        x = xMP[i, 1];
                        y = yMP[i, 1];
                        xP = Dx + x;
                        yP = Dy + y;
                        if (xP * xP + yP * yP < glitchDetect[i, 1]) {
                            flagBreak = true;
                            break;
                        }
                        if (xP * xP + yP * yP > bailout) {
                            flagBreak = true;
                            break;
                        }
                        Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0;
                        Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dy0;
                        Dx = Dxtemp;
                        if (Dx * Dx + Dy * Dy > divThreshold) {
                            flagBreak = true;
                            break;
                        }
                        DxP[i] = Dx;
                        DyP[i] = Dy;
                    }  //next i
                    if (flagBreak == false) i2++;

                    // Compare delta values calculated using series approximation (DxS, DyS) to delta values calculated using
                    // perturbation theory (DxP, DyP) to determine how long the series approximation is valid
                    if (i2 < (10 + 1)) {
                        skipSeriesFlag = true;
                        break;
                    }

                    // Order 3 approximation
                    maxSeries = Math.Min(maxO3Series - 1, i2 - 1);
                    int k2 = 1;
                    double DxS, DyS;
                    flagBreak = false;
                    for (int k = 1; k <= maxSeries - 1; k++) {
                        k2 = k + 1;
                        DxS = A1r[k] * Dx0 + A2r[k] * Dy0 + B1r[k] * Dx0 * Dx0 + B2r[k] * Dy0 * Dy0 + B3r[k] * Dx0 * Dy0 + C1r[k] * Dx0 * Dx0 * Dx0 + C2r[k] * Dy0 * Dy0 * Dy0 + C3r[k] * Dx0 * Dx0 * Dy0 + C4r[k] * Dx0 * Dy0 * Dy0;
                        DyS = A1i[k] * Dx0 + A2i[k] * Dy0 + B1i[k] * Dx0 * Dx0 + B2i[k] * Dy0 * Dy0 + B3i[k] * Dx0 * Dy0 + C1i[k] * Dx0 * Dx0 * Dx0 + C2i[k] * Dy0 * Dy0 * Dy0 + C3i[k] * Dx0 * Dx0 * Dy0 + C4i[k] * Dx0 * Dy0 * Dy0;
                        if (DxP[k] == 0 || DyP[k] == 0) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DxP[k] - DxS) / DxP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DyP[k] - DyS) / DyP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                    }
                    if (flagBreak == true) k2 = k2 - 1;  //go back to when relative difference was below tolerance, this line is surprisingly important
                    if (j == 1) {
                        skipNumberO3 = k2;
                    }
                    else {
                        if (k2 < skipNumberO3) skipNumberO3 = k2;
                    }

                    // Order 5 approximation
                    maxSeries = Math.Min(maxO5Series - 1, i2 - 1);
                    int m = k2;  //purpose of 5th order approximation is to skip more iterations so we only need to check additional iterations
                    if (skipNumberO5 == -1) m = maxSeries + 1;  //avoid entering loop if k2 = m on a previous pass
                    flagBreak = false;
                    for (int k = m - 1; k <= maxSeries - 1; k++) {
                        k2 = k + 1;
                        DxS = A1r[k] * Dx0 + A2r[k] * Dy0 + B1r[k] * Dx0 * Dx0 + B2r[k] * Dy0 * Dy0 + B3r[k] * Dx0 * Dy0 + C1r[k] * Dx0 * Dx0 * Dx0 + C2r[k] * Dy0 * Dy0 * Dy0 + C3r[k] * Dx0 * Dx0 * Dy0 + C4r[k] * Dx0 * Dy0 * Dy0 + D1r[k] * Dx0 * Dx0 * Dx0 * Dx0 + D2r[k] * Dy0 * Dy0 * Dy0 * Dy0 + D3r[k] * Dx0 * Dx0 * Dx0 * Dy0 + D4r[k] * Dx0 * Dy0 * Dy0 * Dy0 + D5r[k] * Dx0 * Dx0 * Dy0 * Dy0 + E1r[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2r[k] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3r[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4r[k] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5r[k] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6r[k] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                        DyS = A1i[k] * Dx0 + A2i[k] * Dy0 + B1i[k] * Dx0 * Dx0 + B2i[k] * Dy0 * Dy0 + B3i[k] * Dx0 * Dy0 + C1i[k] * Dx0 * Dx0 * Dx0 + C2i[k] * Dy0 * Dy0 * Dy0 + C3i[k] * Dx0 * Dx0 * Dy0 + C4i[k] * Dx0 * Dy0 * Dy0 + D1i[k] * Dx0 * Dx0 * Dx0 * Dx0 + D2i[k] * Dy0 * Dy0 * Dy0 * Dy0 + D3i[k] * Dx0 * Dx0 * Dx0 * Dy0 + D4i[k] * Dx0 * Dy0 * Dy0 * Dy0 + D5i[k] * Dx0 * Dx0 * Dy0 * Dy0 + E1i[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2i[k] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3i[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4i[k] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5i[k] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6i[k] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                        if (DxP[k] == 0 || DyP[k] == 0) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DxP[k] - DxS) / DxP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DyP[k] - DyS) / DyP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                    }
                    if (k2 == m) k2 = 0;
                    if (flagBreak == true) k2 = k2 - 1;  //go back to when relative difference was below tolerance, this line is surprisingly important
                    if (j == 1) {
                        skipNumberO5 = k2;
                    }
                    else {
                        if (k2 < skipNumberO5) skipNumberO5 = k2;
                    }

                    // Check if it is worth continuing
                    if (Math.Max(skipNumberO3, skipNumberO5) < 5) {
                        skipSeriesFlag = true;
                        skipNumber = 0;
                        break;
                    }

                }  //next j

                // Determine final number of iterations to skip and what order of approximation to use
                // Order 5 calculations are more expensive so they are penalized slightly in the comparison
                if (skipSeriesFlag == false) {
                    if (skipNumberO3 > (skipNumberO5 - 10)) {
                        skipNumber = skipNumberO3;
                        useOrder3 = true;
                    }
                    else {  //use order 5
                        skipNumber = skipNumberO5;
                        useOrder3 = false;
                    }
                }
            }
            if (numPointsMP == 1) numSkip = skipNumber;  //store number of iterations skipped in initial pass for calculation info output

            // Calculate points about reference point orbit in double precision

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int rP = res * ri + rr1;       //mapping the parallel loop index to the row number relative to the reference point
                    double y0P = ziP - pixelStepP * ri;
                    double x0P = zrP - pixelStepP;
                    double x, y;
                    double xP = 0, xtempP, yP = 0;
                    double Dx = 0, Dxtemp, Dy = 0, Dx0, Dy0;
                    int iterationP = 0;
                    bool skipCalc, goToGlitch;
                    double iterationPD = 0;
                    double log_zn, nu;

                    for (int cP = cc1; cP <= cc2; cP = cP + res) {
                        skipCalc = false;
                        x0P = x0P + pixelStepP;
                        if ((cP == 0 && rP == 0) == false) {
                            if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] == prevMaxIter) || (refreshType == 2 && numPointsMP > 1)) {

                                // Quick pan code
                                if (refreshType == -1 && numPointsMP == 1) {
                                    skipCalc = quickPan(dr, dc, rRef + rP, cRef + cP, navSize);
                                }

                                // If correcting a flagged glitch, only correct location that is flagged
                                if (numPointsMP > 1) {
                                    if (prevGlitchFlag[rRef + rP, cRef + cP] == 0) skipCalc = true;
                                }

                                // Series approximation to skip iterations
                                if (skipCalc == false) {
                                    int m = skipNumber;
                                    int k = 1;
                                    bool tryLessFlag = false;
                                    bool goToTryLess = true;
                                    Dx0 = cP * scaleTerm;
                                    Dy0 = -rP * scaleTerm;
                                    if (skipSeriesFlag == false) {
                                        while (goToTryLess == true) {
                                            goToTryLess = false;
                                            if (useOrder3 == true) {
                                                Dx = A1r[m - 1] * Dx0 + A2r[m - 1] * Dy0 + B1r[m - 1] * Dx0 * Dx0 + B2r[m - 1] * Dy0 * Dy0 + B3r[m - 1] * Dx0 * Dy0 + C1r[m - 1] * Dx0 * Dx0 * Dx0 + C2r[m - 1] * Dy0 * Dy0 * Dy0 + C3r[m - 1] * Dx0 * Dx0 * Dy0 + C4r[m - 1] * Dx0 * Dy0 * Dy0;
                                                Dy = A1i[m - 1] * Dx0 + A2i[m - 1] * Dy0 + B1i[m - 1] * Dx0 * Dx0 + B2i[m - 1] * Dy0 * Dy0 + B3i[m - 1] * Dx0 * Dy0 + C1i[m - 1] * Dx0 * Dx0 * Dx0 + C2i[m - 1] * Dy0 * Dy0 * Dy0 + C3i[m - 1] * Dx0 * Dx0 * Dy0 + C4i[m - 1] * Dx0 * Dy0 * Dy0;
                                            }
                                            else {
                                                Dx = A1r[m - 1] * Dx0 + A2r[m - 1] * Dy0 + B1r[m - 1] * Dx0 * Dx0 + B2r[m - 1] * Dy0 * Dy0 + B3r[m - 1] * Dx0 * Dy0 + C1r[m - 1] * Dx0 * Dx0 * Dx0 + C2r[m - 1] * Dy0 * Dy0 * Dy0 + C3r[m - 1] * Dx0 * Dx0 * Dy0 + C4r[m - 1] * Dx0 * Dy0 * Dy0 + D1r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 + D2r[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 + D3r[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 + D4r[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 + D5r[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 + E1r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2r[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4r[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5r[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6r[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                                                Dy = A1i[m - 1] * Dx0 + A2i[m - 1] * Dy0 + B1i[m - 1] * Dx0 * Dx0 + B2i[m - 1] * Dy0 * Dy0 + B3i[m - 1] * Dx0 * Dy0 + C1i[m - 1] * Dx0 * Dx0 * Dx0 + C2i[m - 1] * Dy0 * Dy0 * Dy0 + C3i[m - 1] * Dx0 * Dx0 * Dy0 + C4i[m - 1] * Dx0 * Dy0 * Dy0 + D1i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 + D2i[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 + D3i[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 + D4i[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 + D5i[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 + E1i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2i[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4i[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5i[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6i[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                                            }
                                            if (Dx * Dx + Dy * Dy > divThreshold) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            xP = Dx + xMP[m, 1];
                                            yP = Dy + yMP[m, 1];
                                            if (xP * xP + yP * yP > bailout) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            if (goToTryLess == false) {
                                                k = m;  //skip first m-1 iterations
                                            }
                                            else {
                                                m = m / 2;
                                            }
                                        }  //Wend
                                    }

                                    // Iteration loop
                                    if (k == 1) {  //no series approximation used
                                        Dx = Dx0;
                                        Dy = Dy0;
                                    }

                                    goToGlitch = false;
                                    for (int j = k; j <= iteration; j++) {
                                        iterationP = j;
                                        x = xMP[j, 1];
                                        y = yMP[j, 1];
                                        xP = Dx + x;
                                        yP = Dy + y;
                                        if (xP * xP + yP * yP < glitchDetect[j, 1]) {
                                            glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                            Interlocked.Increment(ref cntFlagGlitch);
                                            if (flagDebugGlitch == true) iterationPD = maxIter;
                                            goToGlitch = true;
                                            break;
                                        }
                                        if (xP * xP + yP * yP > bailout) break;
                                        Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0;
                                        Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dy0;
                                        Dx = Dxtemp;
                                        if (Dx * Dx + Dy * Dy > divThreshold) break;
                                    }  //next j

                                    if (goToGlitch == false) {

                                        // Discrete coloring
                                        if (flagSmooth == false) {
                                            while (xP * xP + yP * yP < 4 && iterationP < maxIter) {
                                                xtempP = xP * xP - yP * yP + x0P;
                                                yP = 2 * xP * yP + y0P;
                                                xP = xtempP;
                                                iterationP++;
                                            }
                                            iterationPD = iterationP;
                                        }

                                        // Continuous coloring
                                        else {
                                            while (xP * xP + yP * yP < highBailout && iterationP < maxIter) {
                                                xtempP = xP * xP - yP * yP + x0P;
                                                yP = 2 * xP * yP + y0P;
                                                xP = xtempP;
                                                iterationP++;
                                            }

                                            // Additional calculations for continuous coloring
                                            if (iterationP < maxIter) {
                                                log_zn = Math.Log(xP * xP + yP * yP) / 2;
                                                nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                                iterationPD = iterationP + 1 - nu;
                                                iterationPD = Math.Round(iterationPD, 3);
                                            }
                                            else {
                                                iterationPD = iterationP;
                                            }
                                        }

                                    }

                                    // Check if too many additional iterations past the reference were performed
                                    if (maxPastRef != 0 && iterationP > maxPastRef) {
                                        glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                        Interlocked.Increment(ref cntFlagGlitch);
                                        if (flagDebugGlitch == true) iterationPD = maxIter;
                                    }

                                    // Store result to array
                                    if (res == 1) imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] = iterationPD;
                                    else reduceResolution(ref imageDataTemp, rRef + rP, cRef + cP, iterationPD, res);
                                }
                            }
                        }

                    }  //next c
                });  //next r
                if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
                else excelApp.StatusBar = string.Format("Correcting initial results ({2})... {0:0%}      DLL in use on {1} logical processors.{3}", (i + 1) / (double)n, numProcessors, numPointsMP, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
            copyArray(ref glitchFlag, glitchFlagTemp);
            if (cntFlagGlitch > 0) flagGlitch = true;
        }

        public static void bDbl(ref double[,] imageData, bool flagSmooth, int flagDllOption, double zr, double zi, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int r1, int c1, int navSize, int rSize, int cSize)
        {  // Burning Ship fractal, standard double precision
            double pixelStep = scaleTerm * res;
            int rUpperIndex = ((rSize - 1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double[,] imageDataTemp = imageData.Clone() as double[,];  //base 1 array, imageData is not allowed in Parallel.For loop because it is passed by reference

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int r = res * ri + 2;          //mapping the parallel loop index to the row number
                    double y0 = zi - pixelStep * ri;
                    double x0 = zr - pixelStep;
                    double x, xtemp, y;
                    int iteration;
                    double iterationD;
                    bool skipCalc;
                    double log_zn, nu;

                    for (int c = 2; c <= cSize + 1; c = c + res) {
                        skipCalc = false;
                        x0 = x0 + pixelStep;
                        if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[r - 1, c - 1] == prevMaxIter)) {

                            // Quick pan code
                            if (refreshType == -1) {
                                skipCalc = quickPan(r1, c1, r, c, navSize);
                            }

                            // Iteration loop
                            if (skipCalc == false) {
                                iteration = 1;
                                x = x0;
                                y = -y0;

                                // Discrete coloring
                                if (flagSmooth == false) {
                                    while (x * x + y * y < 4 && iteration < maxIter) {
                                        xtemp = x * x - y * y + x0;
                                        y = 2 * Math.Abs(x * y) - y0;
                                        x = xtemp;
                                        iteration++;
                                    }
                                    iterationD = iteration;
                                }

                                // Continuous coloring
                                else {
                                    while (x * x + y * y < highBailout && iteration < maxIter) {
                                        xtemp = x * x - y * y + x0;
                                        y = 2 * Math.Abs(x * y) - y0;
                                        x = xtemp;
                                        iteration++;
                                    }

                                    // Additional calculations for continuous coloring
                                    if (iteration < maxIter) {
                                        log_zn = Math.Log(x * x + y * y) / 2;
                                        nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                        iterationD = iteration + 1 - nu;
                                        iterationD = Math.Round(iterationD, 3);
                                    }
                                    else {
                                        iterationD = iteration;
                                    }
                                }

                                // Store result to array
                                if (res == 1) imageDataTemp[r - 1, c - 1] = iterationD;
                                else reduceResolution(ref imageDataTemp, r, c, iterationD, res);

                            }
                        }
                    }  //next c
                });  //next r
                excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
        }

        public static void bPert(ref double[,] imageData, ref int[,] glitchFlag, ref bool flagGlitch, bool flagSmooth, int flagDllOption, double x0, double y0, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int dr, int dc, int navSize, int rr1, int rr2, int cc1, int cc2, int rRef, int cRef, int numPointsMP,
            double[,] xMP, double[,] yMP, double[,] glitchDetect, double divThreshold, bool flagDebugGlitch, int iteration, int maxPastRef, int[,] prevGlitchFlag)
        {  // Burning Ship fractal, perturbation methods
            double pixelStepP = scaleTerm * res;
            int rUpperIndex = ((rr2 - rr1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double zrP = x0 + scaleTerm * cc1;  //min real position
            double ziP = y0 - scaleTerm * rr1;  //max imaginary position
            int bailout = 4;
            if (flagSmooth == true) bailout = highBailout;

            // The following are used to avoid passing variables by reference in the Parallel.For loop, array bases are unchanged when using the Clone method
            double[,] imageDataTemp = imageData.Clone() as double[,];
            int[,] glitchFlagTemp = glitchFlag.Clone() as int[,];
            flagGlitch = false;  //this is set in the Parallel.For loop using the Interlocked.Increment Method on cntFlagGlitch and then checking for non-zero values
            int cntFlagGlitch = 0;

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);
            else excelApp.StatusBar = string.Format("Correcting initial results ({1})... 0%      DLL in use on {0} logical processors.{2}", numProcessors, numPointsMP, dllOptionString);

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int rP = res * ri + rr1;       //mapping the parallel loop index to the row number relative to the reference point
                    double y0P = ziP - pixelStepP * ri;
                    double x0P = zrP - pixelStepP;
                    double x, y, d1;
                    double xP = 0, xtempP, yP = 0;
                    double Dx, Dxtemp, Dy, Dx0, Dy0;
                    int iterationP = 0;
                    double iterationPD = 0;
                    bool skipCalc, goToGlitch;
                    double log_zn, nu;

                    for (int cP = cc1; cP <= cc2; cP = cP + res) {
                        skipCalc = false;
                        x0P = x0P + pixelStepP;
                        if ((cP == 0 && rP == 0) == false) {
                            if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] == prevMaxIter) || (refreshType == 2 && numPointsMP > 1)) {

                                // Quick pan code
                                if (refreshType == -1 && numPointsMP == 1) {
                                    skipCalc = quickPan(dr, dc, rRef + rP, cRef + cP, navSize);
                                }

                                // If correcting a flagged glitch, only correct location that is flagged
                                if (numPointsMP > 1) {
                                    if (prevGlitchFlag[rRef + rP, cRef + cP] == 0) skipCalc = true;
                                }

                                // Iteration loop
                                if (skipCalc == false) {
                                    Dx0 = cP * scaleTerm;
                                    Dy0 = -rP * scaleTerm;
                                    Dx = Dx0;  //delta1 = Y1 - X1
                                    Dy = -Dy0;
                                    goToGlitch = false;

                                    for (int j = 1; j <= iteration; j++) {
                                        iterationP = j;
                                        x = xMP[j, 1];
                                        y = yMP[j, 1];
                                        xP = Dx + x;
                                        yP = Dy + y;
                                        if (xP * xP + yP * yP < glitchDetect[j, 1]) {
                                            glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                            Interlocked.Increment(ref cntFlagGlitch);
                                            if (flagDebugGlitch == true) iterationPD = maxIter;
                                            goToGlitch = true;
                                            break;
                                        }
                                        if (xP * xP + yP * yP > bailout) break;
                                        d1 = x * Dy + y * Dx + Dx * Dy;
                                        //Dy = 2*(Abs(d1 + x*y) - Abs(x*y)) - Dy0
                                        //The size of the term x*y impacts the accuracy of the expression for Dy
                                        //This term is removed using the following relationships:
                                        //  Abs(A+B)-Abs(A) = Abs(B) except...
                                        //  Abs(A+B)-Abs(A) =-Abs(B) if Sgn(A) is opposite Sgn(B) AND Abs(A) > Abs(B)
                                        //  Abs(A+B)-Abs(A) = Abs(B)-2*Abs(A) if Sgn(A) is opposite Sgn(B) AND Abs(A) < Abs(B)
                                        Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0;
                                        if (Math.Sign(d1) + Math.Sign(x * y) != 0) {
                                            Dy = 2 * Math.Abs(d1) - Dy0;
                                        }
                                        else if (Math.Abs(x * y) > Math.Abs(d1)) {
                                            Dy = -2 * Math.Abs(d1) - Dy0;
                                        }
                                        else {
                                            Dy = 2 * (Math.Abs(d1 + x * y) - Math.Abs(x * y)) - Dy0;
                                        }
                                        Dx = Dxtemp;
                                        if (Dx * Dx + Dy * Dy > divThreshold) break;
                                    }  //next j

                                    if (goToGlitch == false) {

                                        // Discrete coloring
                                        if (flagSmooth == false) {
                                            while (xP * xP + yP * yP < 4 && iterationP < maxIter) {
                                                xtempP = xP * xP - yP * yP + x0P;
                                                yP = 2 * Math.Abs(xP * yP) - y0P;
                                                xP = xtempP;
                                                iterationP++;
                                            }
                                            iterationPD = iterationP;
                                        }

                                        // Continuous coloring
                                        else {
                                            while (xP * xP + yP * yP < highBailout && iterationP < maxIter) {
                                                xtempP = xP * xP - yP * yP + x0P;
                                                yP = 2 * Math.Abs(xP * yP) - y0P;
                                                xP = xtempP;
                                                iterationP++;
                                            }

                                            // Additional calculations for continuous coloring
                                            if (iterationP < maxIter) {
                                                log_zn = Math.Log(xP * xP + yP * yP) / 2;
                                                nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                                iterationPD = iterationP + 1 - nu;
                                                iterationPD = Math.Round(iterationPD, 3);
                                            }
                                            else {
                                                iterationPD = iterationP;
                                            }
                                        }

                                    }

                                    // Check if too many additional iterations past the reference were performed
                                    if (maxPastRef != 0 && iterationP > maxPastRef) {
                                        glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                        Interlocked.Increment(ref cntFlagGlitch);
                                        if (flagDebugGlitch == true) iterationPD = maxIter;
                                    }

                                    // Store result to array
                                    if (res == 1) imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] = iterationPD;
                                    else reduceResolution(ref imageDataTemp, rRef + rP, cRef + cP, iterationPD, res);
                                }
                            }
                        }

                    }  //next c
                });  //next r
                if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
                else excelApp.StatusBar = string.Format("Correcting initial results ({2})... {0:0%}      DLL in use on {1} logical processors.{3}", (i + 1) / (double)n, numProcessors, numPointsMP, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
            copyArray(ref glitchFlag, glitchFlagTemp);
            if (cntFlagGlitch > 0) flagGlitch = true;
        }

        public static void fDbl(ref double[,] imageData, bool flagSmooth, int flagDllOption, double zr, double zi, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int r1, int c1, int navSize, int rSize, int cSize)
        {  // Buffalo fractal, standard double precision
            double pixelStep = scaleTerm * res;
            int rUpperIndex = ((rSize - 1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double[,] imageDataTemp = imageData.Clone() as double[,];  //base 1 array, imageData is not allowed in Parallel.For loop because it is passed by reference

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int r = res * ri + 2;          //mapping the parallel loop index to the row number
                    double y0 = zi - pixelStep * ri;
                    double x0 = zr - pixelStep;
                    double x, xtemp, y;
                    int iteration;
                    double iterationD;
                    bool skipCalc;
                    double log_zn, nu;

                    for (int c = 2; c <= cSize + 1; c = c + res) {
                        skipCalc = false;
                        x0 = x0 + pixelStep;
                        if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[r - 1, c - 1] == prevMaxIter)) {

                            // Quick pan code
                            if (refreshType == -1) {
                                skipCalc = quickPan(r1, c1, r, c, navSize);
                            }

                            // Iteration loop
                            if (skipCalc == false) {
                                iteration = 1;
                                x = x0;
                                y = y0;

                                // Discrete coloring
                                if (flagSmooth == false) {
                                    while (x * x + y * y < 4 && iteration < maxIter) {
                                        xtemp = Math.Abs(x * x - y * y) + x0;
                                        y = 2 * Math.Abs(x * y) + y0;
                                        x = xtemp;
                                        iteration++;
                                    }
                                    iterationD = iteration;
                                }

                                // Continuous coloring
                                else {
                                    while (x * x + y * y < highBailout && iteration < maxIter) {
                                        xtemp = Math.Abs(x * x - y * y) + x0;
                                        y = 2 * Math.Abs(x * y) + y0;
                                        x = xtemp;
                                        iteration++;
                                    }

                                    // Additional calculations for continuous coloring
                                    if (iteration < maxIter) {
                                        log_zn = Math.Log(x * x + y * y) / 2;
                                        nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                        iterationD = iteration + 1 - nu;
                                        iterationD = Math.Round(iterationD, 3);
                                    }
                                    else {
                                        iterationD = iteration;
                                    }
                                }

                                // Store result to array
                                if (res == 1) imageDataTemp[r - 1, c - 1] = iterationD;
                                else reduceResolution(ref imageDataTemp, r, c, iterationD, res);

                            }
                        }
                    }  //next c
                });  //next r
                excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
        }

        public static void fPert(ref double[,] imageData, ref int[,] glitchFlag, ref bool flagGlitch, bool flagSmooth, int flagDllOption, double x0, double y0, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int dr, int dc, int navSize, int rr1, int rr2, int cc1, int cc2, int rRef, int cRef, int numPointsMP,
            double[,] xMP, double[,] yMP, double[,] glitchDetect, double divThreshold, bool flagDebugGlitch, int iteration, int maxPastRef, int[,] prevGlitchFlag)
        {  // Buffalo fractal, perturbation methods
            double pixelStepP = scaleTerm * res;
            int rUpperIndex = ((rr2 - rr1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double zrP = x0 + scaleTerm * cc1;  //min real position
            double ziP = y0 - scaleTerm * rr1;  //max imaginary position
            int bailout = 4;
            if (flagSmooth == true) bailout = highBailout;

            // The following are used to avoid passing variables by reference in the Parallel.For loop, array bases are unchanged when using the Clone method
            double[,] imageDataTemp = imageData.Clone() as double[,];
            int[,] glitchFlagTemp = glitchFlag.Clone() as int[,];
            flagGlitch = false;  //this is set in the Parallel.For loop using the Interlocked.Increment Method on cntFlagGlitch and then checking for non-zero values
            int cntFlagGlitch = 0;

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);
            else excelApp.StatusBar = string.Format("Correcting initial results ({1})... 0%      DLL in use on {0} logical processors.{2}", numProcessors, numPointsMP, dllOptionString);

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int rP = res * ri + rr1;       //mapping the parallel loop index to the row number relative to the reference point
                    double y0P = ziP - pixelStepP * ri;
                    double x0P = zrP - pixelStepP;
                    double x, y, d1;
                    double xP = 0, xtempP, yP = 0;
                    double Dx, Dxtemp, Dy, Dx0, Dy0;
                    int iterationP = 0;
                    double iterationPD = 0;
                    bool skipCalc, goToGlitch;
                    double log_zn, nu;

                    for (int cP = cc1; cP <= cc2; cP = cP + res) {
                        skipCalc = false;
                        x0P = x0P + pixelStepP;
                        if ((cP == 0 && rP == 0) == false) {
                            if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] == prevMaxIter) || (refreshType == 2 && numPointsMP > 1)) {

                                // Quick pan code
                                if (refreshType == -1 && numPointsMP == 1) {
                                    skipCalc = quickPan(dr, dc, rRef + rP, cRef + cP, navSize);
                                }

                                // If correcting a flagged glitch, only correct location that is flagged
                                if (numPointsMP > 1) {
                                    if (prevGlitchFlag[rRef + rP, cRef + cP] == 0) skipCalc = true;
                                }

                                // Iteration loop
                                if (skipCalc == false) {
                                    Dx0 = cP * scaleTerm;
                                    Dy0 = -rP * scaleTerm;
                                    Dx = Dx0;
                                    Dy = Dy0;
                                    goToGlitch = false;

                                    for (int j = 1; j <= iteration; j++) {
                                        iterationP = j;
                                        x = xMP[j, 1];
                                        y = yMP[j, 1];
                                        xP = Dx + x;
                                        yP = Dy + y;
                                        if (xP * xP + yP * yP < glitchDetect[j, 1]) {
                                            glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                            Interlocked.Increment(ref cntFlagGlitch);
                                            if (flagDebugGlitch == true) iterationPD = maxIter;
                                            goToGlitch = true;
                                            break;
                                        }
                                        if (xP * xP + yP * yP > bailout) break;

                                        d1 = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy;
                                        //Dx = Abs(d1 + x*x - y*y) - Abs(x*x - y*y) + Dx0
                                        //The size of the term (x*x - y*y) impacts the accuracy of the expression for Dx
                                        //This term is removed using the following relationships:
                                        //  Abs(A+B)-Abs(A) = Abs(B) except...
                                        //  Abs(A+B)-Abs(A) =-Abs(B) if Sgn(A) is opposite Sgn(B) AND Abs(A) > Abs(B)
                                        //  Abs(A+B)-Abs(A) = Abs(B)-2*Abs(A) if Sgn(A) is opposite Sgn(B) AND Abs(A) < Abs(B)
                                        if (Math.Sign(d1) + Math.Sign(x * x - y * y) != 0) {
                                            Dxtemp = Math.Abs(d1) + Dx0;
                                        }
                                        else if (Math.Abs(x * x - y * y) > Math.Abs(d1)) {
                                            Dxtemp = -Math.Abs(d1) + Dx0;
                                        }
                                        else {
                                            Dxtemp = Math.Abs(d1 + x * x - y * y) - Math.Abs(x * x - y * y) + Dx0;
                                        }

                                        d1 = x * Dy + y * Dx + Dx * Dy;
                                        //Dy = 2*(Abs(d1 + x*y) - Abs(x*y)) - Dy0
                                        //The size of the term x*y impacts the accuracy of the expression for Dy
                                        //This term is removed using the following relationships:
                                        //  Abs(A+B)-Abs(A) = Abs(B) except...
                                        //  Abs(A+B)-Abs(A) =-Abs(B) if Sgn(A) is opposite Sgn(B) AND Abs(A) > Abs(B)
                                        //  Abs(A+B)-Abs(A) = Abs(B)-2*Abs(A) if Sgn(A) is opposite Sgn(B) AND Abs(A) < Abs(B)
                                        if (Math.Sign(d1) + Math.Sign(x * y) != 0) {
                                            Dy = 2 * Math.Abs(d1) + Dy0;
                                        }
                                        else if (Math.Abs(x * y) > Math.Abs(d1)) {
                                            Dy = -2 * Math.Abs(d1) + Dy0;
                                        }
                                        else {
                                            Dy = 2 * (Math.Abs(d1 + x * y) - Math.Abs(x * y)) + Dy0;
                                        }

                                        Dx = Dxtemp;
                                        if (Dx * Dx + Dy * Dy > divThreshold) break;
                                    }  //next j

                                    if (goToGlitch == false) {

                                        // Discrete coloring
                                        if (flagSmooth == false) {
                                            while (xP * xP + yP * yP < 4 && iterationP < maxIter) {
                                                xtempP = Math.Abs(xP * xP - yP * yP) + x0P;
                                                yP = 2 * Math.Abs(xP * yP) + y0P;
                                                xP = xtempP;
                                                iterationP++;
                                            }
                                            iterationPD = iterationP;
                                        }

                                        // Continuous coloring
                                        else {
                                            while (xP * xP + yP * yP < highBailout && iterationP < maxIter) {
                                                xtempP = Math.Abs(xP * xP - yP * yP) + x0P;
                                                yP = 2 * Math.Abs(xP * yP) + y0P;
                                                xP = xtempP;
                                                iterationP++;
                                            }

                                            // Additional calculations for continuous coloring
                                            if (iterationP < maxIter) {
                                                log_zn = Math.Log(xP * xP + yP * yP) / 2;
                                                nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                                iterationPD = iterationP + 1 - nu;
                                                iterationPD = Math.Round(iterationPD, 3);
                                            }
                                            else {
                                                iterationPD = iterationP;
                                            }
                                        }

                                    }

                                    // Check if too many additional iterations past the reference were performed
                                    if (maxPastRef != 0 && iterationP > maxPastRef) {
                                        glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                        Interlocked.Increment(ref cntFlagGlitch);
                                        if (flagDebugGlitch == true) iterationPD = maxIter;
                                    }

                                    // Store result to array
                                    if (res == 1) imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] = iterationPD;
                                    else reduceResolution(ref imageDataTemp, rRef + rP, cRef + cP, iterationPD, res);
                                }
                            }
                        }

                    }  //next c
                });  //next r
                if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
                else excelApp.StatusBar = string.Format("Correcting initial results ({2})... {0:0%}      DLL in use on {1} logical processors.{3}", (i + 1) / (double)n, numProcessors, numPointsMP, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
            copyArray(ref glitchFlag, glitchFlagTemp);
            if (cntFlagGlitch > 0) flagGlitch = true;
        }

        public static void cDbl(ref double[,] imageData, bool flagSmooth, int flagDllOption, double zr, double zi, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int r1, int c1, int navSize, int rSize, int cSize)
        {  // Celtic fractal, standard double precision
            double pixelStep = scaleTerm * res;
            int rUpperIndex = ((rSize - 1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double[,] imageDataTemp = imageData.Clone() as double[,];  //base 1 array, imageData is not allowed in Parallel.For loop because it is passed by reference

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int r = res * ri + 2;          //mapping the parallel loop index to the row number
                    double y0 = zi - pixelStep * ri;
                    double x0 = zr - pixelStep;
                    double x, xtemp, y;
                    int iteration;
                    double iterationD;
                    bool skipCalc;
                    double log_zn, nu;

                    for (int c = 2; c <= cSize + 1; c = c + res) {
                        skipCalc = false;
                        x0 = x0 + pixelStep;
                        if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[r - 1, c - 1] == prevMaxIter)) {

                            // Quick pan code
                            if (refreshType == -1) {
                                skipCalc = quickPan(r1, c1, r, c, navSize);
                            }

                            // Iteration loop
                            if (skipCalc == false) {
                                iteration = 1;
                                x = x0;
                                y = y0;

                                // Discrete coloring
                                if (flagSmooth == false) {
                                    while (x * x + y * y < 4 && iteration < maxIter) {
                                        xtemp = Math.Abs(x * x - y * y) + x0;
                                        y = 2 * x * y + y0;
                                        x = xtemp;
                                        iteration++;
                                    }
                                    iterationD = iteration;
                                }

                                // Continuous coloring
                                else {
                                    while (x * x + y * y < highBailout && iteration < maxIter) {
                                        xtemp = Math.Abs(x * x - y * y) + x0;
                                        y = 2 * x * y + y0;
                                        x = xtemp;
                                        iteration++;
                                    }

                                    // Additional calculations for continuous coloring
                                    if (iteration < maxIter) {
                                        log_zn = Math.Log(x * x + y * y) / 2;
                                        nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                        iterationD = iteration + 1 - nu;
                                        iterationD = Math.Round(iterationD, 3);
                                    }
                                    else {
                                        iterationD = iteration;
                                    }
                                }

                                // Store result to array
                                if (res == 1) imageDataTemp[r - 1, c - 1] = iterationD;
                                else reduceResolution(ref imageDataTemp, r, c, iterationD, res);

                            }
                        }
                    }  //next c
                });  //next r
                excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
        }

        public static void cPert(ref double[,] imageData, ref int[,] glitchFlag, ref bool flagGlitch, bool flagSmooth, int flagDllOption, double x0, double y0, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int dr, int dc, int navSize, int rr1, int rr2, int cc1, int cc2, int rRef, int cRef, int numPointsMP,
            double[,] xMP, double[,] yMP, double[,] glitchDetect, double divThreshold, bool flagDebugGlitch, int iteration, int maxPastRef, int[,] prevGlitchFlag)
        {  // Celtic fractal, perturbation methods
            double pixelStepP = scaleTerm * res;
            int rUpperIndex = ((rr2 - rr1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double zrP = x0 + scaleTerm * cc1;  //min real position
            double ziP = y0 - scaleTerm * rr1;  //max imaginary position
            int bailout = 4;
            if (flagSmooth == true) bailout = highBailout;

            // The following are used to avoid passing variables by reference in the Parallel.For loop, array bases are unchanged when using the Clone method
            double[,] imageDataTemp = imageData.Clone() as double[,];
            int[,] glitchFlagTemp = glitchFlag.Clone() as int[,];
            flagGlitch = false;  //this is set in the Parallel.For loop using the Interlocked.Increment Method on cntFlagGlitch and then checking for non-zero values
            int cntFlagGlitch = 0;

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);
            else excelApp.StatusBar = string.Format("Correcting initial results ({1})... 0%      DLL in use on {0} logical processors.{2}", numProcessors, numPointsMP, dllOptionString);

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int rP = res * ri + rr1;       //mapping the parallel loop index to the row number relative to the reference point
                    double y0P = ziP - pixelStepP * ri;
                    double x0P = zrP - pixelStepP;
                    double x, y, d1;
                    double xP = 0, xtempP, yP = 0;
                    double Dx, Dxtemp, Dy, Dx0, Dy0;
                    int iterationP = 0;
                    double iterationPD = 0;
                    bool skipCalc, goToGlitch;
                    double log_zn, nu;

                    for (int cP = cc1; cP <= cc2; cP = cP + res) {
                        skipCalc = false;
                        x0P = x0P + pixelStepP;
                        if ((cP == 0 && rP == 0) == false) {
                            if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] == prevMaxIter) || (refreshType == 2 && numPointsMP > 1)) {

                                // Quick pan code
                                if (refreshType == -1 && numPointsMP == 1) {
                                    skipCalc = quickPan(dr, dc, rRef + rP, cRef + cP, navSize);
                                }

                                // If correcting a flagged glitch, only correct location that is flagged
                                if (numPointsMP > 1) {
                                    if (prevGlitchFlag[rRef + rP, cRef + cP] == 0) skipCalc = true;
                                }

                                // Iteration loop
                                if (skipCalc == false) {
                                    Dx0 = cP * scaleTerm;
                                    Dy0 = -rP * scaleTerm;
                                    Dx = Dx0;
                                    Dy = Dy0;
                                    goToGlitch = false;

                                    for (int j = 1; j <= iteration; j++) {
                                        iterationP = j;
                                        x = xMP[j, 1];
                                        y = yMP[j, 1];
                                        xP = Dx + x;
                                        yP = Dy + y;
                                        if (xP * xP + yP * yP < glitchDetect[j, 1]) {
                                            glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                            Interlocked.Increment(ref cntFlagGlitch);
                                            if (flagDebugGlitch == true) iterationPD = maxIter;
                                            goToGlitch = true;
                                            break;
                                        }
                                        if (xP * xP + yP * yP > bailout) break;
                                        d1 = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy;
                                        //Dx = Abs(d1 + x*x - y*y) - Abs(x*x - y*y) + Dx0
                                        //The size of the term (x*x - y*y) impacts the accuracy of the expression for Dx
                                        //This term is removed using the following relationships:
                                        //  Abs(A+B)-Abs(A) = Abs(B) except...
                                        //  Abs(A+B)-Abs(A) =-Abs(B) if Sgn(A) is opposite Sgn(B) AND Abs(A) > Abs(B)
                                        //  Abs(A+B)-Abs(A) = Abs(B)-2*Abs(A) if Sgn(A) is opposite Sgn(B) AND Abs(A) < Abs(B)
                                        if (Math.Sign(d1) + Math.Sign(x * x - y * y) != 0) {
                                            Dxtemp = Math.Abs(d1) + Dx0;
                                        }
                                        else if (Math.Abs(x * x - y * y) > Math.Abs(d1)) {
                                            Dxtemp = -Math.Abs(d1) + Dx0;
                                        }
                                        else {
                                            Dxtemp = Math.Abs(d1 + x * x - y * y) - Math.Abs(x * x - y * y) + Dx0;
                                        }
                                        Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dy0;
                                        Dx = Dxtemp;
                                        if (Dx * Dx + Dy * Dy > divThreshold) break;
                                    }  //next j

                                    if (goToGlitch == false) {

                                        // Discrete coloring
                                        if (flagSmooth == false) {
                                            while (xP * xP + yP * yP < 4 && iterationP < maxIter) {
                                                xtempP = Math.Abs(xP * xP - yP * yP) + x0P;
                                                yP = 2 * xP * yP + y0P;
                                                xP = xtempP;
                                                iterationP++;
                                            }
                                            iterationPD = iterationP;
                                        }

                                        // Continuous coloring
                                        else {
                                            while (xP * xP + yP * yP < highBailout && iterationP < maxIter) {
                                                xtempP = Math.Abs(xP * xP - yP * yP) + x0P;
                                                yP = 2 * xP * yP + y0P;
                                                xP = xtempP;
                                                iterationP++;
                                            }

                                            // Additional calculations for continuous coloring
                                            if (iterationP < maxIter) {
                                                log_zn = Math.Log(xP * xP + yP * yP) / 2;
                                                nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                                iterationPD = iterationP + 1 - nu;
                                                iterationPD = Math.Round(iterationPD, 3);
                                            }
                                            else {
                                                iterationPD = iterationP;
                                            }
                                        }

                                    }

                                    // Check if too many additional iterations past the reference were performed
                                    if (maxPastRef != 0 && iterationP > maxPastRef) {
                                        glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                        Interlocked.Increment(ref cntFlagGlitch);
                                        if (flagDebugGlitch == true) iterationPD = maxIter;
                                    }

                                    // Store result to array
                                    if (res == 1) imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] = iterationPD;
                                    else reduceResolution(ref imageDataTemp, rRef + rP, cRef + cP, iterationPD, res);
                                }
                            }
                        }

                    }  //next c
                });  //next r
                if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
                else excelApp.StatusBar = string.Format("Correcting initial results ({2})... {0:0%}      DLL in use on {1} logical processors.{3}", (i + 1) / (double)n, numProcessors, numPointsMP, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
            copyArray(ref glitchFlag, glitchFlagTemp);
            if (cntFlagGlitch > 0) flagGlitch = true;
        }

        public static void rDbl(ref double[,] imageData, double eqnParameter, bool flagSmooth, int flagDllOption, double zr, double zi, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int r1, int c1, int navSize, int rSize, int cSize)
        {  // Rotation Family, standard double precision
            double pixelStep = scaleTerm * res;
            int rUpperIndex = ((rSize - 1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double[,] imageDataTemp = imageData.Clone() as double[,];  //base 1 array, imageData is not allowed in Parallel.For loop because it is passed by reference
            double theta = 2 * Math.PI / eqnParameter;

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int r = res * ri + 2;          //mapping the parallel loop index to the row number
                    double y0 = zi - pixelStep * ri;
                    double x0 = zr - pixelStep;
                    double x, xtemp, y;
                    int iteration;
                    double iterationD;
                    bool skipCalc;
                    double log_zn, nu;
                    double cs, sn;

                    for (int c = 2; c <= cSize + 1; c = c + res) {
                        skipCalc = false;
                        x0 = x0 + pixelStep;
                        if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[r - 1, c - 1] == prevMaxIter)) {

                            // Quick pan code
                            if (refreshType == -1) {
                                skipCalc = quickPan(r1, c1, r, c, navSize);
                            }

                            // Iteration loop
                            if (skipCalc == false) {
                                iteration = 1;
                                cs = Math.Cos(theta);
                                sn = Math.Sin(theta);
                                x = x0 * cs - y0 * sn;
                                y = x0 * sn + y0 * cs;

                                // Discrete coloring
                                if (flagSmooth == false) {
                                    while (x * x + y * y < 4 && iteration < maxIter) {
                                        iteration++;
                                        cs = Math.Cos(theta * iteration);
                                        sn = Math.Sin(theta * iteration);
                                        xtemp = x * x - y * y + x0 * cs - y0 * sn;
                                        y = 2 * x * y + x0 * sn + y0 * cs;
                                        x = xtemp;
                                    }
                                    iterationD = iteration;
                                }

                                // Continuous coloring
                                else {
                                    while (x * x + y * y < highBailout && iteration < maxIter) {
                                        iteration++;
                                        cs = Math.Cos(theta * iteration);
                                        sn = Math.Sin(theta * iteration);
                                        xtemp = x * x - y * y + x0 * cs - y0 * sn;
                                        y = 2 * x * y + x0 * sn + y0 * cs;
                                        x = xtemp;
                                    }

                                    // Additional calculations for continuous coloring
                                    if (iteration < maxIter) {
                                        log_zn = Math.Log(x * x + y * y) / 2;
                                        nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                        iterationD = iteration + 1 - nu;
                                        iterationD = Math.Round(iterationD, 3);
                                    }
                                    else {
                                        iterationD = iteration;
                                    }
                                }

                                // Store result to array
                                if (res == 1) imageDataTemp[r - 1, c - 1] = iterationD;
                                else reduceResolution(ref imageDataTemp, r, c, iterationD, res);

                            }
                        }
                    }  //next c
                });  //next r
                excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
        }

        public static void rPert(ref double[,] imageData, ref int[,] glitchFlag, ref bool flagGlitch, double eqnParameter, bool flagSmooth, int flagDllOption, double x0, double y0, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int dr, int dc, int navSize, int rr1, int rr2, int cc1, int cc2, int rRef, int cRef, int numPointsMP,
            double[,] xMP, double[,] yMP, double[,] glitchDetect, double divThreshold, bool flagDebugGlitch, int iteration, int maxPastRef, int[,] prevGlitchFlag)
        {  // Rotation Family, perturbation methods
            double pixelStepP = scaleTerm * res;
            int rUpperIndex = ((rr2 - rr1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double zrP = x0 + scaleTerm * cc1;  //min real position
            double ziP = y0 - scaleTerm * rr1;  //max imaginary position
            int bailout = 4;
            if (flagSmooth == true) bailout = highBailout;
            double theta = 2 * Math.PI / eqnParameter;

            // The following are used to avoid passing variables by reference in the Parallel.For loop, array bases are unchanged when using the Clone method
            double[,] imageDataTemp = imageData.Clone() as double[,];
            int[,] glitchFlagTemp = glitchFlag.Clone() as int[,];
            flagGlitch = false;  //this is set in the Parallel.For loop using the Interlocked.Increment Method on cntFlagGlitch and then checking for non-zero values
            int cntFlagGlitch = 0;

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);
            else excelApp.StatusBar = string.Format("Correcting initial results ({1})... 0%      DLL in use on {0} logical processors.{2}", numProcessors, numPointsMP, dllOptionString);

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int rP = res * ri + rr1;       //mapping the parallel loop index to the row number relative to the reference point
                    double y0P = ziP - pixelStepP * ri;
                    double x0P = zrP - pixelStepP;
                    double x, y;
                    double xP = 0, xtempP, yP = 0;
                    double Dx, Dxtemp, Dy, Dx0, Dy0;
                    int iterationP = 0;
                    double iterationPD = 0;
                    bool skipCalc, goToGlitch;
                    double log_zn, nu;
                    double cs, sn;

                    for (int cP = cc1; cP <= cc2; cP = cP + res) {
                        skipCalc = false;
                        x0P = x0P + pixelStepP;
                        if ((cP == 0 && rP == 0) == false) {
                            if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] == prevMaxIter) || (refreshType == 2 && numPointsMP > 1)) {

                                // Quick pan code
                                if (refreshType == -1 && numPointsMP == 1) {
                                    skipCalc = quickPan(dr, dc, rRef + rP, cRef + cP, navSize);
                                }

                                // If correcting a flagged glitch, only correct location that is flagged
                                if (numPointsMP > 1) {
                                    if (prevGlitchFlag[rRef + rP, cRef + cP] == 0) skipCalc = true;
                                }

                                // Iteration loop
                                if (skipCalc == false) {
                                    Dx0 = cP * scaleTerm;
                                    Dy0 = -rP * scaleTerm;
                                    cs = Math.Cos(theta);
                                    sn = Math.Sin(theta);
                                    Dx = Dx0 * cs - Dy0 * sn;  //delta1 = Y1 - X1
                                    Dy = Dx0 * sn + Dy0 * cs;
                                    goToGlitch = false;

                                    for (int j = 1; j <= iteration; j++) {
                                        iterationP = j;
                                        x = xMP[j, 1];
                                        y = yMP[j, 1];
                                        xP = Dx + x;
                                        yP = Dy + y;
                                        if (xP * xP + yP * yP < glitchDetect[j, 1]) {
                                            glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                            Interlocked.Increment(ref cntFlagGlitch);
                                            if (flagDebugGlitch == true) iterationPD = maxIter;
                                            goToGlitch = true;
                                            break;
                                        }
                                        if (xP * xP + yP * yP > bailout) break;
                                        cs = Math.Cos(theta * (j + 1));
                                        sn = Math.Sin(theta * (j + 1));
                                        Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0 * cs - Dy0 * sn;
                                        Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dx0 * sn + Dy0 * cs;
                                        Dx = Dxtemp;
                                        if (Dx * Dx + Dy * Dy > divThreshold) break;
                                    }  //next j

                                    if (goToGlitch == false) {

                                        // Discrete coloring
                                        if (flagSmooth == false) {
                                            while (xP * xP + yP * yP < 4 && iterationP < maxIter) {
                                                iterationP++;
                                                cs = Math.Cos(theta * iterationP);
                                                sn = Math.Sin(theta * iterationP);
                                                xtempP = xP * xP - yP * yP + x0P * cs - y0P * sn;
                                                yP = 2 * xP * yP + x0P * sn + y0P * cs;
                                                xP = xtempP;
                                            }
                                            iterationPD = iterationP;
                                        }

                                        // Continuous coloring
                                        else {
                                            while (xP * xP + yP * yP < highBailout && iterationP < maxIter) {
                                                iterationP++;
                                                cs = Math.Cos(theta * iterationP);
                                                sn = Math.Sin(theta * iterationP);
                                                xtempP = xP * xP - yP * yP + x0P * cs - y0P * sn;
                                                yP = 2 * xP * yP + x0P * sn + y0P * cs;
                                                xP = xtempP;
                                            }

                                            // Additional calculations for continuous coloring
                                            if (iterationP < maxIter) {
                                                log_zn = Math.Log(xP * xP + yP * yP) / 2;
                                                nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                                iterationPD = iterationP + 1 - nu;
                                                iterationPD = Math.Round(iterationPD, 3);
                                            }
                                            else {
                                                iterationPD = iterationP;
                                            }
                                        }

                                    }

                                    // Check if too many additional iterations past the reference were performed
                                    if (maxPastRef != 0 && iterationP > maxPastRef) {
                                        glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                        Interlocked.Increment(ref cntFlagGlitch);
                                        if (flagDebugGlitch == true) iterationPD = maxIter;
                                    }

                                    // Store result to array
                                    if (res == 1) imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] = iterationPD;
                                    else reduceResolution(ref imageDataTemp, rRef + rP, cRef + cP, iterationPD, res);
                                }
                            }
                        }

                    }  //next c
                });  //next r
                if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
                else excelApp.StatusBar = string.Format("Correcting initial results ({2})... {0:0%}      DLL in use on {1} logical processors.{3}", (i + 1) / (double)n, numProcessors, numPointsMP, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
            copyArray(ref glitchFlag, glitchFlagTemp);
            if (cntFlagGlitch > 0) flagGlitch = true;
        }

        public static void rSeries3(ref double[,] imageData, ref int[,] glitchFlag, ref bool flagGlitch, ref int numSkip, double eqnParameter, bool flagSmooth, int flagDllOption, double x0, double y0, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int dr, int dc, int navSize, int rr1, int rr2, int cc1, int cc2, int rRef, int cRef, int numPointsMP, double seriesTolerance,
            double[,] xMP, double[,] yMP, double[,] glitchDetect, double divThreshold, bool flagDebugGlitch, int iteration, int maxPastRef, int[,] prevGlitchFlag, int prevGlitchInfo2)
        {  // Rotation Family, perturbation with forced 3rd order series approximation
            double pixelStepP = scaleTerm * res;
            int rUpperIndex = ((rr2 - rr1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double zrP = x0 + scaleTerm * cc1;  //min real position
            double ziP = y0 - scaleTerm * rr1;  //max imaginary position
            bool skipSeriesFlag = false;
            bool flagOverflow = false;
            int maxRefSeries = 0;
            int skipNumber = 0;
            double[] A1r = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
            double[] A2r = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
            double[] A1i = new double[iteration];
            double[] A2i = new double[iteration];
            double[] B1r = new double[iteration];
            double[] B2r = new double[iteration];
            double[] B3r = new double[iteration];
            double[] B1i = new double[iteration];
            double[] B2i = new double[iteration];
            double[] B3i = new double[iteration];
            double[] C1r = new double[iteration];
            double[] C2r = new double[iteration];
            double[] C3r = new double[iteration];
            double[] C4r = new double[iteration];
            double[] C1i = new double[iteration];
            double[] C2i = new double[iteration];
            double[] C3i = new double[iteration];
            double[] C4i = new double[iteration];
            int bailout = 4;
            if (flagSmooth == true) bailout = highBailout;
            double theta = 2 * Math.PI / eqnParameter;
            double cs2, sn2, cs1, sn1;

            // The following are used to avoid passing variables by reference in the Parallel.For loop, array bases are unchanged when using the Clone method
            double[,] imageDataTemp = imageData.Clone() as double[,];
            int[,] glitchFlagTemp = glitchFlag.Clone() as int[,];
            flagGlitch = false;  //this is set in the Parallel.For loop using the Interlocked.Increment Method on cntFlagGlitch and then checking for non-zero values
            int cntFlagGlitch = 0;

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);
            else excelApp.StatusBar = string.Format("Correcting initial results ({1})... 0%      DLL in use on {0} logical processors.{2}", numProcessors, numPointsMP, dllOptionString);

            // Calculate coefficients of series approximation up to a 3rd order approximation
            // When correcting glitches, series approximation is skipped if the largest glitch is small
            if (numPointsMP == 1) {
                if (iteration < 10) skipSeriesFlag = true;
            }
            else {
                if (iteration < 10 || prevGlitchInfo2 < 10) skipSeriesFlag = true;
            }
            if (skipSeriesFlag == false) {
                cs2 = Math.Cos(theta * 2);
                sn2 = Math.Sin(theta * 2);
                cs1 = Math.Cos(theta);
                sn1 = Math.Sin(theta);
                A1r[1] = 2 * (cs1 * xMP[1, 1] - sn1 * yMP[1, 1]) + cs2;
                A2r[1] = -2 * (sn1 * xMP[1, 1] + cs1 * yMP[1, 1]) - sn2;
                B1r[1] = cs1 * cs1 - sn1 * sn1;
                B2r[1] = -B1r[1];
                B3r[1] = -4 * cs1 * sn1;
                A1i[1] = -A2r[1];
                A2i[1] = A1r[1];
                B1i[1] = 2 * cs1 * sn1;
                B2i[1] = -B1i[1];
                B3i[1] = 2 * B1r[1];
                for (int i = 2; i <= iteration - 1; i++) {
                    maxRefSeries = i;
                    cs2 = Math.Cos(theta * (i + 1));
                    sn2 = Math.Sin(theta * (i + 1));
                    A1r[i] = 2 * (-A1i[i - 1] * yMP[i, 1] + A1r[i - 1] * xMP[i, 1]) + cs2;
                    A2r[i] = 2 * (-A2i[i - 1] * yMP[i, 1] + A2r[i - 1] * xMP[i, 1]) - sn2;
                    B1r[i] = -A1i[i - 1] * A1i[i - 1] + A1r[i - 1] * A1r[i - 1] + 2 * (-B1i[i - 1] * yMP[i, 1] + B1r[i - 1] * xMP[i, 1]);
                    B2r[i] = -A2i[i - 1] * A2i[i - 1] + A2r[i - 1] * A2r[i - 1] + 2 * (-B2i[i - 1] * yMP[i, 1] + B2r[i - 1] * xMP[i, 1]);
                    B3r[i] = 2 * (-A1i[i - 1] * A2i[i - 1] + A1r[i - 1] * A2r[i - 1] - B3i[i - 1] * yMP[i, 1] + B3r[i - 1] * xMP[i, 1]);
                    C1r[i] = 2 * (-A1i[i - 1] * B1i[i - 1] + A1r[i - 1] * B1r[i - 1] - C1i[i - 1] * yMP[i, 1] + C1r[i - 1] * xMP[i, 1]);
                    C2r[i] = 2 * (-A2i[i - 1] * B2i[i - 1] + A2r[i - 1] * B2r[i - 1] - C2i[i - 1] * yMP[i, 1] + C2r[i - 1] * xMP[i, 1]);
                    C3r[i] = 2 * (-A1i[i - 1] * B3i[i - 1] + A1r[i - 1] * B3r[i - 1] - A2i[i - 1] * B1i[i - 1] + A2r[i - 1] * B1r[i - 1] - C3i[i - 1] * yMP[i, 1] + C3r[i - 1] * xMP[i, 1]);
                    C4r[i] = 2 * (-A1i[i - 1] * B2i[i - 1] + A1r[i - 1] * B2r[i - 1] - A2i[i - 1] * B3i[i - 1] + A2r[i - 1] * B3r[i - 1] - C4i[i - 1] * yMP[i, 1] + C4r[i - 1] * xMP[i, 1]);
                    A1i[i] = 2 * (A1i[i - 1] * xMP[i, 1] + A1r[i - 1] * yMP[i, 1]) + sn2;
                    A2i[i] = 2 * (A2i[i - 1] * xMP[i, 1] + A2r[i - 1] * yMP[i, 1]) + cs2;
                    B1i[i] = 2 * (A1i[i - 1] * A1r[i - 1] + B1i[i - 1] * xMP[i, 1] + B1r[i - 1] * yMP[i, 1]);
                    B2i[i] = 2 * (A2i[i - 1] * A2r[i - 1] + B2i[i - 1] * xMP[i, 1] + B2r[i - 1] * yMP[i, 1]);
                    B3i[i] = 2 * (A1i[i - 1] * A2r[i - 1] + A1r[i - 1] * A2i[i - 1] + B3i[i - 1] * xMP[i, 1] + B3r[i - 1] * yMP[i, 1]);
                    C1i[i] = 2 * (A1i[i - 1] * B1r[i - 1] + A1r[i - 1] * B1i[i - 1] + C1i[i - 1] * xMP[i, 1] + C1r[i - 1] * yMP[i, 1]);
                    C2i[i] = 2 * (A2i[i - 1] * B2r[i - 1] + A2r[i - 1] * B2i[i - 1] + C2i[i - 1] * xMP[i, 1] + C2r[i - 1] * yMP[i, 1]);
                    C3i[i] = 2 * (A1i[i - 1] * B3r[i - 1] + A1r[i - 1] * B3i[i - 1] + A2i[i - 1] * B1r[i - 1] + A2r[i - 1] * B1i[i - 1] + C3i[i - 1] * xMP[i, 1] + C3r[i - 1] * yMP[i, 1]);
                    C4i[i] = 2 * (A1i[i - 1] * B2r[i - 1] + A1r[i - 1] * B2i[i - 1] + A2i[i - 1] * B3r[i - 1] + A2r[i - 1] * B3i[i - 1] + C4i[i - 1] * xMP[i, 1] + C4r[i - 1] * yMP[i, 1]);
                    // Check for overflow, unlike VBA an exception is not thrown when a double exceeds the maximum magnitude
                    if (Double.IsInfinity(C4i[i]) || Double.IsInfinity(C4r[i]) || Double.IsInfinity(C3i[i]) || Double.IsInfinity(C3r[i]) || Double.IsInfinity(C2i[i]) || Double.IsInfinity(C2r[i]) ||
                        Double.IsInfinity(C1i[i]) || Double.IsInfinity(C1r[i]) || Double.IsInfinity(B3i[i]) || Double.IsInfinity(B3r[i]) || Double.IsInfinity(B2i[i]) || Double.IsInfinity(B2r[i]) ||
                        Double.IsInfinity(B1i[i]) || Double.IsInfinity(B1r[i]) || Double.IsInfinity(A2i[i]) || Double.IsInfinity(A2r[i]) || Double.IsInfinity(A1i[i]) || Double.IsInfinity(A1r[i])) {
                        flagOverflow = true;
                        break;
                    }
                }  //next i

                // Catch overflow condition with flag
                if (flagOverflow == true) {
                    if (maxRefSeries < 10) skipSeriesFlag = true;
                    maxRefSeries = maxRefSeries - 1;
                }
            }

            // Determine how many iterations to skip
            if (skipSeriesFlag == false) {
                for (int j = 1; j <= 4; j++) {  //4 corner pixels of range
                    int rP = 0, cP = 0;
                    if (j == 1) {  //top left, min Re & max Im
                        rP = rr1;
                        cP = cc1;
                    }
                    else if (j == 2) {  //top right, max Re & max Im
                        rP = rr1;
                        cP = cc2;
                    }
                    else if (j == 3) {  //bottom left, min Re & min Im
                        rP = rr2;
                        cP = cc1;
                    }
                    else if (j == 4) {  //bottom right, max Re & min Im
                        rP = rr2;
                        cP = cc2;
                    }
                    // Note rr1 and cc1 are negative while rr2 and cc2 are positive
                    double x0P = x0 + scaleTerm * cP;  //Re position
                    double y0P = y0 - scaleTerm * rP;  //Im position

                    // Iteration loop using perturbation method, save delta values
                    double[] DxP = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
                    double[] DyP = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
                    double x, y, Dxtemp;
                    int i2 = 1;
                    double xP = x0P;
                    double yP = y0P;
                    double Dx0 = cP * scaleTerm;
                    double Dy0 = -rP * scaleTerm;
                    double cs = Math.Cos(theta);
                    double sn = Math.Sin(theta);
                    double Dx = Dx0 * cs - Dy0 * sn;  //delta1 = Y1 - X1
                    double Dy = Dx0 * sn + Dy0 * cs;
                    bool flagBreak = false;
                    DxP[0] = Dx;
                    DyP[0] = Dy;
                    for (int i = 1; i <= maxRefSeries - 1; i++) {
                        i2 = i;
                        x = xMP[i, 1];
                        y = yMP[i, 1];
                        xP = Dx + x;
                        yP = Dy + y;
                        if (xP * xP + yP * yP < glitchDetect[i, 1]) {
                            flagBreak = true;
                            break;
                        }
                        if (xP * xP + yP * yP > bailout) {
                            flagBreak = true;
                            break;
                        }
                        cs = Math.Cos(theta * (i + 1));
                        sn = Math.Sin(theta * (i + 1));
                        Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0 * cs - Dy0 * sn;
                        Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dx0 * sn + Dy0 * cs;
                        Dx = Dxtemp;
                        if (Dx * Dx + Dy * Dy > divThreshold) {
                            flagBreak = true;
                            break;
                        }
                        DxP[i] = Dx;
                        DyP[i] = Dy;
                    }  //next i
                    if (flagBreak == false) i2++;

                    // Compare delta values calculated using series approximation (DxS, DyS) to delta values calculated using
                    // perturbation theory (DxP, DyP) to determine how long the series approximation is valid
                    if (i2 < (10 + 1)) {
                        skipSeriesFlag = true;
                        break;
                    }
                    int k2 = 1;
                    double DxS, DyS;
                    flagBreak = false;
                    for (int k = 1; k <= i2 - 2; k++) {
                        k2 = k + 1;
                        DxS = A1r[k] * Dx0 + A2r[k] * Dy0 + B1r[k] * Dx0 * Dx0 + B2r[k] * Dy0 * Dy0 + B3r[k] * Dx0 * Dy0 + C1r[k] * Dx0 * Dx0 * Dx0 + C2r[k] * Dy0 * Dy0 * Dy0 + C3r[k] * Dx0 * Dx0 * Dy0 + C4r[k] * Dx0 * Dy0 * Dy0;
                        DyS = A1i[k] * Dx0 + A2i[k] * Dy0 + B1i[k] * Dx0 * Dx0 + B2i[k] * Dy0 * Dy0 + B3i[k] * Dx0 * Dy0 + C1i[k] * Dx0 * Dx0 * Dx0 + C2i[k] * Dy0 * Dy0 * Dy0 + C3i[k] * Dx0 * Dx0 * Dy0 + C4i[k] * Dx0 * Dy0 * Dy0;
                        if (DxP[k] == 0 || DyP[k] == 0) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DxP[k] - DxS) / DxP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DyP[k] - DyS) / DyP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                    }
                    if (flagBreak == true) k2 = k2 - 1;  //go back to when relative difference was below tolerance, this line is surprisingly important
                    if (k2 < 5) {
                        skipSeriesFlag = true;
                        skipNumber = 0;
                        break;
                    }
                    if (j == 1) {
                        skipNumber = k2;
                    }
                    else {
                        if (k2 < skipNumber) skipNumber = k2;
                    }
                }  //next j
            }
            if (numPointsMP == 1) numSkip = skipNumber;  //store number of iterations skipped in initial pass for calculation info output

            // Calculate points about reference point orbit in double precision

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int rP = res * ri + rr1;       //mapping the parallel loop index to the row number relative to the reference point
                    double y0P = ziP - pixelStepP * ri;
                    double x0P = zrP - pixelStepP;
                    double x, y;
                    double xP = 0, xtempP, yP = 0;
                    double Dx = 0, Dxtemp, Dy = 0, Dx0, Dy0;
                    int iterationP = 0;
                    bool skipCalc, goToGlitch;
                    double iterationPD = 0;
                    double log_zn, nu;
                    double cs, sn;

                    for (int cP = cc1; cP <= cc2; cP = cP + res) {
                        skipCalc = false;
                        x0P = x0P + pixelStepP;
                        if ((cP == 0 && rP == 0) == false) {
                            if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] == prevMaxIter) || (refreshType == 2 && numPointsMP > 1)) {

                                // Quick pan code
                                if (refreshType == -1 && numPointsMP == 1) {
                                    skipCalc = quickPan(dr, dc, rRef + rP, cRef + cP, navSize);
                                }

                                // If correcting a flagged glitch, only correct location that is flagged
                                if (numPointsMP > 1) {
                                    if (prevGlitchFlag[rRef + rP, cRef + cP] == 0) skipCalc = true;
                                }

                                // Series approximation to skip iterations
                                if (skipCalc == false) {
                                    int m = skipNumber;
                                    int k = 1;
                                    bool tryLessFlag = false;
                                    bool goToTryLess = true;
                                    Dx0 = cP * scaleTerm;
                                    Dy0 = -rP * scaleTerm;
                                    if (skipSeriesFlag == false) {
                                        while (goToTryLess == true) {
                                            goToTryLess = false;
                                            Dx = A1r[m - 1] * Dx0 + A2r[m - 1] * Dy0 + B1r[m - 1] * Dx0 * Dx0 + B2r[m - 1] * Dy0 * Dy0 + B3r[m - 1] * Dx0 * Dy0 + C1r[m - 1] * Dx0 * Dx0 * Dx0 + C2r[m - 1] * Dy0 * Dy0 * Dy0 + C3r[m - 1] * Dx0 * Dx0 * Dy0 + C4r[m - 1] * Dx0 * Dy0 * Dy0;
                                            Dy = A1i[m - 1] * Dx0 + A2i[m - 1] * Dy0 + B1i[m - 1] * Dx0 * Dx0 + B2i[m - 1] * Dy0 * Dy0 + B3i[m - 1] * Dx0 * Dy0 + C1i[m - 1] * Dx0 * Dx0 * Dx0 + C2i[m - 1] * Dy0 * Dy0 * Dy0 + C3i[m - 1] * Dx0 * Dx0 * Dy0 + C4i[m - 1] * Dx0 * Dy0 * Dy0;
                                            if (Dx * Dx + Dy * Dy > divThreshold) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            xP = Dx + xMP[m, 1];
                                            yP = Dy + yMP[m, 1];
                                            if (xP * xP + yP * yP > bailout) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            if (goToTryLess == false) {
                                                k = m;  //skip first m-1 iterations
                                            }
                                            else {
                                                m = m / 2;
                                            }
                                        }  //Wend
                                    }

                                    // Iteration loop
                                    if (k == 1) {  //no series approximation used
                                        cs = Math.Cos(theta);
                                        sn = Math.Sin(theta);
                                        Dx = Dx0 * cs - Dy0 * sn;  //delta1 = Y1 - X1
                                        Dy = Dx0 * sn + Dy0 * cs;
                                    }

                                    goToGlitch = false;
                                    for (int j = k; j <= iteration; j++) {
                                        iterationP = j;
                                        x = xMP[j, 1];
                                        y = yMP[j, 1];
                                        xP = Dx + x;
                                        yP = Dy + y;
                                        if (xP * xP + yP * yP < glitchDetect[j, 1]) {
                                            glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                            Interlocked.Increment(ref cntFlagGlitch);
                                            if (flagDebugGlitch == true) iterationPD = maxIter;
                                            goToGlitch = true;
                                            break;
                                        }
                                        if (xP * xP + yP * yP > bailout) break;
                                        cs = Math.Cos(theta * (j + 1));
                                        sn = Math.Sin(theta * (j + 1));
                                        Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0 * cs - Dy0 * sn;
                                        Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dx0 * sn + Dy0 * cs;
                                        Dx = Dxtemp;
                                        if (Dx * Dx + Dy * Dy > divThreshold) break;
                                    }  //next j

                                    if (goToGlitch == false) {

                                        // Discrete coloring
                                        if (flagSmooth == false) {
                                            while (xP * xP + yP * yP < 4 && iterationP < maxIter) {
                                                iterationP++;
                                                cs = Math.Cos(theta * iterationP);
                                                sn = Math.Sin(theta * iterationP);
                                                xtempP = xP * xP - yP * yP + x0P * cs - y0P * sn;
                                                yP = 2 * xP * yP + x0P * sn + y0P * cs;
                                                xP = xtempP;
                                            }
                                            iterationPD = iterationP;
                                        }

                                        // Continuous coloring
                                        else {
                                            while (xP * xP + yP * yP < highBailout && iterationP < maxIter) {
                                                iterationP++;
                                                cs = Math.Cos(theta * iterationP);
                                                sn = Math.Sin(theta * iterationP);
                                                xtempP = xP * xP - yP * yP + x0P * cs - y0P * sn;
                                                yP = 2 * xP * yP + x0P * sn + y0P * cs;
                                                xP = xtempP;
                                            }

                                            // Additional calculations for continuous coloring
                                            if (iterationP < maxIter) {
                                                log_zn = Math.Log(xP * xP + yP * yP) / 2;
                                                nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                                iterationPD = iterationP + 1 - nu;
                                                iterationPD = Math.Round(iterationPD, 3);
                                            }
                                            else {
                                                iterationPD = iterationP;
                                            }
                                        }

                                    }

                                    // Check if too many additional iterations past the reference were performed
                                    if (maxPastRef != 0 && iterationP > maxPastRef) {
                                        glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                        Interlocked.Increment(ref cntFlagGlitch);
                                        if (flagDebugGlitch == true) iterationPD = maxIter;
                                    }

                                    // Store result to array
                                    if (res == 1) imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] = iterationPD;
                                    else reduceResolution(ref imageDataTemp, rRef + rP, cRef + cP, iterationPD, res);
                                }
                            }
                        }

                    }  //next c
                });  //next r
                if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
                else excelApp.StatusBar = string.Format("Correcting initial results ({2})... {0:0%}      DLL in use on {1} logical processors.{3}", (i + 1) / (double)n, numProcessors, numPointsMP, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
            copyArray(ref glitchFlag, glitchFlagTemp);
            if (cntFlagGlitch > 0) flagGlitch = true;
        }

        public static void rSeries5(ref double[,] imageData, ref int[,] glitchFlag, ref bool flagGlitch, ref int numSkip, double eqnParameter, bool flagSmooth, int flagDllOption, double x0, double y0, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int dr, int dc, int navSize, int rr1, int rr2, int cc1, int cc2, int rRef, int cRef, int numPointsMP, double seriesTolerance,
            double[,] xMP, double[,] yMP, double[,] glitchDetect, double divThreshold, bool flagDebugGlitch, int iteration, int maxPastRef, int[,] prevGlitchFlag, int prevGlitchInfo2)
        {  // Rotation Family, perturbation with forced 5th order series approximation
            double pixelStepP = scaleTerm * res;
            int rUpperIndex = ((rr2 - rr1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double zrP = x0 + scaleTerm * cc1;  //min real position
            double ziP = y0 - scaleTerm * rr1;  //max imaginary position
            bool skipSeriesFlag = false;
            bool flagOverflow = false;
            int maxRefSeries = 0;
            int skipNumber = 0;
            double[] A1r = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
            double[] A2r = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
            double[] A1i = new double[iteration];
            double[] A2i = new double[iteration];
            double[] B1r = new double[iteration];
            double[] B2r = new double[iteration];
            double[] B3r = new double[iteration];
            double[] B1i = new double[iteration];
            double[] B2i = new double[iteration];
            double[] B3i = new double[iteration];
            double[] C1r = new double[iteration];
            double[] C2r = new double[iteration];
            double[] C3r = new double[iteration];
            double[] C4r = new double[iteration];
            double[] C1i = new double[iteration];
            double[] C2i = new double[iteration];
            double[] C3i = new double[iteration];
            double[] C4i = new double[iteration];
            double[] D1r = new double[iteration];
            double[] D2r = new double[iteration];
            double[] D3r = new double[iteration];
            double[] D4r = new double[iteration];
            double[] D5r = new double[iteration];
            double[] D1i = new double[iteration];
            double[] D2i = new double[iteration];
            double[] D3i = new double[iteration];
            double[] D4i = new double[iteration];
            double[] D5i = new double[iteration];
            double[] E1r = new double[iteration];
            double[] E2r = new double[iteration];
            double[] E3r = new double[iteration];
            double[] E4r = new double[iteration];
            double[] E5r = new double[iteration];
            double[] E6r = new double[iteration];
            double[] E1i = new double[iteration];
            double[] E2i = new double[iteration];
            double[] E3i = new double[iteration];
            double[] E4i = new double[iteration];
            double[] E5i = new double[iteration];
            double[] E6i = new double[iteration];
            int bailout = 4;
            if (flagSmooth == true) bailout = highBailout;
            double theta = 2 * Math.PI / eqnParameter;
            double cs2, sn2, cs1, sn1;

            // The following are used to avoid passing variables by reference in the Parallel.For loop, array bases are unchanged when using the Clone method
            double[,] imageDataTemp = imageData.Clone() as double[,];
            int[,] glitchFlagTemp = glitchFlag.Clone() as int[,];
            flagGlitch = false;  //this is set in the Parallel.For loop using the Interlocked.Increment Method on cntFlagGlitch and then checking for non-zero values
            int cntFlagGlitch = 0;

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);
            else excelApp.StatusBar = string.Format("Correcting initial results ({1})... 0%      DLL in use on {0} logical processors.{2}", numProcessors, numPointsMP, dllOptionString);

            // Calculate coefficients of series approximation up to a 5th order approximation
            // When correcting glitches, series approximation is skipped if the largest glitch is small
            if (numPointsMP == 1) {
                if (iteration < 10) skipSeriesFlag = true;
            }
            else {
                if (iteration < 10 || prevGlitchInfo2 < 10) skipSeriesFlag = true;
            }
            if (skipSeriesFlag == false) {
                cs2 = Math.Cos(theta * 2);
                sn2 = Math.Sin(theta * 2);
                cs1 = Math.Cos(theta);
                sn1 = Math.Sin(theta);
                A1r[1] = 2 * (cs1 * xMP[1, 1] - sn1 * yMP[1, 1]) + cs2;
                A2r[1] = -2 * (sn1 * xMP[1, 1] + cs1 * yMP[1, 1]) - sn2;
                B1r[1] = cs1 * cs1 - sn1 * sn1;
                B2r[1] = -B1r[1];
                B3r[1] = -4 * cs1 * sn1;
                A1i[1] = -A2r[1];
                A2i[1] = A1r[1];
                B1i[1] = 2 * cs1 * sn1;
                B2i[1] = -B1i[1];
                B3i[1] = 2 * B1r[1];
                for (int i = 2; i <= iteration - 1; i++) {
                    maxRefSeries = i;
                    cs2 = Math.Cos(theta * (i + 1));
                    sn2 = Math.Sin(theta * (i + 1));
                    A1r[i] = 2 * (-A1i[i - 1] * yMP[i, 1] + A1r[i - 1] * xMP[i, 1]) + cs2;
                    A2r[i] = 2 * (-A2i[i - 1] * yMP[i, 1] + A2r[i - 1] * xMP[i, 1]) - sn2;
                    B1r[i] = -A1i[i - 1] * A1i[i - 1] + A1r[i - 1] * A1r[i - 1] + 2 * (-B1i[i - 1] * yMP[i, 1] + B1r[i - 1] * xMP[i, 1]);
                    B2r[i] = -A2i[i - 1] * A2i[i - 1] + A2r[i - 1] * A2r[i - 1] + 2 * (-B2i[i - 1] * yMP[i, 1] + B2r[i - 1] * xMP[i, 1]);
                    B3r[i] = 2 * (-A1i[i - 1] * A2i[i - 1] + A1r[i - 1] * A2r[i - 1] - B3i[i - 1] * yMP[i, 1] + B3r[i - 1] * xMP[i, 1]);
                    C1r[i] = 2 * (-A1i[i - 1] * B1i[i - 1] + A1r[i - 1] * B1r[i - 1] - C1i[i - 1] * yMP[i, 1] + C1r[i - 1] * xMP[i, 1]);
                    C2r[i] = 2 * (-A2i[i - 1] * B2i[i - 1] + A2r[i - 1] * B2r[i - 1] - C2i[i - 1] * yMP[i, 1] + C2r[i - 1] * xMP[i, 1]);
                    C3r[i] = 2 * (-A1i[i - 1] * B3i[i - 1] + A1r[i - 1] * B3r[i - 1] - A2i[i - 1] * B1i[i - 1] + A2r[i - 1] * B1r[i - 1] - C3i[i - 1] * yMP[i, 1] + C3r[i - 1] * xMP[i, 1]);
                    C4r[i] = 2 * (-A1i[i - 1] * B2i[i - 1] + A1r[i - 1] * B2r[i - 1] - A2i[i - 1] * B3i[i - 1] + A2r[i - 1] * B3r[i - 1] - C4i[i - 1] * yMP[i, 1] + C4r[i - 1] * xMP[i, 1]);
                    A1i[i] = 2 * (A1i[i - 1] * xMP[i, 1] + A1r[i - 1] * yMP[i, 1]) + sn2;
                    A2i[i] = 2 * (A2i[i - 1] * xMP[i, 1] + A2r[i - 1] * yMP[i, 1]) + cs2;
                    B1i[i] = 2 * (A1i[i - 1] * A1r[i - 1] + B1i[i - 1] * xMP[i, 1] + B1r[i - 1] * yMP[i, 1]);
                    B2i[i] = 2 * (A2i[i - 1] * A2r[i - 1] + B2i[i - 1] * xMP[i, 1] + B2r[i - 1] * yMP[i, 1]);
                    B3i[i] = 2 * (A1i[i - 1] * A2r[i - 1] + A1r[i - 1] * A2i[i - 1] + B3i[i - 1] * xMP[i, 1] + B3r[i - 1] * yMP[i, 1]);
                    C1i[i] = 2 * (A1i[i - 1] * B1r[i - 1] + A1r[i - 1] * B1i[i - 1] + C1i[i - 1] * xMP[i, 1] + C1r[i - 1] * yMP[i, 1]);
                    C2i[i] = 2 * (A2i[i - 1] * B2r[i - 1] + A2r[i - 1] * B2i[i - 1] + C2i[i - 1] * xMP[i, 1] + C2r[i - 1] * yMP[i, 1]);
                    C3i[i] = 2 * (A1i[i - 1] * B3r[i - 1] + A1r[i - 1] * B3i[i - 1] + A2i[i - 1] * B1r[i - 1] + A2r[i - 1] * B1i[i - 1] + C3i[i - 1] * xMP[i, 1] + C3r[i - 1] * yMP[i, 1]);
                    C4i[i] = 2 * (A1i[i - 1] * B2r[i - 1] + A1r[i - 1] * B2i[i - 1] + A2i[i - 1] * B3r[i - 1] + A2r[i - 1] * B3i[i - 1] + C4i[i - 1] * xMP[i, 1] + C4r[i - 1] * yMP[i, 1]);
                    D1r[i] = -B1i[i - 1] * B1i[i - 1] + B1r[i - 1] * B1r[i - 1] + 2 * (-A1i[i - 1] * C1i[i - 1] + A1r[i - 1] * C1r[i - 1] - D1i[i - 1] * yMP[i, 1] + D1r[i - 1] * xMP[i, 1]);
                    D2r[i] = -B2i[i - 1] * B2i[i - 1] + B2r[i - 1] * B2r[i - 1] + 2 * (-A2i[i - 1] * C2i[i - 1] + A2r[i - 1] * C2r[i - 1] - D2i[i - 1] * yMP[i, 1] + D2r[i - 1] * xMP[i, 1]);
                    D3r[i] = 2 * (-A1i[i - 1] * C3i[i - 1] + A1r[i - 1] * C3r[i - 1] - A2i[i - 1] * C1i[i - 1] + A2r[i - 1] * C1r[i - 1] - B1i[i - 1] * B3i[i - 1] + B1r[i - 1] * B3r[i - 1] - D3i[i - 1] * yMP[i, 1] + D3r[i - 1] * xMP[i, 1]);
                    D4r[i] = 2 * (-A1i[i - 1] * C2i[i - 1] + A1r[i - 1] * C2r[i - 1] - A2i[i - 1] * C4i[i - 1] + A2r[i - 1] * C4r[i - 1] - B2i[i - 1] * B3i[i - 1] + B2r[i - 1] * B3r[i - 1] - D4i[i - 1] * yMP[i, 1] + D4r[i - 1] * xMP[i, 1]);
                    D5r[i] = -B3i[i - 1] * B3i[i - 1] + B3r[i - 1] * B3r[i - 1] + 2 * (-A1i[i - 1] * C4i[i - 1] + A1r[i - 1] * C4r[i - 1] - A2i[i - 1] * C3i[i - 1] + A2r[i - 1] * C3r[i - 1] - B1i[i - 1] * B2i[i - 1] + B1r[i - 1] * B2r[i - 1] - D5i[i - 1] * yMP[i, 1] + D5r[i - 1] * xMP[i, 1]);
                    E1r[i] = 2 * (-A1i[i - 1] * D1i[i - 1] + A1r[i - 1] * D1r[i - 1] - B1i[i - 1] * C1i[i - 1] + B1r[i - 1] * C1r[i - 1] - E1i[i - 1] * yMP[i, 1] + E1r[i - 1] * xMP[i, 1]);
                    E2r[i] = 2 * (-A2i[i - 1] * D2i[i - 1] + A2r[i - 1] * D2r[i - 1] - B2i[i - 1] * C2i[i - 1] + B2r[i - 1] * C2r[i - 1] - E2i[i - 1] * yMP[i, 1] + E2r[i - 1] * xMP[i, 1]);
                    E3r[i] = 2 * (-A1i[i - 1] * D3i[i - 1] + A1r[i - 1] * D3r[i - 1] - A2i[i - 1] * D1i[i - 1] + A2r[i - 1] * D1r[i - 1] - B1i[i - 1] * C3i[i - 1] + B1r[i - 1] * C3r[i - 1] - B3i[i - 1] * C1i[i - 1] + B3r[i - 1] * C1r[i - 1] - E3i[i - 1] * yMP[i, 1] + E3r[i - 1] * xMP[i, 1]);
                    E4r[i] = 2 * (-A1i[i - 1] * D2i[i - 1] + A1r[i - 1] * D2r[i - 1] - A2i[i - 1] * D4i[i - 1] + A2r[i - 1] * D4r[i - 1] - B2i[i - 1] * C4i[i - 1] + B2r[i - 1] * C4r[i - 1] - B3i[i - 1] * C2i[i - 1] + B3r[i - 1] * C2r[i - 1] - E4i[i - 1] * yMP[i, 1] + E4r[i - 1] * xMP[i, 1]);
                    E5r[i] = 2 * (-A1i[i - 1] * D5i[i - 1] + A1r[i - 1] * D5r[i - 1] - A2i[i - 1] * D3i[i - 1] + A2r[i - 1] * D3r[i - 1] - B1i[i - 1] * C4i[i - 1] + B1r[i - 1] * C4r[i - 1] - B2i[i - 1] * C1i[i - 1] + B2r[i - 1] * C1r[i - 1] - B3i[i - 1] * C3i[i - 1] + B3r[i - 1] * C3r[i - 1] - E5i[i - 1] * yMP[i, 1] + E5r[i - 1] * xMP[i, 1]);
                    E6r[i] = 2 * (-A1i[i - 1] * D4i[i - 1] + A1r[i - 1] * D4r[i - 1] - A2i[i - 1] * D5i[i - 1] + A2r[i - 1] * D5r[i - 1] - B1i[i - 1] * C2i[i - 1] + B1r[i - 1] * C2r[i - 1] - B2i[i - 1] * C3i[i - 1] + B2r[i - 1] * C3r[i - 1] - B3i[i - 1] * C4i[i - 1] + B3r[i - 1] * C4r[i - 1] - E6i[i - 1] * yMP[i, 1] + E6r[i - 1] * xMP[i, 1]);
                    D1i[i] = 2 * (A1i[i - 1] * C1r[i - 1] + A1r[i - 1] * C1i[i - 1] + B1i[i - 1] * B1r[i - 1] + D1i[i - 1] * xMP[i, 1] + D1r[i - 1] * yMP[i, 1]);
                    D2i[i] = 2 * (A2i[i - 1] * C2r[i - 1] + A2r[i - 1] * C2i[i - 1] + B2i[i - 1] * B2r[i - 1] + D2i[i - 1] * xMP[i, 1] + D2r[i - 1] * yMP[i, 1]);
                    D3i[i] = 2 * (A1i[i - 1] * C3r[i - 1] + A1r[i - 1] * C3i[i - 1] + A2i[i - 1] * C1r[i - 1] + A2r[i - 1] * C1i[i - 1] + B1i[i - 1] * B3r[i - 1] + B1r[i - 1] * B3i[i - 1] + D3i[i - 1] * xMP[i, 1] + D3r[i - 1] * yMP[i, 1]);
                    D4i[i] = 2 * (A1i[i - 1] * C2r[i - 1] + A1r[i - 1] * C2i[i - 1] + A2i[i - 1] * C4r[i - 1] + A2r[i - 1] * C4i[i - 1] + B2i[i - 1] * B3r[i - 1] + B2r[i - 1] * B3i[i - 1] + D4i[i - 1] * xMP[i, 1] + D4r[i - 1] * yMP[i, 1]);
                    D5i[i] = 2 * (A1i[i - 1] * C4r[i - 1] + A1r[i - 1] * C4i[i - 1] + A2i[i - 1] * C3r[i - 1] + A2r[i - 1] * C3i[i - 1] + B1i[i - 1] * B2r[i - 1] + B1r[i - 1] * B2i[i - 1] + B3i[i - 1] * B3r[i - 1] + D5i[i - 1] * xMP[i, 1] + D5r[i - 1] * yMP[i, 1]);
                    E1i[i] = 2 * (A1i[i - 1] * D1r[i - 1] + A1r[i - 1] * D1i[i - 1] + B1i[i - 1] * C1r[i - 1] + B1r[i - 1] * C1i[i - 1] + E1i[i - 1] * xMP[i, 1] + E1r[i - 1] * yMP[i, 1]);
                    E2i[i] = 2 * (A2i[i - 1] * D2r[i - 1] + A2r[i - 1] * D2i[i - 1] + B2i[i - 1] * C2r[i - 1] + B2r[i - 1] * C2i[i - 1] + E2i[i - 1] * xMP[i, 1] + E2r[i - 1] * yMP[i, 1]);
                    E3i[i] = 2 * (A1i[i - 1] * D3r[i - 1] + A1r[i - 1] * D3i[i - 1] + A2i[i - 1] * D1r[i - 1] + A2r[i - 1] * D1i[i - 1] + B1i[i - 1] * C3r[i - 1] + B1r[i - 1] * C3i[i - 1] + B3i[i - 1] * C1r[i - 1] + B3r[i - 1] * C1i[i - 1] + E3i[i - 1] * xMP[i, 1] + E3r[i - 1] * yMP[i, 1]);
                    E4i[i] = 2 * (A1i[i - 1] * D2r[i - 1] + A1r[i - 1] * D2i[i - 1] + A2i[i - 1] * D4r[i - 1] + A2r[i - 1] * D4i[i - 1] + B2i[i - 1] * C4r[i - 1] + B2r[i - 1] * C4i[i - 1] + B3i[i - 1] * C2r[i - 1] + B3r[i - 1] * C2i[i - 1] + E4i[i - 1] * xMP[i, 1] + E4r[i - 1] * yMP[i, 1]);
                    E5i[i] = 2 * (A1i[i - 1] * D5r[i - 1] + A1r[i - 1] * D5i[i - 1] + A2i[i - 1] * D3r[i - 1] + A2r[i - 1] * D3i[i - 1] + B1i[i - 1] * C4r[i - 1] + B1r[i - 1] * C4i[i - 1] + B2i[i - 1] * C1r[i - 1] + B2r[i - 1] * C1i[i - 1] + B3i[i - 1] * C3r[i - 1] + B3r[i - 1] * C3i[i - 1] + E5i[i - 1] * xMP[i, 1] + E5r[i - 1] * yMP[i, 1]);
                    E6i[i] = 2 * (A1i[i - 1] * D4r[i - 1] + A1r[i - 1] * D4i[i - 1] + A2i[i - 1] * D5r[i - 1] + A2r[i - 1] * D5i[i - 1] + B1i[i - 1] * C2r[i - 1] + B1r[i - 1] * C2i[i - 1] + B2i[i - 1] * C3r[i - 1] + B2r[i - 1] * C3i[i - 1] + B3i[i - 1] * C4r[i - 1] + B3r[i - 1] * C4i[i - 1] + E6i[i - 1] * xMP[i, 1] + E6r[i - 1] * yMP[i, 1]);
                    // Check for overflow, unlike VBA an exception is not thrown when a double exceeds the maximum magnitude
                    if (Double.IsInfinity(E6i[i]) || Double.IsInfinity(E6r[i]) || Double.IsInfinity(E5i[i]) || Double.IsInfinity(E5r[i]) || Double.IsInfinity(E4i[i]) || Double.IsInfinity(E4r[i]) ||
                        Double.IsInfinity(E3i[i]) || Double.IsInfinity(E3r[i]) || Double.IsInfinity(E2i[i]) || Double.IsInfinity(E2r[i]) || Double.IsInfinity(E1i[i]) || Double.IsInfinity(E1r[i]) ||
                        Double.IsInfinity(D5i[i]) || Double.IsInfinity(D5r[i]) || Double.IsInfinity(D4i[i]) || Double.IsInfinity(D4r[i]) || Double.IsInfinity(D3i[i]) || Double.IsInfinity(D3r[i]) ||
                        Double.IsInfinity(D2i[i]) || Double.IsInfinity(D2r[i]) || Double.IsInfinity(D1i[i]) || Double.IsInfinity(D1r[i]) ||
                        Double.IsInfinity(C4i[i]) || Double.IsInfinity(C4r[i]) || Double.IsInfinity(C3i[i]) || Double.IsInfinity(C3r[i]) || Double.IsInfinity(C2i[i]) || Double.IsInfinity(C2r[i]) ||
                        Double.IsInfinity(C1i[i]) || Double.IsInfinity(C1r[i]) || Double.IsInfinity(B3i[i]) || Double.IsInfinity(B3r[i]) || Double.IsInfinity(B2i[i]) || Double.IsInfinity(B2r[i]) ||
                        Double.IsInfinity(B1i[i]) || Double.IsInfinity(B1r[i]) || Double.IsInfinity(A2i[i]) || Double.IsInfinity(A2r[i]) || Double.IsInfinity(A1i[i]) || Double.IsInfinity(A1r[i])) {
                        flagOverflow = true;
                        break;
                    }
                }  //next i

                // Catch overflow condition with flag
                if (flagOverflow == true) {
                    if (maxRefSeries < 10) skipSeriesFlag = true;
                    maxRefSeries = maxRefSeries - 1;
                }
            }

            // Determine how many iterations to skip
            if (skipSeriesFlag == false) {
                for (int j = 1; j <= 4; j++) {  //4 corner pixels of range
                    int rP = 0, cP = 0;
                    if (j == 1) {  //top left, min Re & max Im
                        rP = rr1;
                        cP = cc1;
                    }
                    else if (j == 2) {  //top right, max Re & max Im
                        rP = rr1;
                        cP = cc2;
                    }
                    else if (j == 3) {  //bottom left, min Re & min Im
                        rP = rr2;
                        cP = cc1;
                    }
                    else if (j == 4) {  //bottom right, max Re & min Im
                        rP = rr2;
                        cP = cc2;
                    }
                    // Note rr1 and cc1 are negative while rr2 and cc2 are positive
                    double x0P = x0 + scaleTerm * cP;  //Re position
                    double y0P = y0 - scaleTerm * rP;  //Im position

                    // Iteration loop using perturbation method, save delta values
                    double[] DxP = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
                    double[] DyP = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
                    double x, y, Dxtemp;
                    int i2 = 1;
                    double xP = x0P;
                    double yP = y0P;
                    double Dx0 = cP * scaleTerm;
                    double Dy0 = -rP * scaleTerm;
                    double cs = Math.Cos(theta);
                    double sn = Math.Sin(theta);
                    double Dx = Dx0 * cs - Dy0 * sn;  //delta1 = Y1 - X1
                    double Dy = Dx0 * sn + Dy0 * cs;
                    bool flagBreak = false;
                    DxP[0] = Dx;
                    DyP[0] = Dy;
                    for (int i = 1; i <= maxRefSeries - 1; i++) {
                        i2 = i;
                        x = xMP[i, 1];
                        y = yMP[i, 1];
                        xP = Dx + x;
                        yP = Dy + y;
                        if (xP * xP + yP * yP < glitchDetect[i, 1]) {
                            flagBreak = true;
                            break;
                        }
                        if (xP * xP + yP * yP > bailout) {
                            flagBreak = true;
                            break;
                        }
                        cs = Math.Cos(theta * (i + 1));
                        sn = Math.Sin(theta * (i + 1));
                        Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0 * cs - Dy0 * sn;
                        Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dx0 * sn + Dy0 * cs;
                        Dx = Dxtemp;
                        if (Dx * Dx + Dy * Dy > divThreshold) {
                            flagBreak = true;
                            break;
                        }
                        DxP[i] = Dx;
                        DyP[i] = Dy;
                    }  //next i
                    if (flagBreak == false) i2++;

                    // Compare delta values calculated using series approximation (DxS, DyS) to delta values calculated using
                    // perturbation theory (DxP, DyP) to determine how long the series approximation is valid
                    if (i2 < (10 + 1)) {
                        skipSeriesFlag = true;
                        break;
                    }
                    int k2 = 1;
                    double DxS, DyS;
                    flagBreak = false;
                    for (int k = 1; k <= i2 - 2; k++) {
                        k2 = k + 1;
                        DxS = A1r[k] * Dx0 + A2r[k] * Dy0 + B1r[k] * Dx0 * Dx0 + B2r[k] * Dy0 * Dy0 + B3r[k] * Dx0 * Dy0 + C1r[k] * Dx0 * Dx0 * Dx0 + C2r[k] * Dy0 * Dy0 * Dy0 + C3r[k] * Dx0 * Dx0 * Dy0 + C4r[k] * Dx0 * Dy0 * Dy0 + D1r[k] * Dx0 * Dx0 * Dx0 * Dx0 + D2r[k] * Dy0 * Dy0 * Dy0 * Dy0 + D3r[k] * Dx0 * Dx0 * Dx0 * Dy0 + D4r[k] * Dx0 * Dy0 * Dy0 * Dy0 + D5r[k] * Dx0 * Dx0 * Dy0 * Dy0 + E1r[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2r[k] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3r[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4r[k] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5r[k] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6r[k] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                        DyS = A1i[k] * Dx0 + A2i[k] * Dy0 + B1i[k] * Dx0 * Dx0 + B2i[k] * Dy0 * Dy0 + B3i[k] * Dx0 * Dy0 + C1i[k] * Dx0 * Dx0 * Dx0 + C2i[k] * Dy0 * Dy0 * Dy0 + C3i[k] * Dx0 * Dx0 * Dy0 + C4i[k] * Dx0 * Dy0 * Dy0 + D1i[k] * Dx0 * Dx0 * Dx0 * Dx0 + D2i[k] * Dy0 * Dy0 * Dy0 * Dy0 + D3i[k] * Dx0 * Dx0 * Dx0 * Dy0 + D4i[k] * Dx0 * Dy0 * Dy0 * Dy0 + D5i[k] * Dx0 * Dx0 * Dy0 * Dy0 + E1i[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2i[k] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3i[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4i[k] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5i[k] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6i[k] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                        if (DxP[k] == 0 || DyP[k] == 0) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DxP[k] - DxS) / DxP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DyP[k] - DyS) / DyP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                    }
                    if (flagBreak == true) k2 = k2 - 1;  //go back to when relative difference was below tolerance, this line is surprisingly important
                    if (k2 < 5) {
                        skipSeriesFlag = true;
                        skipNumber = 0;
                        break;
                    }
                    if (j == 1) {
                        skipNumber = k2;
                    }
                    else {
                        if (k2 < skipNumber) skipNumber = k2;
                    }
                }  //next j
            }
            if (numPointsMP == 1) numSkip = skipNumber;  //store number of iterations skipped in initial pass for calculation info output

            // Calculate points about reference point orbit in double precision

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int rP = res * ri + rr1;       //mapping the parallel loop index to the row number relative to the reference point
                    double y0P = ziP - pixelStepP * ri;
                    double x0P = zrP - pixelStepP;
                    double x, y;
                    double xP = 0, xtempP, yP = 0;
                    double Dx = 0, Dxtemp, Dy = 0, Dx0, Dy0;
                    int iterationP = 0;
                    bool skipCalc, goToGlitch;
                    double iterationPD = 0;
                    double log_zn, nu;
                    double cs, sn;

                    for (int cP = cc1; cP <= cc2; cP = cP + res) {
                        skipCalc = false;
                        x0P = x0P + pixelStepP;
                        if ((cP == 0 && rP == 0) == false) {
                            if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] == prevMaxIter) || (refreshType == 2 && numPointsMP > 1)) {

                                // Quick pan code
                                if (refreshType == -1 && numPointsMP == 1) {
                                    skipCalc = quickPan(dr, dc, rRef + rP, cRef + cP, navSize);
                                }

                                // If correcting a flagged glitch, only correct location that is flagged
                                if (numPointsMP > 1) {
                                    if (prevGlitchFlag[rRef + rP, cRef + cP] == 0) skipCalc = true;
                                }

                                // Series approximation to skip iterations
                                if (skipCalc == false) {
                                    int m = skipNumber;
                                    int k = 1;
                                    bool tryLessFlag = false;
                                    bool goToTryLess = true;
                                    Dx0 = cP * scaleTerm;
                                    Dy0 = -rP * scaleTerm;
                                    if (skipSeriesFlag == false) {
                                        while (goToTryLess == true) {
                                            goToTryLess = false;
                                            Dx = A1r[m - 1] * Dx0 + A2r[m - 1] * Dy0 + B1r[m - 1] * Dx0 * Dx0 + B2r[m - 1] * Dy0 * Dy0 + B3r[m - 1] * Dx0 * Dy0 + C1r[m - 1] * Dx0 * Dx0 * Dx0 + C2r[m - 1] * Dy0 * Dy0 * Dy0 + C3r[m - 1] * Dx0 * Dx0 * Dy0 + C4r[m - 1] * Dx0 * Dy0 * Dy0 + D1r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 + D2r[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 + D3r[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 + D4r[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 + D5r[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 + E1r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2r[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4r[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5r[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6r[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                                            Dy = A1i[m - 1] * Dx0 + A2i[m - 1] * Dy0 + B1i[m - 1] * Dx0 * Dx0 + B2i[m - 1] * Dy0 * Dy0 + B3i[m - 1] * Dx0 * Dy0 + C1i[m - 1] * Dx0 * Dx0 * Dx0 + C2i[m - 1] * Dy0 * Dy0 * Dy0 + C3i[m - 1] * Dx0 * Dx0 * Dy0 + C4i[m - 1] * Dx0 * Dy0 * Dy0 + D1i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 + D2i[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 + D3i[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 + D4i[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 + D5i[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 + E1i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2i[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4i[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5i[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6i[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                                            if (Dx * Dx + Dy * Dy > divThreshold) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            xP = Dx + xMP[m, 1];
                                            yP = Dy + yMP[m, 1];
                                            if (xP * xP + yP * yP > bailout) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            if (goToTryLess == false) {
                                                k = m;  //skip first m-1 iterations
                                            }
                                            else {
                                                m = m / 2;
                                            }
                                        }  //Wend
                                    }

                                    // Iteration loop
                                    if (k == 1) {  //no series approximation used
                                        cs = Math.Cos(theta);
                                        sn = Math.Sin(theta);
                                        Dx = Dx0 * cs - Dy0 * sn;  //delta1 = Y1 - X1
                                        Dy = Dx0 * sn + Dy0 * cs;
                                    }

                                    goToGlitch = false;
                                    for (int j = k; j <= iteration; j++) {
                                        iterationP = j;
                                        x = xMP[j, 1];
                                        y = yMP[j, 1];
                                        xP = Dx + x;
                                        yP = Dy + y;
                                        if (xP * xP + yP * yP < glitchDetect[j, 1]) {
                                            glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                            Interlocked.Increment(ref cntFlagGlitch);
                                            if (flagDebugGlitch == true) iterationPD = maxIter;
                                            goToGlitch = true;
                                            break;
                                        }
                                        if (xP * xP + yP * yP > bailout) break;
                                        cs = Math.Cos(theta * (j + 1));
                                        sn = Math.Sin(theta * (j + 1));
                                        Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0 * cs - Dy0 * sn;
                                        Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dx0 * sn + Dy0 * cs;
                                        Dx = Dxtemp;
                                        if (Dx * Dx + Dy * Dy > divThreshold) break;
                                    }  //next j

                                    if (goToGlitch == false) {

                                        // Discrete coloring
                                        if (flagSmooth == false) {
                                            while (xP * xP + yP * yP < 4 && iterationP < maxIter) {
                                                iterationP++;
                                                cs = Math.Cos(theta * iterationP);
                                                sn = Math.Sin(theta * iterationP);
                                                xtempP = xP * xP - yP * yP + x0P * cs - y0P * sn;
                                                yP = 2 * xP * yP + x0P * sn + y0P * cs;
                                                xP = xtempP;
                                            }
                                            iterationPD = iterationP;
                                        }

                                        // Continuous coloring
                                        else {
                                            while (xP * xP + yP * yP < highBailout && iterationP < maxIter) {
                                                iterationP++;
                                                cs = Math.Cos(theta * iterationP);
                                                sn = Math.Sin(theta * iterationP);
                                                xtempP = xP * xP - yP * yP + x0P * cs - y0P * sn;
                                                yP = 2 * xP * yP + x0P * sn + y0P * cs;
                                                xP = xtempP;
                                            }

                                            // Additional calculations for continuous coloring
                                            if (iterationP < maxIter) {
                                                log_zn = Math.Log(xP * xP + yP * yP) / 2;
                                                nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                                iterationPD = iterationP + 1 - nu;
                                                iterationPD = Math.Round(iterationPD, 3);
                                            }
                                            else {
                                                iterationPD = iterationP;
                                            }
                                        }

                                    }

                                    // Check if too many additional iterations past the reference were performed
                                    if (maxPastRef != 0 && iterationP > maxPastRef) {
                                        glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                        Interlocked.Increment(ref cntFlagGlitch);
                                        if (flagDebugGlitch == true) iterationPD = maxIter;
                                    }

                                    // Store result to array
                                    if (res == 1) imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] = iterationPD;
                                    else reduceResolution(ref imageDataTemp, rRef + rP, cRef + cP, iterationPD, res);
                                }
                            }
                        }

                    }  //next c
                });  //next r
                if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
                else excelApp.StatusBar = string.Format("Correcting initial results ({2})... {0:0%}      DLL in use on {1} logical processors.{3}", (i + 1) / (double)n, numProcessors, numPointsMP, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
            copyArray(ref glitchFlag, glitchFlagTemp);
            if (cntFlagGlitch > 0) flagGlitch = true;
        }

        public static void rSeries(ref double[,] imageData, ref int[,] glitchFlag, ref bool flagGlitch, ref int numSkip, double eqnParameter, bool flagSmooth, int flagDllOption, double x0, double y0, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int dr, int dc, int navSize, int rr1, int rr2, int cc1, int cc2, int rRef, int cRef, int numPointsMP, double seriesTolerance,
            double[,] xMP, double[,] yMP, double[,] glitchDetect, double divThreshold, bool flagDebugGlitch, int iteration, int maxPastRef, int[,] prevGlitchFlag, int prevGlitchInfo2)
        {  // Rotation Family, perturbation with series approximation
            double pixelStepP = scaleTerm * res;
            int rUpperIndex = ((rr2 - rr1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double zrP = x0 + scaleTerm * cc1;  //min real position
            double ziP = y0 - scaleTerm * rr1;  //max imaginary position
            bool skipSeriesFlag = false;
            bool flagOverflow = false;
            int maxRefSeries = 0;
            int skipNumber = 0;
            double[] A1r = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
            double[] A2r = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
            double[] A1i = new double[iteration];
            double[] A2i = new double[iteration];
            double[] B1r = new double[iteration];
            double[] B2r = new double[iteration];
            double[] B3r = new double[iteration];
            double[] B1i = new double[iteration];
            double[] B2i = new double[iteration];
            double[] B3i = new double[iteration];
            double[] C1r = new double[iteration];
            double[] C2r = new double[iteration];
            double[] C3r = new double[iteration];
            double[] C4r = new double[iteration];
            double[] C1i = new double[iteration];
            double[] C2i = new double[iteration];
            double[] C3i = new double[iteration];
            double[] C4i = new double[iteration];
            double[] D1r = new double[iteration];
            double[] D2r = new double[iteration];
            double[] D3r = new double[iteration];
            double[] D4r = new double[iteration];
            double[] D5r = new double[iteration];
            double[] D1i = new double[iteration];
            double[] D2i = new double[iteration];
            double[] D3i = new double[iteration];
            double[] D4i = new double[iteration];
            double[] D5i = new double[iteration];
            double[] E1r = new double[iteration];
            double[] E2r = new double[iteration];
            double[] E3r = new double[iteration];
            double[] E4r = new double[iteration];
            double[] E5r = new double[iteration];
            double[] E6r = new double[iteration];
            double[] E1i = new double[iteration];
            double[] E2i = new double[iteration];
            double[] E3i = new double[iteration];
            double[] E4i = new double[iteration];
            double[] E5i = new double[iteration];
            double[] E6i = new double[iteration];
            bool calculatingO3Flag = false, calcO5flag = false, useOrder3 = false;
            int maxO3Series = 0, maxO5Series = 0, maxSeries = 0;
            int skipNumberO3 = 0, skipNumberO5 = 0;
            int i3 = 0;
            int bailout = 4;
            if (flagSmooth == true) bailout = highBailout;
            double theta = 2 * Math.PI / eqnParameter;
            double cs2, sn2, cs1, sn1;

            // The following are used to avoid passing variables by reference in the Parallel.For loop, array bases are unchanged when using the Clone method
            double[,] imageDataTemp = imageData.Clone() as double[,];
            int[,] glitchFlagTemp = glitchFlag.Clone() as int[,];
            flagGlitch = false;  //this is set in the Parallel.For loop using the Interlocked.Increment Method on cntFlagGlitch and then checking for non-zero values
            int cntFlagGlitch = 0;

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);
            else excelApp.StatusBar = string.Format("Correcting initial results ({1})... 0%      DLL in use on {0} logical processors.{2}", numProcessors, numPointsMP, dllOptionString);

            // Calculate coefficients of series approximation up to a 5th order approximation
            // When correcting glitches, series approximation is skipped if the largest glitch is small
            if (numPointsMP == 1) {
                if (iteration < 10) skipSeriesFlag = true;
            }
            else {
                if (iteration < 10 || prevGlitchInfo2 < 10) skipSeriesFlag = true;
            }
            if (skipSeriesFlag == false) {
                cs2 = Math.Cos(theta * 2);
                sn2 = Math.Sin(theta * 2);
                cs1 = Math.Cos(theta);
                sn1 = Math.Sin(theta);
                A1r[1] = 2 * (cs1 * xMP[1, 1] - sn1 * yMP[1, 1]) + cs2;
                A2r[1] = -2 * (sn1 * xMP[1, 1] + cs1 * yMP[1, 1]) - sn2;
                B1r[1] = cs1 * cs1 - sn1 * sn1;
                B2r[1] = -B1r[1];
                B3r[1] = -4 * cs1 * sn1;
                A1i[1] = -A2r[1];
                A2i[1] = A1r[1];
                B1i[1] = 2 * cs1 * sn1;
                B2i[1] = -B1i[1];
                B3i[1] = 2 * B1r[1];
                calcO5flag = true;
                for (int i = 2; i <= iteration - 1; i++) {
                    maxRefSeries = i;
                    calculatingO3Flag = true;
                    cs2 = Math.Cos(theta * (i + 1));
                    sn2 = Math.Sin(theta * (i + 1));
                    A1r[i] = 2 * (-A1i[i - 1] * yMP[i, 1] + A1r[i - 1] * xMP[i, 1]) + cs2;
                    A2r[i] = 2 * (-A2i[i - 1] * yMP[i, 1] + A2r[i - 1] * xMP[i, 1]) - sn2;
                    B1r[i] = -A1i[i - 1] * A1i[i - 1] + A1r[i - 1] * A1r[i - 1] + 2 * (-B1i[i - 1] * yMP[i, 1] + B1r[i - 1] * xMP[i, 1]);
                    B2r[i] = -A2i[i - 1] * A2i[i - 1] + A2r[i - 1] * A2r[i - 1] + 2 * (-B2i[i - 1] * yMP[i, 1] + B2r[i - 1] * xMP[i, 1]);
                    B3r[i] = 2 * (-A1i[i - 1] * A2i[i - 1] + A1r[i - 1] * A2r[i - 1] - B3i[i - 1] * yMP[i, 1] + B3r[i - 1] * xMP[i, 1]);
                    C1r[i] = 2 * (-A1i[i - 1] * B1i[i - 1] + A1r[i - 1] * B1r[i - 1] - C1i[i - 1] * yMP[i, 1] + C1r[i - 1] * xMP[i, 1]);
                    C2r[i] = 2 * (-A2i[i - 1] * B2i[i - 1] + A2r[i - 1] * B2r[i - 1] - C2i[i - 1] * yMP[i, 1] + C2r[i - 1] * xMP[i, 1]);
                    C3r[i] = 2 * (-A1i[i - 1] * B3i[i - 1] + A1r[i - 1] * B3r[i - 1] - A2i[i - 1] * B1i[i - 1] + A2r[i - 1] * B1r[i - 1] - C3i[i - 1] * yMP[i, 1] + C3r[i - 1] * xMP[i, 1]);
                    C4r[i] = 2 * (-A1i[i - 1] * B2i[i - 1] + A1r[i - 1] * B2r[i - 1] - A2i[i - 1] * B3i[i - 1] + A2r[i - 1] * B3r[i - 1] - C4i[i - 1] * yMP[i, 1] + C4r[i - 1] * xMP[i, 1]);
                    A1i[i] = 2 * (A1i[i - 1] * xMP[i, 1] + A1r[i - 1] * yMP[i, 1]) + sn2;
                    A2i[i] = 2 * (A2i[i - 1] * xMP[i, 1] + A2r[i - 1] * yMP[i, 1]) + cs2;
                    B1i[i] = 2 * (A1i[i - 1] * A1r[i - 1] + B1i[i - 1] * xMP[i, 1] + B1r[i - 1] * yMP[i, 1]);
                    B2i[i] = 2 * (A2i[i - 1] * A2r[i - 1] + B2i[i - 1] * xMP[i, 1] + B2r[i - 1] * yMP[i, 1]);
                    B3i[i] = 2 * (A1i[i - 1] * A2r[i - 1] + A1r[i - 1] * A2i[i - 1] + B3i[i - 1] * xMP[i, 1] + B3r[i - 1] * yMP[i, 1]);
                    C1i[i] = 2 * (A1i[i - 1] * B1r[i - 1] + A1r[i - 1] * B1i[i - 1] + C1i[i - 1] * xMP[i, 1] + C1r[i - 1] * yMP[i, 1]);
                    C2i[i] = 2 * (A2i[i - 1] * B2r[i - 1] + A2r[i - 1] * B2i[i - 1] + C2i[i - 1] * xMP[i, 1] + C2r[i - 1] * yMP[i, 1]);
                    C3i[i] = 2 * (A1i[i - 1] * B3r[i - 1] + A1r[i - 1] * B3i[i - 1] + A2i[i - 1] * B1r[i - 1] + A2r[i - 1] * B1i[i - 1] + C3i[i - 1] * xMP[i, 1] + C3r[i - 1] * yMP[i, 1]);
                    C4i[i] = 2 * (A1i[i - 1] * B2r[i - 1] + A1r[i - 1] * B2i[i - 1] + A2i[i - 1] * B3r[i - 1] + A2r[i - 1] * B3i[i - 1] + C4i[i - 1] * xMP[i, 1] + C4r[i - 1] * yMP[i, 1]);
                    // Check for overflow, unlike VBA an exception is not thrown when a double exceeds the maximum magnitude
                    if (Double.IsInfinity(C4i[i]) || Double.IsInfinity(C4r[i]) || Double.IsInfinity(C3i[i]) || Double.IsInfinity(C3r[i]) || Double.IsInfinity(C2i[i]) || Double.IsInfinity(C2r[i]) ||
                        Double.IsInfinity(C1i[i]) || Double.IsInfinity(C1r[i]) || Double.IsInfinity(B3i[i]) || Double.IsInfinity(B3r[i]) || Double.IsInfinity(B2i[i]) || Double.IsInfinity(B2r[i]) ||
                        Double.IsInfinity(B1i[i]) || Double.IsInfinity(B1r[i]) || Double.IsInfinity(A2i[i]) || Double.IsInfinity(A2r[i]) || Double.IsInfinity(A1i[i]) || Double.IsInfinity(A1r[i])) {
                        flagOverflow = true;
                        break;
                    }
                    calculatingO3Flag = false;
                    D1r[i] = -B1i[i - 1] * B1i[i - 1] + B1r[i - 1] * B1r[i - 1] + 2 * (-A1i[i - 1] * C1i[i - 1] + A1r[i - 1] * C1r[i - 1] - D1i[i - 1] * yMP[i, 1] + D1r[i - 1] * xMP[i, 1]);
                    D2r[i] = -B2i[i - 1] * B2i[i - 1] + B2r[i - 1] * B2r[i - 1] + 2 * (-A2i[i - 1] * C2i[i - 1] + A2r[i - 1] * C2r[i - 1] - D2i[i - 1] * yMP[i, 1] + D2r[i - 1] * xMP[i, 1]);
                    D3r[i] = 2 * (-A1i[i - 1] * C3i[i - 1] + A1r[i - 1] * C3r[i - 1] - A2i[i - 1] * C1i[i - 1] + A2r[i - 1] * C1r[i - 1] - B1i[i - 1] * B3i[i - 1] + B1r[i - 1] * B3r[i - 1] - D3i[i - 1] * yMP[i, 1] + D3r[i - 1] * xMP[i, 1]);
                    D4r[i] = 2 * (-A1i[i - 1] * C2i[i - 1] + A1r[i - 1] * C2r[i - 1] - A2i[i - 1] * C4i[i - 1] + A2r[i - 1] * C4r[i - 1] - B2i[i - 1] * B3i[i - 1] + B2r[i - 1] * B3r[i - 1] - D4i[i - 1] * yMP[i, 1] + D4r[i - 1] * xMP[i, 1]);
                    D5r[i] = -B3i[i - 1] * B3i[i - 1] + B3r[i - 1] * B3r[i - 1] + 2 * (-A1i[i - 1] * C4i[i - 1] + A1r[i - 1] * C4r[i - 1] - A2i[i - 1] * C3i[i - 1] + A2r[i - 1] * C3r[i - 1] - B1i[i - 1] * B2i[i - 1] + B1r[i - 1] * B2r[i - 1] - D5i[i - 1] * yMP[i, 1] + D5r[i - 1] * xMP[i, 1]);
                    E1r[i] = 2 * (-A1i[i - 1] * D1i[i - 1] + A1r[i - 1] * D1r[i - 1] - B1i[i - 1] * C1i[i - 1] + B1r[i - 1] * C1r[i - 1] - E1i[i - 1] * yMP[i, 1] + E1r[i - 1] * xMP[i, 1]);
                    E2r[i] = 2 * (-A2i[i - 1] * D2i[i - 1] + A2r[i - 1] * D2r[i - 1] - B2i[i - 1] * C2i[i - 1] + B2r[i - 1] * C2r[i - 1] - E2i[i - 1] * yMP[i, 1] + E2r[i - 1] * xMP[i, 1]);
                    E3r[i] = 2 * (-A1i[i - 1] * D3i[i - 1] + A1r[i - 1] * D3r[i - 1] - A2i[i - 1] * D1i[i - 1] + A2r[i - 1] * D1r[i - 1] - B1i[i - 1] * C3i[i - 1] + B1r[i - 1] * C3r[i - 1] - B3i[i - 1] * C1i[i - 1] + B3r[i - 1] * C1r[i - 1] - E3i[i - 1] * yMP[i, 1] + E3r[i - 1] * xMP[i, 1]);
                    E4r[i] = 2 * (-A1i[i - 1] * D2i[i - 1] + A1r[i - 1] * D2r[i - 1] - A2i[i - 1] * D4i[i - 1] + A2r[i - 1] * D4r[i - 1] - B2i[i - 1] * C4i[i - 1] + B2r[i - 1] * C4r[i - 1] - B3i[i - 1] * C2i[i - 1] + B3r[i - 1] * C2r[i - 1] - E4i[i - 1] * yMP[i, 1] + E4r[i - 1] * xMP[i, 1]);
                    E5r[i] = 2 * (-A1i[i - 1] * D5i[i - 1] + A1r[i - 1] * D5r[i - 1] - A2i[i - 1] * D3i[i - 1] + A2r[i - 1] * D3r[i - 1] - B1i[i - 1] * C4i[i - 1] + B1r[i - 1] * C4r[i - 1] - B2i[i - 1] * C1i[i - 1] + B2r[i - 1] * C1r[i - 1] - B3i[i - 1] * C3i[i - 1] + B3r[i - 1] * C3r[i - 1] - E5i[i - 1] * yMP[i, 1] + E5r[i - 1] * xMP[i, 1]);
                    E6r[i] = 2 * (-A1i[i - 1] * D4i[i - 1] + A1r[i - 1] * D4r[i - 1] - A2i[i - 1] * D5i[i - 1] + A2r[i - 1] * D5r[i - 1] - B1i[i - 1] * C2i[i - 1] + B1r[i - 1] * C2r[i - 1] - B2i[i - 1] * C3i[i - 1] + B2r[i - 1] * C3r[i - 1] - B3i[i - 1] * C4i[i - 1] + B3r[i - 1] * C4r[i - 1] - E6i[i - 1] * yMP[i, 1] + E6r[i - 1] * xMP[i, 1]);
                    D1i[i] = 2 * (A1i[i - 1] * C1r[i - 1] + A1r[i - 1] * C1i[i - 1] + B1i[i - 1] * B1r[i - 1] + D1i[i - 1] * xMP[i, 1] + D1r[i - 1] * yMP[i, 1]);
                    D2i[i] = 2 * (A2i[i - 1] * C2r[i - 1] + A2r[i - 1] * C2i[i - 1] + B2i[i - 1] * B2r[i - 1] + D2i[i - 1] * xMP[i, 1] + D2r[i - 1] * yMP[i, 1]);
                    D3i[i] = 2 * (A1i[i - 1] * C3r[i - 1] + A1r[i - 1] * C3i[i - 1] + A2i[i - 1] * C1r[i - 1] + A2r[i - 1] * C1i[i - 1] + B1i[i - 1] * B3r[i - 1] + B1r[i - 1] * B3i[i - 1] + D3i[i - 1] * xMP[i, 1] + D3r[i - 1] * yMP[i, 1]);
                    D4i[i] = 2 * (A1i[i - 1] * C2r[i - 1] + A1r[i - 1] * C2i[i - 1] + A2i[i - 1] * C4r[i - 1] + A2r[i - 1] * C4i[i - 1] + B2i[i - 1] * B3r[i - 1] + B2r[i - 1] * B3i[i - 1] + D4i[i - 1] * xMP[i, 1] + D4r[i - 1] * yMP[i, 1]);
                    D5i[i] = 2 * (A1i[i - 1] * C4r[i - 1] + A1r[i - 1] * C4i[i - 1] + A2i[i - 1] * C3r[i - 1] + A2r[i - 1] * C3i[i - 1] + B1i[i - 1] * B2r[i - 1] + B1r[i - 1] * B2i[i - 1] + B3i[i - 1] * B3r[i - 1] + D5i[i - 1] * xMP[i, 1] + D5r[i - 1] * yMP[i, 1]);
                    E1i[i] = 2 * (A1i[i - 1] * D1r[i - 1] + A1r[i - 1] * D1i[i - 1] + B1i[i - 1] * C1r[i - 1] + B1r[i - 1] * C1i[i - 1] + E1i[i - 1] * xMP[i, 1] + E1r[i - 1] * yMP[i, 1]);
                    E2i[i] = 2 * (A2i[i - 1] * D2r[i - 1] + A2r[i - 1] * D2i[i - 1] + B2i[i - 1] * C2r[i - 1] + B2r[i - 1] * C2i[i - 1] + E2i[i - 1] * xMP[i, 1] + E2r[i - 1] * yMP[i, 1]);
                    E3i[i] = 2 * (A1i[i - 1] * D3r[i - 1] + A1r[i - 1] * D3i[i - 1] + A2i[i - 1] * D1r[i - 1] + A2r[i - 1] * D1i[i - 1] + B1i[i - 1] * C3r[i - 1] + B1r[i - 1] * C3i[i - 1] + B3i[i - 1] * C1r[i - 1] + B3r[i - 1] * C1i[i - 1] + E3i[i - 1] * xMP[i, 1] + E3r[i - 1] * yMP[i, 1]);
                    E4i[i] = 2 * (A1i[i - 1] * D2r[i - 1] + A1r[i - 1] * D2i[i - 1] + A2i[i - 1] * D4r[i - 1] + A2r[i - 1] * D4i[i - 1] + B2i[i - 1] * C4r[i - 1] + B2r[i - 1] * C4i[i - 1] + B3i[i - 1] * C2r[i - 1] + B3r[i - 1] * C2i[i - 1] + E4i[i - 1] * xMP[i, 1] + E4r[i - 1] * yMP[i, 1]);
                    E5i[i] = 2 * (A1i[i - 1] * D5r[i - 1] + A1r[i - 1] * D5i[i - 1] + A2i[i - 1] * D3r[i - 1] + A2r[i - 1] * D3i[i - 1] + B1i[i - 1] * C4r[i - 1] + B1r[i - 1] * C4i[i - 1] + B2i[i - 1] * C1r[i - 1] + B2r[i - 1] * C1i[i - 1] + B3i[i - 1] * C3r[i - 1] + B3r[i - 1] * C3i[i - 1] + E5i[i - 1] * xMP[i, 1] + E5r[i - 1] * yMP[i, 1]);
                    E6i[i] = 2 * (A1i[i - 1] * D4r[i - 1] + A1r[i - 1] * D4i[i - 1] + A2i[i - 1] * D5r[i - 1] + A2r[i - 1] * D5i[i - 1] + B1i[i - 1] * C2r[i - 1] + B1r[i - 1] * C2i[i - 1] + B2i[i - 1] * C3r[i - 1] + B2r[i - 1] * C3i[i - 1] + B3i[i - 1] * C4r[i - 1] + B3r[i - 1] * C4i[i - 1] + E6i[i - 1] * xMP[i, 1] + E6r[i - 1] * yMP[i, 1]);
                    if (Double.IsInfinity(E6i[i]) || Double.IsInfinity(E6r[i]) || Double.IsInfinity(E5i[i]) || Double.IsInfinity(E5r[i]) || Double.IsInfinity(E4i[i]) || Double.IsInfinity(E4r[i]) ||
                        Double.IsInfinity(E3i[i]) || Double.IsInfinity(E3r[i]) || Double.IsInfinity(E2i[i]) || Double.IsInfinity(E2r[i]) || Double.IsInfinity(E1i[i]) || Double.IsInfinity(E1r[i]) ||
                        Double.IsInfinity(D5i[i]) || Double.IsInfinity(D5r[i]) || Double.IsInfinity(D4i[i]) || Double.IsInfinity(D4r[i]) || Double.IsInfinity(D3i[i]) || Double.IsInfinity(D3r[i]) ||
                        Double.IsInfinity(D2i[i]) || Double.IsInfinity(D2r[i]) || Double.IsInfinity(D1i[i]) || Double.IsInfinity(D1r[i])) {
                        flagOverflow = true;
                        break;
                    }
                }  //next i

                // Catch overflow condition with flag
                if (flagOverflow == true) {
                    if (calculatingO3Flag == false) {  //overflow occurred in order 4 or 5 term
                        maxO5Series = maxRefSeries - 1;
                        flagOverflow = false;  //reset flag
                        calcO5flag = false;
                        i3 = maxRefSeries;
                    }
                    else {  //overflow occurred in order 3 term
                        maxO3Series = maxRefSeries - 1;
                        maxRefSeries = maxRefSeries - 1;
                    }
                }

                // If overflow occurred in order 4 or 5 term, calculate additional order 3 terms if possible
                if (calcO5flag == false) {  //OverflowReturnIn
                    for (int i = i3 + 1; i <= iteration - 1; i++) {
                        maxRefSeries = i;
                        cs2 = Math.Cos(theta * (i + 1));
                        sn2 = Math.Sin(theta * (i + 1));
                        A1r[i] = 2 * (-A1i[i - 1] * yMP[i, 1] + A1r[i - 1] * xMP[i, 1]) + cs2;
                        A2r[i] = 2 * (-A2i[i - 1] * yMP[i, 1] + A2r[i - 1] * xMP[i, 1]) - sn2;
                        B1r[i] = -A1i[i - 1] * A1i[i - 1] + A1r[i - 1] * A1r[i - 1] + 2 * (-B1i[i - 1] * yMP[i, 1] + B1r[i - 1] * xMP[i, 1]);
                        B2r[i] = -A2i[i - 1] * A2i[i - 1] + A2r[i - 1] * A2r[i - 1] + 2 * (-B2i[i - 1] * yMP[i, 1] + B2r[i - 1] * xMP[i, 1]);
                        B3r[i] = 2 * (-A1i[i - 1] * A2i[i - 1] + A1r[i - 1] * A2r[i - 1] - B3i[i - 1] * yMP[i, 1] + B3r[i - 1] * xMP[i, 1]);
                        C1r[i] = 2 * (-A1i[i - 1] * B1i[i - 1] + A1r[i - 1] * B1r[i - 1] - C1i[i - 1] * yMP[i, 1] + C1r[i - 1] * xMP[i, 1]);
                        C2r[i] = 2 * (-A2i[i - 1] * B2i[i - 1] + A2r[i - 1] * B2r[i - 1] - C2i[i - 1] * yMP[i, 1] + C2r[i - 1] * xMP[i, 1]);
                        C3r[i] = 2 * (-A1i[i - 1] * B3i[i - 1] + A1r[i - 1] * B3r[i - 1] - A2i[i - 1] * B1i[i - 1] + A2r[i - 1] * B1r[i - 1] - C3i[i - 1] * yMP[i, 1] + C3r[i - 1] * xMP[i, 1]);
                        C4r[i] = 2 * (-A1i[i - 1] * B2i[i - 1] + A1r[i - 1] * B2r[i - 1] - A2i[i - 1] * B3i[i - 1] + A2r[i - 1] * B3r[i - 1] - C4i[i - 1] * yMP[i, 1] + C4r[i - 1] * xMP[i, 1]);
                        A1i[i] = 2 * (A1i[i - 1] * xMP[i, 1] + A1r[i - 1] * yMP[i, 1]) + sn2;
                        A2i[i] = 2 * (A2i[i - 1] * xMP[i, 1] + A2r[i - 1] * yMP[i, 1]) + cs2;
                        B1i[i] = 2 * (A1i[i - 1] * A1r[i - 1] + B1i[i - 1] * xMP[i, 1] + B1r[i - 1] * yMP[i, 1]);
                        B2i[i] = 2 * (A2i[i - 1] * A2r[i - 1] + B2i[i - 1] * xMP[i, 1] + B2r[i - 1] * yMP[i, 1]);
                        B3i[i] = 2 * (A1i[i - 1] * A2r[i - 1] + A1r[i - 1] * A2i[i - 1] + B3i[i - 1] * xMP[i, 1] + B3r[i - 1] * yMP[i, 1]);
                        C1i[i] = 2 * (A1i[i - 1] * B1r[i - 1] + A1r[i - 1] * B1i[i - 1] + C1i[i - 1] * xMP[i, 1] + C1r[i - 1] * yMP[i, 1]);
                        C2i[i] = 2 * (A2i[i - 1] * B2r[i - 1] + A2r[i - 1] * B2i[i - 1] + C2i[i - 1] * xMP[i, 1] + C2r[i - 1] * yMP[i, 1]);
                        C3i[i] = 2 * (A1i[i - 1] * B3r[i - 1] + A1r[i - 1] * B3i[i - 1] + A2i[i - 1] * B1r[i - 1] + A2r[i - 1] * B1i[i - 1] + C3i[i - 1] * xMP[i, 1] + C3r[i - 1] * yMP[i, 1]);
                        C4i[i] = 2 * (A1i[i - 1] * B2r[i - 1] + A1r[i - 1] * B2i[i - 1] + A2i[i - 1] * B3r[i - 1] + A2r[i - 1] * B3i[i - 1] + C4i[i - 1] * xMP[i, 1] + C4r[i - 1] * yMP[i, 1]);
                        if (Double.IsInfinity(C4i[i]) || Double.IsInfinity(C4r[i]) || Double.IsInfinity(C3i[i]) || Double.IsInfinity(C3r[i]) || Double.IsInfinity(C2i[i]) || Double.IsInfinity(C2r[i]) ||
                            Double.IsInfinity(C1i[i]) || Double.IsInfinity(C1r[i]) || Double.IsInfinity(B3i[i]) || Double.IsInfinity(B3r[i]) || Double.IsInfinity(B2i[i]) || Double.IsInfinity(B2r[i]) ||
                            Double.IsInfinity(B1i[i]) || Double.IsInfinity(B1r[i]) || Double.IsInfinity(A2i[i]) || Double.IsInfinity(A2r[i]) || Double.IsInfinity(A1i[i]) || Double.IsInfinity(A1r[i])) {
                            flagOverflow = true;
                            break;
                        }
                    }  //next i
                    if (flagOverflow == true) {
                        maxO3Series = maxRefSeries - 1;
                        maxRefSeries = maxRefSeries - 1;
                    }
                }
            }
            if (maxO3Series == 0) maxO3Series = maxRefSeries;
            if (maxO5Series == 0) maxO5Series = maxRefSeries;
            if (maxRefSeries < 10) skipSeriesFlag = true;

            // Determine how many iterations to skip
            if (skipSeriesFlag == false) {
                for (int j = 1; j <= 4; j++) {  //4 corner pixels of range
                    int rP = 0, cP = 0;
                    if (j == 1) {  //top left, min Re & max Im
                        rP = rr1;
                        cP = cc1;
                    }
                    else if (j == 2) {  //top right, max Re & max Im
                        rP = rr1;
                        cP = cc2;
                    }
                    else if (j == 3) {  //bottom left, min Re & min Im
                        rP = rr2;
                        cP = cc1;
                    }
                    else if (j == 4) {  //bottom right, max Re & min Im
                        rP = rr2;
                        cP = cc2;
                    }
                    // Note rr1 and cc1 are negative while rr2 and cc2 are positive
                    double x0P = x0 + scaleTerm * cP;  //Re position
                    double y0P = y0 - scaleTerm * rP;  //Im position

                    // Iteration loop using perturbation method, save delta values
                    double[] DxP = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
                    double[] DyP = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
                    double x, y, Dxtemp;
                    int i2 = 1;
                    double xP = x0P;
                    double yP = y0P;
                    double Dx0 = cP * scaleTerm;
                    double Dy0 = -rP * scaleTerm;
                    double cs = Math.Cos(theta);
                    double sn = Math.Sin(theta);
                    double Dx = Dx0 * cs - Dy0 * sn;  //delta1 = Y1 - X1
                    double Dy = Dx0 * sn + Dy0 * cs;
                    bool flagBreak = false;
                    DxP[0] = Dx;
                    DyP[0] = Dy;
                    for (int i = 1; i <= maxRefSeries - 1; i++) {
                        i2 = i;
                        x = xMP[i, 1];
                        y = yMP[i, 1];
                        xP = Dx + x;
                        yP = Dy + y;
                        if (xP * xP + yP * yP < glitchDetect[i, 1]) {
                            flagBreak = true;
                            break;
                        }
                        if (xP * xP + yP * yP > bailout) {
                            flagBreak = true;
                            break;
                        }
                        cs = Math.Cos(theta * (i + 1));
                        sn = Math.Sin(theta * (i + 1));
                        Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0 * cs - Dy0 * sn;
                        Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dx0 * sn + Dy0 * cs;
                        Dx = Dxtemp;
                        if (Dx * Dx + Dy * Dy > divThreshold) {
                            flagBreak = true;
                            break;
                        }
                        DxP[i] = Dx;
                        DyP[i] = Dy;
                    }  //next i
                    if (flagBreak == false) i2++;

                    // Compare delta values calculated using series approximation (DxS, DyS) to delta values calculated using
                    // perturbation theory (DxP, DyP) to determine how long the series approximation is valid
                    if (i2 < (10 + 1)) {
                        skipSeriesFlag = true;
                        break;
                    }

                    // Order 3 approximation
                    maxSeries = Math.Min(maxO3Series - 1, i2 - 1);
                    int k2 = 1;
                    double DxS, DyS;
                    flagBreak = false;
                    for (int k = 1; k <= maxSeries - 1; k++) {
                        k2 = k + 1;
                        DxS = A1r[k] * Dx0 + A2r[k] * Dy0 + B1r[k] * Dx0 * Dx0 + B2r[k] * Dy0 * Dy0 + B3r[k] * Dx0 * Dy0 + C1r[k] * Dx0 * Dx0 * Dx0 + C2r[k] * Dy0 * Dy0 * Dy0 + C3r[k] * Dx0 * Dx0 * Dy0 + C4r[k] * Dx0 * Dy0 * Dy0;
                        DyS = A1i[k] * Dx0 + A2i[k] * Dy0 + B1i[k] * Dx0 * Dx0 + B2i[k] * Dy0 * Dy0 + B3i[k] * Dx0 * Dy0 + C1i[k] * Dx0 * Dx0 * Dx0 + C2i[k] * Dy0 * Dy0 * Dy0 + C3i[k] * Dx0 * Dx0 * Dy0 + C4i[k] * Dx0 * Dy0 * Dy0;
                        if (DxP[k] == 0 || DyP[k] == 0) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DxP[k] - DxS) / DxP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DyP[k] - DyS) / DyP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                    }
                    if (flagBreak == true) k2 = k2 - 1;  //go back to when relative difference was below tolerance, this line is surprisingly important
                    if (j == 1) {
                        skipNumberO3 = k2;
                    }
                    else {
                        if (k2 < skipNumberO3) skipNumberO3 = k2;
                    }

                    // Order 5 approximation
                    maxSeries = Math.Min(maxO5Series - 1, i2 - 1);
                    int m = k2;  //purpose of 5th order approximation is to skip more iterations so we only need to check additional iterations
                    if (skipNumberO5 == -1) m = maxSeries + 1;  //avoid entering loop if k2 = m on a previous pass
                    flagBreak = false;
                    for (int k = m - 1; k <= maxSeries - 1; k++) {
                        k2 = k + 1;
                        DxS = A1r[k] * Dx0 + A2r[k] * Dy0 + B1r[k] * Dx0 * Dx0 + B2r[k] * Dy0 * Dy0 + B3r[k] * Dx0 * Dy0 + C1r[k] * Dx0 * Dx0 * Dx0 + C2r[k] * Dy0 * Dy0 * Dy0 + C3r[k] * Dx0 * Dx0 * Dy0 + C4r[k] * Dx0 * Dy0 * Dy0 + D1r[k] * Dx0 * Dx0 * Dx0 * Dx0 + D2r[k] * Dy0 * Dy0 * Dy0 * Dy0 + D3r[k] * Dx0 * Dx0 * Dx0 * Dy0 + D4r[k] * Dx0 * Dy0 * Dy0 * Dy0 + D5r[k] * Dx0 * Dx0 * Dy0 * Dy0 + E1r[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2r[k] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3r[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4r[k] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5r[k] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6r[k] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                        DyS = A1i[k] * Dx0 + A2i[k] * Dy0 + B1i[k] * Dx0 * Dx0 + B2i[k] * Dy0 * Dy0 + B3i[k] * Dx0 * Dy0 + C1i[k] * Dx0 * Dx0 * Dx0 + C2i[k] * Dy0 * Dy0 * Dy0 + C3i[k] * Dx0 * Dx0 * Dy0 + C4i[k] * Dx0 * Dy0 * Dy0 + D1i[k] * Dx0 * Dx0 * Dx0 * Dx0 + D2i[k] * Dy0 * Dy0 * Dy0 * Dy0 + D3i[k] * Dx0 * Dx0 * Dx0 * Dy0 + D4i[k] * Dx0 * Dy0 * Dy0 * Dy0 + D5i[k] * Dx0 * Dx0 * Dy0 * Dy0 + E1i[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2i[k] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3i[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4i[k] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5i[k] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6i[k] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                        if (DxP[k] == 0 || DyP[k] == 0) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DxP[k] - DxS) / DxP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DyP[k] - DyS) / DyP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                    }
                    if (k2 == m) k2 = 0;
                    if (flagBreak == true) k2 = k2 - 1;  //go back to when relative difference was below tolerance, this line is surprisingly important
                    if (j == 1) {
                        skipNumberO5 = k2;
                    }
                    else {
                        if (k2 < skipNumberO5) skipNumberO5 = k2;
                    }

                    // Check if it is worth continuing
                    if (Math.Max(skipNumberO3, skipNumberO5) < 5) {
                        skipSeriesFlag = true;
                        skipNumber = 0;
                        break;
                    }

                }  //next j

                // Determine final number of iterations to skip and what order of approximation to use
                // Order 5 calculations are more expensive so they are penalized slightly in the comparison
                if (skipSeriesFlag == false) {
                    if (skipNumberO3 > (skipNumberO5 - 10)) {
                        skipNumber = skipNumberO3;
                        useOrder3 = true;
                    }
                    else {  //use order 5
                        skipNumber = skipNumberO5;
                        useOrder3 = false;
                    }
                }
            }
            if (numPointsMP == 1) numSkip = skipNumber;  //store number of iterations skipped in initial pass for calculation info output

            // Calculate points about reference point orbit in double precision

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int rP = res * ri + rr1;       //mapping the parallel loop index to the row number relative to the reference point
                    double y0P = ziP - pixelStepP * ri;
                    double x0P = zrP - pixelStepP;
                    double x, y;
                    double xP = 0, xtempP, yP = 0;
                    double Dx = 0, Dxtemp, Dy = 0, Dx0, Dy0;
                    int iterationP = 0;
                    bool skipCalc, goToGlitch;
                    double iterationPD = 0;
                    double log_zn, nu;
                    double cs, sn;

                    for (int cP = cc1; cP <= cc2; cP = cP + res) {
                        skipCalc = false;
                        x0P = x0P + pixelStepP;
                        if ((cP == 0 && rP == 0) == false) {
                            if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] == prevMaxIter) || (refreshType == 2 && numPointsMP > 1)) {

                                // Quick pan code
                                if (refreshType == -1 && numPointsMP == 1) {
                                    skipCalc = quickPan(dr, dc, rRef + rP, cRef + cP, navSize);
                                }

                                // If correcting a flagged glitch, only correct location that is flagged
                                if (numPointsMP > 1) {
                                    if (prevGlitchFlag[rRef + rP, cRef + cP] == 0) skipCalc = true;
                                }

                                // Series approximation to skip iterations
                                if (skipCalc == false) {
                                    int m = skipNumber;
                                    int k = 1;
                                    bool tryLessFlag = false;
                                    bool goToTryLess = true;
                                    Dx0 = cP * scaleTerm;
                                    Dy0 = -rP * scaleTerm;
                                    if (skipSeriesFlag == false) {
                                        while (goToTryLess == true) {
                                            goToTryLess = false;
                                            if (useOrder3 == true) {
                                                Dx = A1r[m - 1] * Dx0 + A2r[m - 1] * Dy0 + B1r[m - 1] * Dx0 * Dx0 + B2r[m - 1] * Dy0 * Dy0 + B3r[m - 1] * Dx0 * Dy0 + C1r[m - 1] * Dx0 * Dx0 * Dx0 + C2r[m - 1] * Dy0 * Dy0 * Dy0 + C3r[m - 1] * Dx0 * Dx0 * Dy0 + C4r[m - 1] * Dx0 * Dy0 * Dy0;
                                                Dy = A1i[m - 1] * Dx0 + A2i[m - 1] * Dy0 + B1i[m - 1] * Dx0 * Dx0 + B2i[m - 1] * Dy0 * Dy0 + B3i[m - 1] * Dx0 * Dy0 + C1i[m - 1] * Dx0 * Dx0 * Dx0 + C2i[m - 1] * Dy0 * Dy0 * Dy0 + C3i[m - 1] * Dx0 * Dx0 * Dy0 + C4i[m - 1] * Dx0 * Dy0 * Dy0;
                                            }
                                            else {
                                                Dx = A1r[m - 1] * Dx0 + A2r[m - 1] * Dy0 + B1r[m - 1] * Dx0 * Dx0 + B2r[m - 1] * Dy0 * Dy0 + B3r[m - 1] * Dx0 * Dy0 + C1r[m - 1] * Dx0 * Dx0 * Dx0 + C2r[m - 1] * Dy0 * Dy0 * Dy0 + C3r[m - 1] * Dx0 * Dx0 * Dy0 + C4r[m - 1] * Dx0 * Dy0 * Dy0 + D1r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 + D2r[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 + D3r[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 + D4r[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 + D5r[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 + E1r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2r[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4r[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5r[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6r[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                                                Dy = A1i[m - 1] * Dx0 + A2i[m - 1] * Dy0 + B1i[m - 1] * Dx0 * Dx0 + B2i[m - 1] * Dy0 * Dy0 + B3i[m - 1] * Dx0 * Dy0 + C1i[m - 1] * Dx0 * Dx0 * Dx0 + C2i[m - 1] * Dy0 * Dy0 * Dy0 + C3i[m - 1] * Dx0 * Dx0 * Dy0 + C4i[m - 1] * Dx0 * Dy0 * Dy0 + D1i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 + D2i[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 + D3i[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 + D4i[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 + D5i[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 + E1i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2i[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4i[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5i[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6i[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                                            }
                                            if (Dx * Dx + Dy * Dy > divThreshold) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            xP = Dx + xMP[m, 1];
                                            yP = Dy + yMP[m, 1];
                                            if (xP * xP + yP * yP > bailout) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            if (goToTryLess == false) {
                                                k = m;  //skip first m-1 iterations
                                            }
                                            else {
                                                m = m / 2;
                                            }
                                        }  //Wend
                                    }

                                    // Iteration loop
                                    if (k == 1) {  //no series approximation used
                                        cs = Math.Cos(theta);
                                        sn = Math.Sin(theta);
                                        Dx = Dx0 * cs - Dy0 * sn;  //delta1 = Y1 - X1
                                        Dy = Dx0 * sn + Dy0 * cs;
                                    }

                                    goToGlitch = false;
                                    for (int j = k; j <= iteration; j++) {
                                        iterationP = j;
                                        x = xMP[j, 1];
                                        y = yMP[j, 1];
                                        xP = Dx + x;
                                        yP = Dy + y;
                                        if (xP * xP + yP * yP < glitchDetect[j, 1]) {
                                            glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                            Interlocked.Increment(ref cntFlagGlitch);
                                            if (flagDebugGlitch == true) iterationPD = maxIter;
                                            goToGlitch = true;
                                            break;
                                        }
                                        if (xP * xP + yP * yP > bailout) break;
                                        cs = Math.Cos(theta * (j + 1));
                                        sn = Math.Sin(theta * (j + 1));
                                        Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0 * cs - Dy0 * sn;
                                        Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dx0 * sn + Dy0 * cs;
                                        Dx = Dxtemp;
                                        if (Dx * Dx + Dy * Dy > divThreshold) break;
                                    }  //next j

                                    if (goToGlitch == false) {

                                        // Discrete coloring
                                        if (flagSmooth == false) {
                                            while (xP * xP + yP * yP < 4 && iterationP < maxIter) {
                                                iterationP++;
                                                cs = Math.Cos(theta * iterationP);
                                                sn = Math.Sin(theta * iterationP);
                                                xtempP = xP * xP - yP * yP + x0P * cs - y0P * sn;
                                                yP = 2 * xP * yP + x0P * sn + y0P * cs;
                                                xP = xtempP;
                                            }
                                            iterationPD = iterationP;
                                        }

                                        // Continuous coloring
                                        else {
                                            while (xP * xP + yP * yP < highBailout && iterationP < maxIter) {
                                                iterationP++;
                                                cs = Math.Cos(theta * iterationP);
                                                sn = Math.Sin(theta * iterationP);
                                                xtempP = xP * xP - yP * yP + x0P * cs - y0P * sn;
                                                yP = 2 * xP * yP + x0P * sn + y0P * cs;
                                                xP = xtempP;
                                            }

                                            // Additional calculations for continuous coloring
                                            if (iterationP < maxIter) {
                                                log_zn = Math.Log(xP * xP + yP * yP) / 2;
                                                nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                                iterationPD = iterationP + 1 - nu;
                                                iterationPD = Math.Round(iterationPD, 3);
                                            }
                                            else {
                                                iterationPD = iterationP;
                                            }
                                        }

                                    }

                                    // Check if too many additional iterations past the reference were performed
                                    if (maxPastRef != 0 && iterationP > maxPastRef) {
                                        glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                        Interlocked.Increment(ref cntFlagGlitch);
                                        if (flagDebugGlitch == true) iterationPD = maxIter;
                                    }

                                    // Store result to array
                                    if (res == 1) imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] = iterationPD;
                                    else reduceResolution(ref imageDataTemp, rRef + rP, cRef + cP, iterationPD, res);
                                }
                            }
                        }

                    }  //next c
                });  //next r
                if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
                else excelApp.StatusBar = string.Format("Correcting initial results ({2})... {0:0%}      DLL in use on {1} logical processors.{3}", (i + 1) / (double)n, numProcessors, numPointsMP, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
            copyArray(ref glitchFlag, glitchFlagTemp);
            if (cntFlagGlitch > 0) flagGlitch = true;
        }

        public static void tDbl(ref double[,] imageData, bool flagSmooth, int flagDllOption, double zr, double zi, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int r1, int c1, int navSize, int rSize, int cSize)
        {  // Beetle fractal, standard double precision
            double pixelStep = scaleTerm * res;
            int rUpperIndex = ((rSize - 1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double[,] imageDataTemp = imageData.Clone() as double[,];  //base 1 array, imageData is not allowed in Parallel.For loop because it is passed by reference

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int r = res * ri + 2;          //mapping the parallel loop index to the row number
                    double y0 = zi - pixelStep * ri;
                    double x0 = zr - pixelStep;
                    double x, xtemp, y;
                    int iteration;
                    double iterationD;
                    bool skipCalc;
                    double log_zn, nu;

                    for (int c = 2; c <= cSize + 1; c = c + res) {
                        skipCalc = false;
                        x0 = x0 + pixelStep;
                        if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[r - 1, c - 1] == prevMaxIter)) {

                            // Quick pan code
                            if (refreshType == -1) {
                                skipCalc = quickPan(r1, c1, r, c, navSize);
                            }

                            // Iteration loop
                            if (skipCalc == false) {
                                iteration = 1;
                                x = -x0;
                                y = -y0;

                                // Discrete coloring
                                if (flagSmooth == false) {
                                    while (x * x + y * y < 4 && iteration < maxIter) {
                                        iteration++;
                                        if (iteration % 2 == 1) {
                                            xtemp = x * x - y * y - x0;
                                            y = 2 * x * y - y0;
                                        }
                                        else {
                                            xtemp = x * x - y * y + x0;
                                            y = 2 * x * y + y0;
                                        }
                                        x = xtemp;
                                    }
                                    iterationD = iteration;
                                }

                                // Continuous coloring
                                else {
                                    while (x * x + y * y < highBailout && iteration < maxIter) {
                                        iteration++;
                                        if (iteration % 2 == 1) {
                                            xtemp = x * x - y * y - x0;
                                            y = 2 * x * y - y0;
                                        }
                                        else {
                                            xtemp = x * x - y * y + x0;
                                            y = 2 * x * y + y0;
                                        }
                                        x = xtemp;
                                    }

                                    // Additional calculations for continuous coloring
                                    if (iteration < maxIter) {
                                        log_zn = Math.Log(x * x + y * y) / 2;
                                        nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                        iterationD = iteration + 1 - nu;
                                        iterationD = Math.Round(iterationD, 3);
                                    }
                                    else {
                                        iterationD = iteration;
                                    }
                                }

                                // Store result to array
                                if (res == 1) imageDataTemp[r - 1, c - 1] = iterationD;
                                else reduceResolution(ref imageDataTemp, r, c, iterationD, res);

                            }
                        }
                    }  //next c
                });  //next r
                excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
        }

        public static void tPert(ref double[,] imageData, ref int[,] glitchFlag, ref bool flagGlitch, bool flagSmooth, int flagDllOption, double x0, double y0, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int dr, int dc, int navSize, int rr1, int rr2, int cc1, int cc2, int rRef, int cRef, int numPointsMP,
            double[,] xMP, double[,] yMP, double[,] glitchDetect, double divThreshold, bool flagDebugGlitch, int iteration, int maxPastRef, int[,] prevGlitchFlag)
        {  // Beetle fractal, perturbation methods
            double pixelStepP = scaleTerm * res;
            int rUpperIndex = ((rr2 - rr1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double zrP = x0 + scaleTerm * cc1;  //min real position
            double ziP = y0 - scaleTerm * rr1;  //max imaginary position
            int bailout = 4;
            if (flagSmooth == true) bailout = highBailout;

            // The following are used to avoid passing variables by reference in the Parallel.For loop, array bases are unchanged when using the Clone method
            double[,] imageDataTemp = imageData.Clone() as double[,];
            int[,] glitchFlagTemp = glitchFlag.Clone() as int[,];
            flagGlitch = false;  //this is set in the Parallel.For loop using the Interlocked.Increment Method on cntFlagGlitch and then checking for non-zero values
            int cntFlagGlitch = 0;

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);
            else excelApp.StatusBar = string.Format("Correcting initial results ({1})... 0%      DLL in use on {0} logical processors.{2}", numProcessors, numPointsMP, dllOptionString);

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int rP = res * ri + rr1;       //mapping the parallel loop index to the row number relative to the reference point
                    double y0P = ziP - pixelStepP * ri;
                    double x0P = zrP - pixelStepP;
                    double x, y;
                    double xP = 0, xtempP, yP = 0;
                    double Dx, Dxtemp, Dy, Dx0, Dy0;
                    int iterationP = 0;
                    double iterationPD = 0;
                    bool skipCalc, goToGlitch;
                    double log_zn, nu;

                    for (int cP = cc1; cP <= cc2; cP = cP + res) {
                        skipCalc = false;
                        x0P = x0P + pixelStepP;
                        if ((cP == 0 && rP == 0) == false) {
                            if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] == prevMaxIter) || (refreshType == 2 && numPointsMP > 1)) {

                                // Quick pan code
                                if (refreshType == -1 && numPointsMP == 1) {
                                    skipCalc = quickPan(dr, dc, rRef + rP, cRef + cP, navSize);
                                }

                                // If correcting a flagged glitch, only correct location that is flagged
                                if (numPointsMP > 1) {
                                    if (prevGlitchFlag[rRef + rP, cRef + cP] == 0) skipCalc = true;
                                }

                                // Iteration loop
                                if (skipCalc == false) {
                                    Dx0 = cP * scaleTerm;
                                    Dy0 = -rP * scaleTerm;
                                    Dx = -Dx0;  //delta1 = Y1 - X1
                                    Dy = -Dy0;
                                    goToGlitch = false;

                                    for (int j = 1; j <= iteration; j++) {
                                        iterationP = j;
                                        x = xMP[j, 1];
                                        y = yMP[j, 1];
                                        xP = Dx + x;
                                        yP = Dy + y;
                                        if (xP * xP + yP * yP < glitchDetect[j, 1]) {
                                            glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                            Interlocked.Increment(ref cntFlagGlitch);
                                            if (flagDebugGlitch == true) iterationPD = maxIter;
                                            goToGlitch = true;
                                            break;
                                        }
                                        if (xP * xP + yP * yP > bailout) break;
                                        if ((j + 1) % 2 == 1) {
                                            Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy - Dx0;
                                            Dy = 2 * (x * Dy + y * Dx + Dx * Dy) - Dy0;
                                        }
                                        else {
                                            Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0;
                                            Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dy0;
                                        }
                                        Dx = Dxtemp;
                                        if (Dx * Dx + Dy * Dy > divThreshold) break;
                                    }  //next j

                                    if (goToGlitch == false) {

                                        // Discrete coloring
                                        if (flagSmooth == false) {
                                            while (xP * xP + yP * yP < 4 && iterationP < maxIter) {
                                                iterationP++;
                                                if (iterationP % 2 == 1) {
                                                    xtempP = xP * xP - yP * yP - x0P;
                                                    yP = 2 * xP * yP - y0P;
                                                }
                                                else {
                                                    xtempP = xP * xP - yP * yP + x0P;
                                                    yP = 2 * xP * yP + y0P;
                                                }
                                                xP = xtempP;
                                            }
                                            iterationPD = iterationP;
                                        }

                                        // Continuous coloring
                                        else {
                                            while (xP * xP + yP * yP < highBailout && iterationP < maxIter) {
                                                iterationP++;
                                                if (iterationP % 2 == 1) {
                                                    xtempP = xP * xP - yP * yP - x0P;
                                                    yP = 2 * xP * yP - y0P;
                                                }
                                                else {
                                                    xtempP = xP * xP - yP * yP + x0P;
                                                    yP = 2 * xP * yP + y0P;
                                                }
                                                xP = xtempP;
                                            }

                                            // Additional calculations for continuous coloring
                                            if (iterationP < maxIter) {
                                                log_zn = Math.Log(xP * xP + yP * yP) / 2;
                                                nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                                iterationPD = iterationP + 1 - nu;
                                                iterationPD = Math.Round(iterationPD, 3);
                                            }
                                            else {
                                                iterationPD = iterationP;
                                            }
                                        }

                                    }

                                    // Check if too many additional iterations past the reference were performed
                                    if (maxPastRef != 0 && iterationP > maxPastRef) {
                                        glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                        Interlocked.Increment(ref cntFlagGlitch);
                                        if (flagDebugGlitch == true) iterationPD = maxIter;
                                    }

                                    // Store result to array
                                    if (res == 1) imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] = iterationPD;
                                    else reduceResolution(ref imageDataTemp, rRef + rP, cRef + cP, iterationPD, res);
                                }
                            }
                        }

                    }  //next c
                });  //next r
                if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
                else excelApp.StatusBar = string.Format("Correcting initial results ({2})... {0:0%}      DLL in use on {1} logical processors.{3}", (i + 1) / (double)n, numProcessors, numPointsMP, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
            copyArray(ref glitchFlag, glitchFlagTemp);
            if (cntFlagGlitch > 0) flagGlitch = true;
        }

        public static void tSeries3(ref double[,] imageData, ref int[,] glitchFlag, ref bool flagGlitch, ref int numSkip, bool flagSmooth, int flagDllOption, double x0, double y0, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int dr, int dc, int navSize, int rr1, int rr2, int cc1, int cc2, int rRef, int cRef, int numPointsMP, double seriesTolerance,
            double[,] xMP, double[,] yMP, double[,] glitchDetect, double divThreshold, bool flagDebugGlitch, int iteration, int maxPastRef, int[,] prevGlitchFlag, int prevGlitchInfo2)
        {  // Beetle fractal, perturbation with forced 3rd order series approximation
            double pixelStepP = scaleTerm * res;
            int rUpperIndex = ((rr2 - rr1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double zrP = x0 + scaleTerm * cc1;  //min real position
            double ziP = y0 - scaleTerm * rr1;  //max imaginary position
            bool skipSeriesFlag = false;
            bool flagOverflow = false;
            int maxRefSeries = 0;
            int skipNumber = 0;
            double[] A1r = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
            double[] A2r = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
            double[] A1i = new double[iteration];
            double[] A2i = new double[iteration];
            double[] B1r = new double[iteration];
            double[] B2r = new double[iteration];
            double[] B3r = new double[iteration];
            double[] B1i = new double[iteration];
            double[] B2i = new double[iteration];
            double[] B3i = new double[iteration];
            double[] C1r = new double[iteration];
            double[] C2r = new double[iteration];
            double[] C3r = new double[iteration];
            double[] C4r = new double[iteration];
            double[] C1i = new double[iteration];
            double[] C2i = new double[iteration];
            double[] C3i = new double[iteration];
            double[] C4i = new double[iteration];
            int bailout = 4;
            if (flagSmooth == true) bailout = highBailout;
            double cs2;

            // The following are used to avoid passing variables by reference in the Parallel.For loop, array bases are unchanged when using the Clone method
            double[,] imageDataTemp = imageData.Clone() as double[,];
            int[,] glitchFlagTemp = glitchFlag.Clone() as int[,];
            flagGlitch = false;  //this is set in the Parallel.For loop using the Interlocked.Increment Method on cntFlagGlitch and then checking for non-zero values
            int cntFlagGlitch = 0;

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);
            else excelApp.StatusBar = string.Format("Correcting initial results ({1})... 0%      DLL in use on {0} logical processors.{2}", numProcessors, numPointsMP, dllOptionString);

            // Calculate coefficients of series approximation up to a 3rd order approximation
            // When correcting glitches, series approximation is skipped if the largest glitch is small
            if (numPointsMP == 1) {
                if (iteration < 10) skipSeriesFlag = true;
            }
            else {
                if (iteration < 10 || prevGlitchInfo2 < 10) skipSeriesFlag = true;
            }
            if (skipSeriesFlag == false) {
                A1r[1] = -2 * xMP[1, 1] + 1;
                A2r[1] = 2 * yMP[1, 1];
                B1r[1] = 1;
                B2r[1] = -1;
                A1i[1] = -A2r[1];
                A2i[1] = A1r[1];
                B3i[1] = 2;
                for (int i = 2; i <= iteration - 1; i++) {
                    maxRefSeries = i;
                    if ((i + 1) % 2 == 1) {
                        cs2 = -1;
                    }
                    else {
                        cs2 = 1;
                    }
                    A1r[i] = 2 * (-A1i[i - 1] * yMP[i, 1] + A1r[i - 1] * xMP[i, 1]) + cs2;
                    A2r[i] = 2 * (-A2i[i - 1] * yMP[i, 1] + A2r[i - 1] * xMP[i, 1]);
                    B1r[i] = -A1i[i - 1] * A1i[i - 1] + A1r[i - 1] * A1r[i - 1] + 2 * (-B1i[i - 1] * yMP[i, 1] + B1r[i - 1] * xMP[i, 1]);
                    B2r[i] = -A2i[i - 1] * A2i[i - 1] + A2r[i - 1] * A2r[i - 1] + 2 * (-B2i[i - 1] * yMP[i, 1] + B2r[i - 1] * xMP[i, 1]);
                    B3r[i] = 2 * (-A1i[i - 1] * A2i[i - 1] + A1r[i - 1] * A2r[i - 1] - B3i[i - 1] * yMP[i, 1] + B3r[i - 1] * xMP[i, 1]);
                    C1r[i] = 2 * (-A1i[i - 1] * B1i[i - 1] + A1r[i - 1] * B1r[i - 1] - C1i[i - 1] * yMP[i, 1] + C1r[i - 1] * xMP[i, 1]);
                    C2r[i] = 2 * (-A2i[i - 1] * B2i[i - 1] + A2r[i - 1] * B2r[i - 1] - C2i[i - 1] * yMP[i, 1] + C2r[i - 1] * xMP[i, 1]);
                    C3r[i] = 2 * (-A1i[i - 1] * B3i[i - 1] + A1r[i - 1] * B3r[i - 1] - A2i[i - 1] * B1i[i - 1] + A2r[i - 1] * B1r[i - 1] - C3i[i - 1] * yMP[i, 1] + C3r[i - 1] * xMP[i, 1]);
                    C4r[i] = 2 * (-A1i[i - 1] * B2i[i - 1] + A1r[i - 1] * B2r[i - 1] - A2i[i - 1] * B3i[i - 1] + A2r[i - 1] * B3r[i - 1] - C4i[i - 1] * yMP[i, 1] + C4r[i - 1] * xMP[i, 1]);
                    A1i[i] = 2 * (A1i[i - 1] * xMP[i, 1] + A1r[i - 1] * yMP[i, 1]);
                    A2i[i] = 2 * (A2i[i - 1] * xMP[i, 1] + A2r[i - 1] * yMP[i, 1]) + cs2;
                    B1i[i] = 2 * (A1i[i - 1] * A1r[i - 1] + B1i[i - 1] * xMP[i, 1] + B1r[i - 1] * yMP[i, 1]);
                    B2i[i] = 2 * (A2i[i - 1] * A2r[i - 1] + B2i[i - 1] * xMP[i, 1] + B2r[i - 1] * yMP[i, 1]);
                    B3i[i] = 2 * (A1i[i - 1] * A2r[i - 1] + A1r[i - 1] * A2i[i - 1] + B3i[i - 1] * xMP[i, 1] + B3r[i - 1] * yMP[i, 1]);
                    C1i[i] = 2 * (A1i[i - 1] * B1r[i - 1] + A1r[i - 1] * B1i[i - 1] + C1i[i - 1] * xMP[i, 1] + C1r[i - 1] * yMP[i, 1]);
                    C2i[i] = 2 * (A2i[i - 1] * B2r[i - 1] + A2r[i - 1] * B2i[i - 1] + C2i[i - 1] * xMP[i, 1] + C2r[i - 1] * yMP[i, 1]);
                    C3i[i] = 2 * (A1i[i - 1] * B3r[i - 1] + A1r[i - 1] * B3i[i - 1] + A2i[i - 1] * B1r[i - 1] + A2r[i - 1] * B1i[i - 1] + C3i[i - 1] * xMP[i, 1] + C3r[i - 1] * yMP[i, 1]);
                    C4i[i] = 2 * (A1i[i - 1] * B2r[i - 1] + A1r[i - 1] * B2i[i - 1] + A2i[i - 1] * B3r[i - 1] + A2r[i - 1] * B3i[i - 1] + C4i[i - 1] * xMP[i, 1] + C4r[i - 1] * yMP[i, 1]);
                    // Check for overflow, unlike VBA an exception is not thrown when a double exceeds the maximum magnitude
                    if (Double.IsInfinity(C4i[i]) || Double.IsInfinity(C4r[i]) || Double.IsInfinity(C3i[i]) || Double.IsInfinity(C3r[i]) || Double.IsInfinity(C2i[i]) || Double.IsInfinity(C2r[i]) ||
                        Double.IsInfinity(C1i[i]) || Double.IsInfinity(C1r[i]) || Double.IsInfinity(B3i[i]) || Double.IsInfinity(B3r[i]) || Double.IsInfinity(B2i[i]) || Double.IsInfinity(B2r[i]) ||
                        Double.IsInfinity(B1i[i]) || Double.IsInfinity(B1r[i]) || Double.IsInfinity(A2i[i]) || Double.IsInfinity(A2r[i]) || Double.IsInfinity(A1i[i]) || Double.IsInfinity(A1r[i])) {
                        flagOverflow = true;
                        break;
                    }
                }  //next i

                // Catch overflow condition with flag
                if (flagOverflow == true) {
                    if (maxRefSeries < 10) skipSeriesFlag = true;
                    maxRefSeries = maxRefSeries - 1;
                }
            }

            // Determine how many iterations to skip
            if (skipSeriesFlag == false) {
                for (int j = 1; j <= 4; j++) {  //4 corner pixels of range
                    int rP = 0, cP = 0;
                    if (j == 1) {  //top left, min Re & max Im
                        rP = rr1;
                        cP = cc1;
                    }
                    else if (j == 2) {  //top right, max Re & max Im
                        rP = rr1;
                        cP = cc2;
                    }
                    else if (j == 3) {  //bottom left, min Re & min Im
                        rP = rr2;
                        cP = cc1;
                    }
                    else if (j == 4) {  //bottom right, max Re & min Im
                        rP = rr2;
                        cP = cc2;
                    }
                    // Note rr1 and cc1 are negative while rr2 and cc2 are positive
                    double x0P = x0 + scaleTerm * cP;  //Re position
                    double y0P = y0 - scaleTerm * rP;  //Im position

                    // Iteration loop using perturbation method, save delta values
                    double[] DxP = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
                    double[] DyP = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
                    double x, y, Dxtemp;
                    int i2 = 1;
                    double xP = x0P;
                    double yP = y0P;
                    double Dx0 = cP * scaleTerm;
                    double Dy0 = -rP * scaleTerm;
                    double Dx = -Dx0;  //delta1 = Y1 - X1
                    double Dy = -Dy0;
                    bool flagBreak = false;
                    DxP[0] = Dx;
                    DyP[0] = Dy;
                    for (int i = 1; i <= maxRefSeries - 1; i++) {
                        i2 = i;
                        x = xMP[i, 1];
                        y = yMP[i, 1];
                        xP = Dx + x;
                        yP = Dy + y;
                        if (xP * xP + yP * yP < glitchDetect[i, 1]) {
                            flagBreak = true;
                            break;
                        }
                        if (xP * xP + yP * yP > bailout) {
                            flagBreak = true;
                            break;
                        }
                        if ((i + 1) % 2 == 1) {
                            Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy - Dx0;
                            Dy = 2 * (x * Dy + y * Dx + Dx * Dy) - Dy0;
                        }
                        else {
                            Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0;
                            Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dy0;
                        }
                        Dx = Dxtemp;
                        if (Dx * Dx + Dy * Dy > divThreshold) {
                            flagBreak = true;
                            break;
                        }
                        DxP[i] = Dx;
                        DyP[i] = Dy;
                    }  //next i
                    if (flagBreak == false) i2++;

                    // Compare delta values calculated using series approximation (DxS, DyS) to delta values calculated using
                    // perturbation theory (DxP, DyP) to determine how long the series approximation is valid
                    if (i2 < (10 + 1)) {
                        skipSeriesFlag = true;
                        break;
                    }
                    int k2 = 1;
                    double DxS, DyS;
                    flagBreak = false;
                    for (int k = 1; k <= i2 - 2; k++) {
                        k2 = k + 1;
                        DxS = A1r[k] * Dx0 + A2r[k] * Dy0 + B1r[k] * Dx0 * Dx0 + B2r[k] * Dy0 * Dy0 + B3r[k] * Dx0 * Dy0 + C1r[k] * Dx0 * Dx0 * Dx0 + C2r[k] * Dy0 * Dy0 * Dy0 + C3r[k] * Dx0 * Dx0 * Dy0 + C4r[k] * Dx0 * Dy0 * Dy0;
                        DyS = A1i[k] * Dx0 + A2i[k] * Dy0 + B1i[k] * Dx0 * Dx0 + B2i[k] * Dy0 * Dy0 + B3i[k] * Dx0 * Dy0 + C1i[k] * Dx0 * Dx0 * Dx0 + C2i[k] * Dy0 * Dy0 * Dy0 + C3i[k] * Dx0 * Dx0 * Dy0 + C4i[k] * Dx0 * Dy0 * Dy0;
                        if (DxP[k] == 0 || DyP[k] == 0) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DxP[k] - DxS) / DxP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DyP[k] - DyS) / DyP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                    }
                    if (flagBreak == true) k2 = k2 - 1;  //go back to when relative difference was below tolerance, this line is surprisingly important
                    if (k2 < 5) {
                        skipSeriesFlag = true;
                        skipNumber = 0;
                        break;
                    }
                    if (j == 1) {
                        skipNumber = k2;
                    }
                    else {
                        if (k2 < skipNumber) skipNumber = k2;
                    }
                }  //next j
            }
            if (numPointsMP == 1) numSkip = skipNumber;  //store number of iterations skipped in initial pass for calculation info output

            // Calculate points about reference point orbit in double precision

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int rP = res * ri + rr1;       //mapping the parallel loop index to the row number relative to the reference point
                    double y0P = ziP - pixelStepP * ri;
                    double x0P = zrP - pixelStepP;
                    double x, y;
                    double xP = 0, xtempP, yP = 0;
                    double Dx = 0, Dxtemp, Dy = 0, Dx0, Dy0;
                    int iterationP = 0;
                    bool skipCalc, goToGlitch;
                    double iterationPD = 0;
                    double log_zn, nu;

                    for (int cP = cc1; cP <= cc2; cP = cP + res) {
                        skipCalc = false;
                        x0P = x0P + pixelStepP;
                        if ((cP == 0 && rP == 0) == false) {
                            if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] == prevMaxIter) || (refreshType == 2 && numPointsMP > 1)) {

                                // Quick pan code
                                if (refreshType == -1 && numPointsMP == 1) {
                                    skipCalc = quickPan(dr, dc, rRef + rP, cRef + cP, navSize);
                                }

                                // If correcting a flagged glitch, only correct location that is flagged
                                if (numPointsMP > 1) {
                                    if (prevGlitchFlag[rRef + rP, cRef + cP] == 0) skipCalc = true;
                                }

                                // Series approximation to skip iterations
                                if (skipCalc == false) {
                                    int m = skipNumber;
                                    int k = 1;
                                    bool tryLessFlag = false;
                                    bool goToTryLess = true;
                                    Dx0 = cP * scaleTerm;
                                    Dy0 = -rP * scaleTerm;
                                    if (skipSeriesFlag == false) {
                                        while (goToTryLess == true) {
                                            goToTryLess = false;
                                            Dx = A1r[m - 1] * Dx0 + A2r[m - 1] * Dy0 + B1r[m - 1] * Dx0 * Dx0 + B2r[m - 1] * Dy0 * Dy0 + B3r[m - 1] * Dx0 * Dy0 + C1r[m - 1] * Dx0 * Dx0 * Dx0 + C2r[m - 1] * Dy0 * Dy0 * Dy0 + C3r[m - 1] * Dx0 * Dx0 * Dy0 + C4r[m - 1] * Dx0 * Dy0 * Dy0;
                                            Dy = A1i[m - 1] * Dx0 + A2i[m - 1] * Dy0 + B1i[m - 1] * Dx0 * Dx0 + B2i[m - 1] * Dy0 * Dy0 + B3i[m - 1] * Dx0 * Dy0 + C1i[m - 1] * Dx0 * Dx0 * Dx0 + C2i[m - 1] * Dy0 * Dy0 * Dy0 + C3i[m - 1] * Dx0 * Dx0 * Dy0 + C4i[m - 1] * Dx0 * Dy0 * Dy0;
                                            if (Dx * Dx + Dy * Dy > divThreshold) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            xP = Dx + xMP[m, 1];
                                            yP = Dy + yMP[m, 1];
                                            if (xP * xP + yP * yP > bailout) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            if (goToTryLess == false) {
                                                k = m;  //skip first m-1 iterations
                                            }
                                            else {
                                                m = m / 2;
                                            }
                                        }  //Wend
                                    }

                                    // Iteration loop
                                    if (k == 1) {  //no series approximation used
                                        Dx = -Dx0;  //delta1 = Y1 - X1
                                        Dy = -Dy0;
                                    }

                                    goToGlitch = false;
                                    for (int j = k; j <= iteration; j++) {
                                        iterationP = j;
                                        x = xMP[j, 1];
                                        y = yMP[j, 1];
                                        xP = Dx + x;
                                        yP = Dy + y;
                                        if (xP * xP + yP * yP < glitchDetect[j, 1]) {
                                            glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                            Interlocked.Increment(ref cntFlagGlitch);
                                            if (flagDebugGlitch == true) iterationPD = maxIter;
                                            goToGlitch = true;
                                            break;
                                        }
                                        if (xP * xP + yP * yP > bailout) break;
                                        if ((j + 1) % 2 == 1) {
                                            Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy - Dx0;
                                            Dy = 2 * (x * Dy + y * Dx + Dx * Dy) - Dy0;
                                        }
                                        else {
                                            Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0;
                                            Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dy0;
                                        }
                                        Dx = Dxtemp;
                                        if (Dx * Dx + Dy * Dy > divThreshold) break;
                                    }  //next j

                                    if (goToGlitch == false) {

                                        // Discrete coloring
                                        if (flagSmooth == false) {
                                            while (xP * xP + yP * yP < 4 && iterationP < maxIter) {
                                                iterationP++;
                                                if (iterationP % 2 == 1) {
                                                    xtempP = xP * xP - yP * yP - x0P;
                                                    yP = 2 * xP * yP - y0P;
                                                }
                                                else {
                                                    xtempP = xP * xP - yP * yP + x0P;
                                                    yP = 2 * xP * yP + y0P;
                                                }
                                                xP = xtempP;
                                            }
                                            iterationPD = iterationP;
                                        }

                                        // Continuous coloring
                                        else {
                                            while (xP * xP + yP * yP < highBailout && iterationP < maxIter) {
                                                iterationP++;
                                                if (iterationP % 2 == 1) {
                                                    xtempP = xP * xP - yP * yP - x0P;
                                                    yP = 2 * xP * yP - y0P;
                                                }
                                                else {
                                                    xtempP = xP * xP - yP * yP + x0P;
                                                    yP = 2 * xP * yP + y0P;
                                                }
                                                xP = xtempP;
                                            }

                                            // Additional calculations for continuous coloring
                                            if (iterationP < maxIter) {
                                                log_zn = Math.Log(xP * xP + yP * yP) / 2;
                                                nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                                iterationPD = iterationP + 1 - nu;
                                                iterationPD = Math.Round(iterationPD, 3);
                                            }
                                            else {
                                                iterationPD = iterationP;
                                            }
                                        }

                                    }

                                    // Check if too many additional iterations past the reference were performed
                                    if (maxPastRef != 0 && iterationP > maxPastRef) {
                                        glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                        Interlocked.Increment(ref cntFlagGlitch);
                                        if (flagDebugGlitch == true) iterationPD = maxIter;
                                    }

                                    // Store result to array
                                    if (res == 1) imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] = iterationPD;
                                    else reduceResolution(ref imageDataTemp, rRef + rP, cRef + cP, iterationPD, res);
                                }
                            }
                        }

                    }  //next c
                });  //next r
                if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
                else excelApp.StatusBar = string.Format("Correcting initial results ({2})... {0:0%}      DLL in use on {1} logical processors.{3}", (i + 1) / (double)n, numProcessors, numPointsMP, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
            copyArray(ref glitchFlag, glitchFlagTemp);
            if (cntFlagGlitch > 0) flagGlitch = true;
        }

        public static void tSeries5(ref double[,] imageData, ref int[,] glitchFlag, ref bool flagGlitch, ref int numSkip, bool flagSmooth, int flagDllOption, double x0, double y0, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int dr, int dc, int navSize, int rr1, int rr2, int cc1, int cc2, int rRef, int cRef, int numPointsMP, double seriesTolerance,
            double[,] xMP, double[,] yMP, double[,] glitchDetect, double divThreshold, bool flagDebugGlitch, int iteration, int maxPastRef, int[,] prevGlitchFlag, int prevGlitchInfo2)
        {  // Beetle fractal, perturbation with forced 5th order series approximation
            double pixelStepP = scaleTerm * res;
            int rUpperIndex = ((rr2 - rr1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double zrP = x0 + scaleTerm * cc1;  //min real position
            double ziP = y0 - scaleTerm * rr1;  //max imaginary position
            bool skipSeriesFlag = false;
            bool flagOverflow = false;
            int maxRefSeries = 0;
            int skipNumber = 0;
            double[] A1r = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
            double[] A2r = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
            double[] A1i = new double[iteration];
            double[] A2i = new double[iteration];
            double[] B1r = new double[iteration];
            double[] B2r = new double[iteration];
            double[] B3r = new double[iteration];
            double[] B1i = new double[iteration];
            double[] B2i = new double[iteration];
            double[] B3i = new double[iteration];
            double[] C1r = new double[iteration];
            double[] C2r = new double[iteration];
            double[] C3r = new double[iteration];
            double[] C4r = new double[iteration];
            double[] C1i = new double[iteration];
            double[] C2i = new double[iteration];
            double[] C3i = new double[iteration];
            double[] C4i = new double[iteration];
            double[] D1r = new double[iteration];
            double[] D2r = new double[iteration];
            double[] D3r = new double[iteration];
            double[] D4r = new double[iteration];
            double[] D5r = new double[iteration];
            double[] D1i = new double[iteration];
            double[] D2i = new double[iteration];
            double[] D3i = new double[iteration];
            double[] D4i = new double[iteration];
            double[] D5i = new double[iteration];
            double[] E1r = new double[iteration];
            double[] E2r = new double[iteration];
            double[] E3r = new double[iteration];
            double[] E4r = new double[iteration];
            double[] E5r = new double[iteration];
            double[] E6r = new double[iteration];
            double[] E1i = new double[iteration];
            double[] E2i = new double[iteration];
            double[] E3i = new double[iteration];
            double[] E4i = new double[iteration];
            double[] E5i = new double[iteration];
            double[] E6i = new double[iteration];
            int bailout = 4;
            if (flagSmooth == true) bailout = highBailout;
            double cs2;

            // The following are used to avoid passing variables by reference in the Parallel.For loop, array bases are unchanged when using the Clone method
            double[,] imageDataTemp = imageData.Clone() as double[,];
            int[,] glitchFlagTemp = glitchFlag.Clone() as int[,];
            flagGlitch = false;  //this is set in the Parallel.For loop using the Interlocked.Increment Method on cntFlagGlitch and then checking for non-zero values
            int cntFlagGlitch = 0;

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);
            else excelApp.StatusBar = string.Format("Correcting initial results ({1})... 0%      DLL in use on {0} logical processors.{2}", numProcessors, numPointsMP, dllOptionString);

            // Calculate coefficients of series approximation up to a 5th order approximation
            // When correcting glitches, series approximation is skipped if the largest glitch is small
            if (numPointsMP == 1) {
                if (iteration < 10) skipSeriesFlag = true;
            }
            else {
                if (iteration < 10 || prevGlitchInfo2 < 10) skipSeriesFlag = true;
            }
            if (skipSeriesFlag == false) {
                A1r[1] = -2 * xMP[1, 1] + 1;
                A2r[1] = 2 * yMP[1, 1];
                B1r[1] = 1;
                B2r[1] = -1;
                A1i[1] = -A2r[1];
                A2i[1] = A1r[1];
                B3i[1] = 2;
                for (int i = 2; i <= iteration - 1; i++) {
                    maxRefSeries = i;
                    if ((i + 1) % 2 == 1) {
                        cs2 = -1;
                    }
                    else {
                        cs2 = 1;
                    }
                    A1r[i] = 2 * (-A1i[i - 1] * yMP[i, 1] + A1r[i - 1] * xMP[i, 1]) + cs2;
                    A2r[i] = 2 * (-A2i[i - 1] * yMP[i, 1] + A2r[i - 1] * xMP[i, 1]);
                    B1r[i] = -A1i[i - 1] * A1i[i - 1] + A1r[i - 1] * A1r[i - 1] + 2 * (-B1i[i - 1] * yMP[i, 1] + B1r[i - 1] * xMP[i, 1]);
                    B2r[i] = -A2i[i - 1] * A2i[i - 1] + A2r[i - 1] * A2r[i - 1] + 2 * (-B2i[i - 1] * yMP[i, 1] + B2r[i - 1] * xMP[i, 1]);
                    B3r[i] = 2 * (-A1i[i - 1] * A2i[i - 1] + A1r[i - 1] * A2r[i - 1] - B3i[i - 1] * yMP[i, 1] + B3r[i - 1] * xMP[i, 1]);
                    C1r[i] = 2 * (-A1i[i - 1] * B1i[i - 1] + A1r[i - 1] * B1r[i - 1] - C1i[i - 1] * yMP[i, 1] + C1r[i - 1] * xMP[i, 1]);
                    C2r[i] = 2 * (-A2i[i - 1] * B2i[i - 1] + A2r[i - 1] * B2r[i - 1] - C2i[i - 1] * yMP[i, 1] + C2r[i - 1] * xMP[i, 1]);
                    C3r[i] = 2 * (-A1i[i - 1] * B3i[i - 1] + A1r[i - 1] * B3r[i - 1] - A2i[i - 1] * B1i[i - 1] + A2r[i - 1] * B1r[i - 1] - C3i[i - 1] * yMP[i, 1] + C3r[i - 1] * xMP[i, 1]);
                    C4r[i] = 2 * (-A1i[i - 1] * B2i[i - 1] + A1r[i - 1] * B2r[i - 1] - A2i[i - 1] * B3i[i - 1] + A2r[i - 1] * B3r[i - 1] - C4i[i - 1] * yMP[i, 1] + C4r[i - 1] * xMP[i, 1]);
                    A1i[i] = 2 * (A1i[i - 1] * xMP[i, 1] + A1r[i - 1] * yMP[i, 1]);
                    A2i[i] = 2 * (A2i[i - 1] * xMP[i, 1] + A2r[i - 1] * yMP[i, 1]) + cs2;
                    B1i[i] = 2 * (A1i[i - 1] * A1r[i - 1] + B1i[i - 1] * xMP[i, 1] + B1r[i - 1] * yMP[i, 1]);
                    B2i[i] = 2 * (A2i[i - 1] * A2r[i - 1] + B2i[i - 1] * xMP[i, 1] + B2r[i - 1] * yMP[i, 1]);
                    B3i[i] = 2 * (A1i[i - 1] * A2r[i - 1] + A1r[i - 1] * A2i[i - 1] + B3i[i - 1] * xMP[i, 1] + B3r[i - 1] * yMP[i, 1]);
                    C1i[i] = 2 * (A1i[i - 1] * B1r[i - 1] + A1r[i - 1] * B1i[i - 1] + C1i[i - 1] * xMP[i, 1] + C1r[i - 1] * yMP[i, 1]);
                    C2i[i] = 2 * (A2i[i - 1] * B2r[i - 1] + A2r[i - 1] * B2i[i - 1] + C2i[i - 1] * xMP[i, 1] + C2r[i - 1] * yMP[i, 1]);
                    C3i[i] = 2 * (A1i[i - 1] * B3r[i - 1] + A1r[i - 1] * B3i[i - 1] + A2i[i - 1] * B1r[i - 1] + A2r[i - 1] * B1i[i - 1] + C3i[i - 1] * xMP[i, 1] + C3r[i - 1] * yMP[i, 1]);
                    C4i[i] = 2 * (A1i[i - 1] * B2r[i - 1] + A1r[i - 1] * B2i[i - 1] + A2i[i - 1] * B3r[i - 1] + A2r[i - 1] * B3i[i - 1] + C4i[i - 1] * xMP[i, 1] + C4r[i - 1] * yMP[i, 1]);
                    D1r[i] = -B1i[i - 1] * B1i[i - 1] + B1r[i - 1] * B1r[i - 1] + 2 * (-A1i[i - 1] * C1i[i - 1] + A1r[i - 1] * C1r[i - 1] - D1i[i - 1] * yMP[i, 1] + D1r[i - 1] * xMP[i, 1]);
                    D2r[i] = -B2i[i - 1] * B2i[i - 1] + B2r[i - 1] * B2r[i - 1] + 2 * (-A2i[i - 1] * C2i[i - 1] + A2r[i - 1] * C2r[i - 1] - D2i[i - 1] * yMP[i, 1] + D2r[i - 1] * xMP[i, 1]);
                    D3r[i] = 2 * (-A1i[i - 1] * C3i[i - 1] + A1r[i - 1] * C3r[i - 1] - A2i[i - 1] * C1i[i - 1] + A2r[i - 1] * C1r[i - 1] - B1i[i - 1] * B3i[i - 1] + B1r[i - 1] * B3r[i - 1] - D3i[i - 1] * yMP[i, 1] + D3r[i - 1] * xMP[i, 1]);
                    D4r[i] = 2 * (-A1i[i - 1] * C2i[i - 1] + A1r[i - 1] * C2r[i - 1] - A2i[i - 1] * C4i[i - 1] + A2r[i - 1] * C4r[i - 1] - B2i[i - 1] * B3i[i - 1] + B2r[i - 1] * B3r[i - 1] - D4i[i - 1] * yMP[i, 1] + D4r[i - 1] * xMP[i, 1]);
                    D5r[i] = -B3i[i - 1] * B3i[i - 1] + B3r[i - 1] * B3r[i - 1] + 2 * (-A1i[i - 1] * C4i[i - 1] + A1r[i - 1] * C4r[i - 1] - A2i[i - 1] * C3i[i - 1] + A2r[i - 1] * C3r[i - 1] - B1i[i - 1] * B2i[i - 1] + B1r[i - 1] * B2r[i - 1] - D5i[i - 1] * yMP[i, 1] + D5r[i - 1] * xMP[i, 1]);
                    E1r[i] = 2 * (-A1i[i - 1] * D1i[i - 1] + A1r[i - 1] * D1r[i - 1] - B1i[i - 1] * C1i[i - 1] + B1r[i - 1] * C1r[i - 1] - E1i[i - 1] * yMP[i, 1] + E1r[i - 1] * xMP[i, 1]);
                    E2r[i] = 2 * (-A2i[i - 1] * D2i[i - 1] + A2r[i - 1] * D2r[i - 1] - B2i[i - 1] * C2i[i - 1] + B2r[i - 1] * C2r[i - 1] - E2i[i - 1] * yMP[i, 1] + E2r[i - 1] * xMP[i, 1]);
                    E3r[i] = 2 * (-A1i[i - 1] * D3i[i - 1] + A1r[i - 1] * D3r[i - 1] - A2i[i - 1] * D1i[i - 1] + A2r[i - 1] * D1r[i - 1] - B1i[i - 1] * C3i[i - 1] + B1r[i - 1] * C3r[i - 1] - B3i[i - 1] * C1i[i - 1] + B3r[i - 1] * C1r[i - 1] - E3i[i - 1] * yMP[i, 1] + E3r[i - 1] * xMP[i, 1]);
                    E4r[i] = 2 * (-A1i[i - 1] * D2i[i - 1] + A1r[i - 1] * D2r[i - 1] - A2i[i - 1] * D4i[i - 1] + A2r[i - 1] * D4r[i - 1] - B2i[i - 1] * C4i[i - 1] + B2r[i - 1] * C4r[i - 1] - B3i[i - 1] * C2i[i - 1] + B3r[i - 1] * C2r[i - 1] - E4i[i - 1] * yMP[i, 1] + E4r[i - 1] * xMP[i, 1]);
                    E5r[i] = 2 * (-A1i[i - 1] * D5i[i - 1] + A1r[i - 1] * D5r[i - 1] - A2i[i - 1] * D3i[i - 1] + A2r[i - 1] * D3r[i - 1] - B1i[i - 1] * C4i[i - 1] + B1r[i - 1] * C4r[i - 1] - B2i[i - 1] * C1i[i - 1] + B2r[i - 1] * C1r[i - 1] - B3i[i - 1] * C3i[i - 1] + B3r[i - 1] * C3r[i - 1] - E5i[i - 1] * yMP[i, 1] + E5r[i - 1] * xMP[i, 1]);
                    E6r[i] = 2 * (-A1i[i - 1] * D4i[i - 1] + A1r[i - 1] * D4r[i - 1] - A2i[i - 1] * D5i[i - 1] + A2r[i - 1] * D5r[i - 1] - B1i[i - 1] * C2i[i - 1] + B1r[i - 1] * C2r[i - 1] - B2i[i - 1] * C3i[i - 1] + B2r[i - 1] * C3r[i - 1] - B3i[i - 1] * C4i[i - 1] + B3r[i - 1] * C4r[i - 1] - E6i[i - 1] * yMP[i, 1] + E6r[i - 1] * xMP[i, 1]);
                    D1i[i] = 2 * (A1i[i - 1] * C1r[i - 1] + A1r[i - 1] * C1i[i - 1] + B1i[i - 1] * B1r[i - 1] + D1i[i - 1] * xMP[i, 1] + D1r[i - 1] * yMP[i, 1]);
                    D2i[i] = 2 * (A2i[i - 1] * C2r[i - 1] + A2r[i - 1] * C2i[i - 1] + B2i[i - 1] * B2r[i - 1] + D2i[i - 1] * xMP[i, 1] + D2r[i - 1] * yMP[i, 1]);
                    D3i[i] = 2 * (A1i[i - 1] * C3r[i - 1] + A1r[i - 1] * C3i[i - 1] + A2i[i - 1] * C1r[i - 1] + A2r[i - 1] * C1i[i - 1] + B1i[i - 1] * B3r[i - 1] + B1r[i - 1] * B3i[i - 1] + D3i[i - 1] * xMP[i, 1] + D3r[i - 1] * yMP[i, 1]);
                    D4i[i] = 2 * (A1i[i - 1] * C2r[i - 1] + A1r[i - 1] * C2i[i - 1] + A2i[i - 1] * C4r[i - 1] + A2r[i - 1] * C4i[i - 1] + B2i[i - 1] * B3r[i - 1] + B2r[i - 1] * B3i[i - 1] + D4i[i - 1] * xMP[i, 1] + D4r[i - 1] * yMP[i, 1]);
                    D5i[i] = 2 * (A1i[i - 1] * C4r[i - 1] + A1r[i - 1] * C4i[i - 1] + A2i[i - 1] * C3r[i - 1] + A2r[i - 1] * C3i[i - 1] + B1i[i - 1] * B2r[i - 1] + B1r[i - 1] * B2i[i - 1] + B3i[i - 1] * B3r[i - 1] + D5i[i - 1] * xMP[i, 1] + D5r[i - 1] * yMP[i, 1]);
                    E1i[i] = 2 * (A1i[i - 1] * D1r[i - 1] + A1r[i - 1] * D1i[i - 1] + B1i[i - 1] * C1r[i - 1] + B1r[i - 1] * C1i[i - 1] + E1i[i - 1] * xMP[i, 1] + E1r[i - 1] * yMP[i, 1]);
                    E2i[i] = 2 * (A2i[i - 1] * D2r[i - 1] + A2r[i - 1] * D2i[i - 1] + B2i[i - 1] * C2r[i - 1] + B2r[i - 1] * C2i[i - 1] + E2i[i - 1] * xMP[i, 1] + E2r[i - 1] * yMP[i, 1]);
                    E3i[i] = 2 * (A1i[i - 1] * D3r[i - 1] + A1r[i - 1] * D3i[i - 1] + A2i[i - 1] * D1r[i - 1] + A2r[i - 1] * D1i[i - 1] + B1i[i - 1] * C3r[i - 1] + B1r[i - 1] * C3i[i - 1] + B3i[i - 1] * C1r[i - 1] + B3r[i - 1] * C1i[i - 1] + E3i[i - 1] * xMP[i, 1] + E3r[i - 1] * yMP[i, 1]);
                    E4i[i] = 2 * (A1i[i - 1] * D2r[i - 1] + A1r[i - 1] * D2i[i - 1] + A2i[i - 1] * D4r[i - 1] + A2r[i - 1] * D4i[i - 1] + B2i[i - 1] * C4r[i - 1] + B2r[i - 1] * C4i[i - 1] + B3i[i - 1] * C2r[i - 1] + B3r[i - 1] * C2i[i - 1] + E4i[i - 1] * xMP[i, 1] + E4r[i - 1] * yMP[i, 1]);
                    E5i[i] = 2 * (A1i[i - 1] * D5r[i - 1] + A1r[i - 1] * D5i[i - 1] + A2i[i - 1] * D3r[i - 1] + A2r[i - 1] * D3i[i - 1] + B1i[i - 1] * C4r[i - 1] + B1r[i - 1] * C4i[i - 1] + B2i[i - 1] * C1r[i - 1] + B2r[i - 1] * C1i[i - 1] + B3i[i - 1] * C3r[i - 1] + B3r[i - 1] * C3i[i - 1] + E5i[i - 1] * xMP[i, 1] + E5r[i - 1] * yMP[i, 1]);
                    E6i[i] = 2 * (A1i[i - 1] * D4r[i - 1] + A1r[i - 1] * D4i[i - 1] + A2i[i - 1] * D5r[i - 1] + A2r[i - 1] * D5i[i - 1] + B1i[i - 1] * C2r[i - 1] + B1r[i - 1] * C2i[i - 1] + B2i[i - 1] * C3r[i - 1] + B2r[i - 1] * C3i[i - 1] + B3i[i - 1] * C4r[i - 1] + B3r[i - 1] * C4i[i - 1] + E6i[i - 1] * xMP[i, 1] + E6r[i - 1] * yMP[i, 1]);
                    // Check for overflow, unlike VBA an exception is not thrown when a double exceeds the maximum magnitude
                    if (Double.IsInfinity(E6i[i]) || Double.IsInfinity(E6r[i]) || Double.IsInfinity(E5i[i]) || Double.IsInfinity(E5r[i]) || Double.IsInfinity(E4i[i]) || Double.IsInfinity(E4r[i]) ||
                        Double.IsInfinity(E3i[i]) || Double.IsInfinity(E3r[i]) || Double.IsInfinity(E2i[i]) || Double.IsInfinity(E2r[i]) || Double.IsInfinity(E1i[i]) || Double.IsInfinity(E1r[i]) ||
                        Double.IsInfinity(D5i[i]) || Double.IsInfinity(D5r[i]) || Double.IsInfinity(D4i[i]) || Double.IsInfinity(D4r[i]) || Double.IsInfinity(D3i[i]) || Double.IsInfinity(D3r[i]) ||
                        Double.IsInfinity(D2i[i]) || Double.IsInfinity(D2r[i]) || Double.IsInfinity(D1i[i]) || Double.IsInfinity(D1r[i]) ||
                        Double.IsInfinity(C4i[i]) || Double.IsInfinity(C4r[i]) || Double.IsInfinity(C3i[i]) || Double.IsInfinity(C3r[i]) || Double.IsInfinity(C2i[i]) || Double.IsInfinity(C2r[i]) ||
                        Double.IsInfinity(C1i[i]) || Double.IsInfinity(C1r[i]) || Double.IsInfinity(B3i[i]) || Double.IsInfinity(B3r[i]) || Double.IsInfinity(B2i[i]) || Double.IsInfinity(B2r[i]) ||
                        Double.IsInfinity(B1i[i]) || Double.IsInfinity(B1r[i]) || Double.IsInfinity(A2i[i]) || Double.IsInfinity(A2r[i]) || Double.IsInfinity(A1i[i]) || Double.IsInfinity(A1r[i])) {
                        flagOverflow = true;
                        break;
                    }
                }  //next i

                // Catch overflow condition with flag
                if (flagOverflow == true) {
                    if (maxRefSeries < 10) skipSeriesFlag = true;
                    maxRefSeries = maxRefSeries - 1;
                }
            }

            // Determine how many iterations to skip
            if (skipSeriesFlag == false) {
                for (int j = 1; j <= 4; j++) {  //4 corner pixels of range
                    int rP = 0, cP = 0;
                    if (j == 1) {  //top left, min Re & max Im
                        rP = rr1;
                        cP = cc1;
                    }
                    else if (j == 2) {  //top right, max Re & max Im
                        rP = rr1;
                        cP = cc2;
                    }
                    else if (j == 3) {  //bottom left, min Re & min Im
                        rP = rr2;
                        cP = cc1;
                    }
                    else if (j == 4) {  //bottom right, max Re & min Im
                        rP = rr2;
                        cP = cc2;
                    }
                    // Note rr1 and cc1 are negative while rr2 and cc2 are positive
                    double x0P = x0 + scaleTerm * cP;  //Re position
                    double y0P = y0 - scaleTerm * rP;  //Im position

                    // Iteration loop using perturbation method, save delta values
                    double[] DxP = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
                    double[] DyP = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
                    double x, y, Dxtemp;
                    int i2 = 1;
                    double xP = x0P;
                    double yP = y0P;
                    double Dx0 = cP * scaleTerm;
                    double Dy0 = -rP * scaleTerm;
                    double Dx = -Dx0;  //delta1 = Y1 - X1
                    double Dy = -Dy0;
                    bool flagBreak = false;
                    DxP[0] = Dx;
                    DyP[0] = Dy;
                    for (int i = 1; i <= maxRefSeries - 1; i++) {
                        i2 = i;
                        x = xMP[i, 1];
                        y = yMP[i, 1];
                        xP = Dx + x;
                        yP = Dy + y;
                        if (xP * xP + yP * yP < glitchDetect[i, 1]) {
                            flagBreak = true;
                            break;
                        }
                        if (xP * xP + yP * yP > bailout) {
                            flagBreak = true;
                            break;
                        }
                        if ((i + 1) % 2 == 1) {
                            Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy - Dx0;
                            Dy = 2 * (x * Dy + y * Dx + Dx * Dy) - Dy0;
                        }
                        else {
                            Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0;
                            Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dy0;
                        }
                        Dx = Dxtemp;
                        if (Dx * Dx + Dy * Dy > divThreshold) {
                            flagBreak = true;
                            break;
                        }
                        DxP[i] = Dx;
                        DyP[i] = Dy;
                    }  //next i
                    if (flagBreak == false) i2++;

                    // Compare delta values calculated using series approximation (DxS, DyS) to delta values calculated using
                    // perturbation theory (DxP, DyP) to determine how long the series approximation is valid
                    if (i2 < (10 + 1)) {
                        skipSeriesFlag = true;
                        break;
                    }
                    int k2 = 1;
                    double DxS, DyS;
                    flagBreak = false;
                    for (int k = 1; k <= i2 - 2; k++) {
                        k2 = k + 1;
                        DxS = A1r[k] * Dx0 + A2r[k] * Dy0 + B1r[k] * Dx0 * Dx0 + B2r[k] * Dy0 * Dy0 + B3r[k] * Dx0 * Dy0 + C1r[k] * Dx0 * Dx0 * Dx0 + C2r[k] * Dy0 * Dy0 * Dy0 + C3r[k] * Dx0 * Dx0 * Dy0 + C4r[k] * Dx0 * Dy0 * Dy0 + D1r[k] * Dx0 * Dx0 * Dx0 * Dx0 + D2r[k] * Dy0 * Dy0 * Dy0 * Dy0 + D3r[k] * Dx0 * Dx0 * Dx0 * Dy0 + D4r[k] * Dx0 * Dy0 * Dy0 * Dy0 + D5r[k] * Dx0 * Dx0 * Dy0 * Dy0 + E1r[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2r[k] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3r[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4r[k] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5r[k] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6r[k] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                        DyS = A1i[k] * Dx0 + A2i[k] * Dy0 + B1i[k] * Dx0 * Dx0 + B2i[k] * Dy0 * Dy0 + B3i[k] * Dx0 * Dy0 + C1i[k] * Dx0 * Dx0 * Dx0 + C2i[k] * Dy0 * Dy0 * Dy0 + C3i[k] * Dx0 * Dx0 * Dy0 + C4i[k] * Dx0 * Dy0 * Dy0 + D1i[k] * Dx0 * Dx0 * Dx0 * Dx0 + D2i[k] * Dy0 * Dy0 * Dy0 * Dy0 + D3i[k] * Dx0 * Dx0 * Dx0 * Dy0 + D4i[k] * Dx0 * Dy0 * Dy0 * Dy0 + D5i[k] * Dx0 * Dx0 * Dy0 * Dy0 + E1i[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2i[k] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3i[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4i[k] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5i[k] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6i[k] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                        if (DxP[k] == 0 || DyP[k] == 0) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DxP[k] - DxS) / DxP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DyP[k] - DyS) / DyP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                    }
                    if (flagBreak == true) k2 = k2 - 1;  //go back to when relative difference was below tolerance, this line is surprisingly important
                    if (k2 < 5) {
                        skipSeriesFlag = true;
                        skipNumber = 0;
                        break;
                    }
                    if (j == 1) {
                        skipNumber = k2;
                    }
                    else {
                        if (k2 < skipNumber) skipNumber = k2;
                    }
                }  //next j
            }
            if (numPointsMP == 1) numSkip = skipNumber;  //store number of iterations skipped in initial pass for calculation info output

            // Calculate points about reference point orbit in double precision

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int rP = res * ri + rr1;       //mapping the parallel loop index to the row number relative to the reference point
                    double y0P = ziP - pixelStepP * ri;
                    double x0P = zrP - pixelStepP;
                    double x, y;
                    double xP = 0, xtempP, yP = 0;
                    double Dx = 0, Dxtemp, Dy = 0, Dx0, Dy0;
                    int iterationP = 0;
                    bool skipCalc, goToGlitch;
                    double iterationPD = 0;
                    double log_zn, nu;

                    for (int cP = cc1; cP <= cc2; cP = cP + res) {
                        skipCalc = false;
                        x0P = x0P + pixelStepP;
                        if ((cP == 0 && rP == 0) == false) {
                            if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] == prevMaxIter) || (refreshType == 2 && numPointsMP > 1)) {

                                // Quick pan code
                                if (refreshType == -1 && numPointsMP == 1) {
                                    skipCalc = quickPan(dr, dc, rRef + rP, cRef + cP, navSize);
                                }

                                // If correcting a flagged glitch, only correct location that is flagged
                                if (numPointsMP > 1) {
                                    if (prevGlitchFlag[rRef + rP, cRef + cP] == 0) skipCalc = true;
                                }

                                // Series approximation to skip iterations
                                if (skipCalc == false) {
                                    int m = skipNumber;
                                    int k = 1;
                                    bool tryLessFlag = false;
                                    bool goToTryLess = true;
                                    Dx0 = cP * scaleTerm;
                                    Dy0 = -rP * scaleTerm;
                                    if (skipSeriesFlag == false) {
                                        while (goToTryLess == true) {
                                            goToTryLess = false;
                                            Dx = A1r[m - 1] * Dx0 + A2r[m - 1] * Dy0 + B1r[m - 1] * Dx0 * Dx0 + B2r[m - 1] * Dy0 * Dy0 + B3r[m - 1] * Dx0 * Dy0 + C1r[m - 1] * Dx0 * Dx0 * Dx0 + C2r[m - 1] * Dy0 * Dy0 * Dy0 + C3r[m - 1] * Dx0 * Dx0 * Dy0 + C4r[m - 1] * Dx0 * Dy0 * Dy0 + D1r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 + D2r[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 + D3r[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 + D4r[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 + D5r[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 + E1r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2r[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4r[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5r[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6r[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                                            Dy = A1i[m - 1] * Dx0 + A2i[m - 1] * Dy0 + B1i[m - 1] * Dx0 * Dx0 + B2i[m - 1] * Dy0 * Dy0 + B3i[m - 1] * Dx0 * Dy0 + C1i[m - 1] * Dx0 * Dx0 * Dx0 + C2i[m - 1] * Dy0 * Dy0 * Dy0 + C3i[m - 1] * Dx0 * Dx0 * Dy0 + C4i[m - 1] * Dx0 * Dy0 * Dy0 + D1i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 + D2i[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 + D3i[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 + D4i[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 + D5i[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 + E1i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2i[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4i[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5i[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6i[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                                            if (Dx * Dx + Dy * Dy > divThreshold) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            xP = Dx + xMP[m, 1];
                                            yP = Dy + yMP[m, 1];
                                            if (xP * xP + yP * yP > bailout) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            if (goToTryLess == false) {
                                                k = m;  //skip first m-1 iterations
                                            }
                                            else {
                                                m = m / 2;
                                            }
                                        }  //Wend
                                    }

                                    // Iteration loop
                                    if (k == 1) {  //no series approximation used
                                        Dx = -Dx0;  //delta1 = Y1 - X1
                                        Dy = -Dy0;
                                    }

                                    goToGlitch = false;
                                    for (int j = k; j <= iteration; j++) {
                                        iterationP = j;
                                        x = xMP[j, 1];
                                        y = yMP[j, 1];
                                        xP = Dx + x;
                                        yP = Dy + y;
                                        if (xP * xP + yP * yP < glitchDetect[j, 1]) {
                                            glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                            Interlocked.Increment(ref cntFlagGlitch);
                                            if (flagDebugGlitch == true) iterationPD = maxIter;
                                            goToGlitch = true;
                                            break;
                                        }
                                        if (xP * xP + yP * yP > bailout) break;
                                        if ((j + 1) % 2 == 1) {
                                            Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy - Dx0;
                                            Dy = 2 * (x * Dy + y * Dx + Dx * Dy) - Dy0;
                                        }
                                        else {
                                            Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0;
                                            Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dy0;
                                        }
                                        Dx = Dxtemp;
                                        if (Dx * Dx + Dy * Dy > divThreshold) break;
                                    }  //next j

                                    if (goToGlitch == false) {

                                        // Discrete coloring
                                        if (flagSmooth == false) {
                                            while (xP * xP + yP * yP < 4 && iterationP < maxIter) {
                                                iterationP++;
                                                if (iterationP % 2 == 1) {
                                                    xtempP = xP * xP - yP * yP - x0P;
                                                    yP = 2 * xP * yP - y0P;
                                                }
                                                else {
                                                    xtempP = xP * xP - yP * yP + x0P;
                                                    yP = 2 * xP * yP + y0P;
                                                }
                                                xP = xtempP;
                                            }
                                            iterationPD = iterationP;
                                        }

                                        // Continuous coloring
                                        else {
                                            while (xP * xP + yP * yP < highBailout && iterationP < maxIter) {
                                                iterationP++;
                                                if (iterationP % 2 == 1) {
                                                    xtempP = xP * xP - yP * yP - x0P;
                                                    yP = 2 * xP * yP - y0P;
                                                }
                                                else {
                                                    xtempP = xP * xP - yP * yP + x0P;
                                                    yP = 2 * xP * yP + y0P;
                                                }
                                                xP = xtempP;
                                            }

                                            // Additional calculations for continuous coloring
                                            if (iterationP < maxIter) {
                                                log_zn = Math.Log(xP * xP + yP * yP) / 2;
                                                nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                                iterationPD = iterationP + 1 - nu;
                                                iterationPD = Math.Round(iterationPD, 3);
                                            }
                                            else {
                                                iterationPD = iterationP;
                                            }
                                        }

                                    }

                                    // Check if too many additional iterations past the reference were performed
                                    if (maxPastRef != 0 && iterationP > maxPastRef) {
                                        glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                        Interlocked.Increment(ref cntFlagGlitch);
                                        if (flagDebugGlitch == true) iterationPD = maxIter;
                                    }

                                    // Store result to array
                                    if (res == 1) imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] = iterationPD;
                                    else reduceResolution(ref imageDataTemp, rRef + rP, cRef + cP, iterationPD, res);
                                }
                            }
                        }

                    }  //next c
                });  //next r
                if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
                else excelApp.StatusBar = string.Format("Correcting initial results ({2})... {0:0%}      DLL in use on {1} logical processors.{3}", (i + 1) / (double)n, numProcessors, numPointsMP, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
            copyArray(ref glitchFlag, glitchFlagTemp);
            if (cntFlagGlitch > 0) flagGlitch = true;
        }

        public static void tSeries(ref double[,] imageData, ref int[,] glitchFlag, ref bool flagGlitch, ref int numSkip, bool flagSmooth, int flagDllOption, double x0, double y0, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int dr, int dc, int navSize, int rr1, int rr2, int cc1, int cc2, int rRef, int cRef, int numPointsMP, double seriesTolerance,
            double[,] xMP, double[,] yMP, double[,] glitchDetect, double divThreshold, bool flagDebugGlitch, int iteration, int maxPastRef, int[,] prevGlitchFlag, int prevGlitchInfo2)
        {  // Beetle fractal, perturbation with series approximation
            double pixelStepP = scaleTerm * res;
            int rUpperIndex = ((rr2 - rr1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double zrP = x0 + scaleTerm * cc1;  //min real position
            double ziP = y0 - scaleTerm * rr1;  //max imaginary position
            bool skipSeriesFlag = false;
            bool flagOverflow = false;
            int maxRefSeries = 0;
            int skipNumber = 0;
            double[] A1r = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
            double[] A2r = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
            double[] A1i = new double[iteration];
            double[] A2i = new double[iteration];
            double[] B1r = new double[iteration];
            double[] B2r = new double[iteration];
            double[] B3r = new double[iteration];
            double[] B1i = new double[iteration];
            double[] B2i = new double[iteration];
            double[] B3i = new double[iteration];
            double[] C1r = new double[iteration];
            double[] C2r = new double[iteration];
            double[] C3r = new double[iteration];
            double[] C4r = new double[iteration];
            double[] C1i = new double[iteration];
            double[] C2i = new double[iteration];
            double[] C3i = new double[iteration];
            double[] C4i = new double[iteration];
            double[] D1r = new double[iteration];
            double[] D2r = new double[iteration];
            double[] D3r = new double[iteration];
            double[] D4r = new double[iteration];
            double[] D5r = new double[iteration];
            double[] D1i = new double[iteration];
            double[] D2i = new double[iteration];
            double[] D3i = new double[iteration];
            double[] D4i = new double[iteration];
            double[] D5i = new double[iteration];
            double[] E1r = new double[iteration];
            double[] E2r = new double[iteration];
            double[] E3r = new double[iteration];
            double[] E4r = new double[iteration];
            double[] E5r = new double[iteration];
            double[] E6r = new double[iteration];
            double[] E1i = new double[iteration];
            double[] E2i = new double[iteration];
            double[] E3i = new double[iteration];
            double[] E4i = new double[iteration];
            double[] E5i = new double[iteration];
            double[] E6i = new double[iteration];
            bool calculatingO3Flag = false, calcO5flag = false, useOrder3 = false;
            int maxO3Series = 0, maxO5Series = 0, maxSeries = 0;
            int skipNumberO3 = 0, skipNumberO5 = 0;
            int i3 = 0;
            int bailout = 4;
            if (flagSmooth == true) bailout = highBailout;
            double cs2;

            // The following are used to avoid passing variables by reference in the Parallel.For loop, array bases are unchanged when using the Clone method
            double[,] imageDataTemp = imageData.Clone() as double[,];
            int[,] glitchFlagTemp = glitchFlag.Clone() as int[,];
            flagGlitch = false;  //this is set in the Parallel.For loop using the Interlocked.Increment Method on cntFlagGlitch and then checking for non-zero values
            int cntFlagGlitch = 0;

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);
            else excelApp.StatusBar = string.Format("Correcting initial results ({1})... 0%      DLL in use on {0} logical processors.{2}", numProcessors, numPointsMP, dllOptionString);

            // Calculate coefficients of series approximation up to a 5th order approximation
            // When correcting glitches, series approximation is skipped if the largest glitch is small
            if (numPointsMP == 1) {
                if (iteration < 10) skipSeriesFlag = true;
            }
            else {
                if (iteration < 10 || prevGlitchInfo2 < 10) skipSeriesFlag = true;
            }
            if (skipSeriesFlag == false) {
                A1r[1] = -2 * xMP[1, 1] + 1;
                A2r[1] = 2 * yMP[1, 1];
                B1r[1] = 1;
                B2r[1] = -1;
                A1i[1] = -A2r[1];
                A2i[1] = A1r[1];
                B3i[1] = 2;
                calcO5flag = true;
                for (int i = 2; i <= iteration - 1; i++) {
                    maxRefSeries = i;
                    calculatingO3Flag = true;
                    if ((i + 1) % 2 == 1) {
                        cs2 = -1;
                    }
                    else {
                        cs2 = 1;
                    }
                    A1r[i] = 2 * (-A1i[i - 1] * yMP[i, 1] + A1r[i - 1] * xMP[i, 1]) + cs2;
                    A2r[i] = 2 * (-A2i[i - 1] * yMP[i, 1] + A2r[i - 1] * xMP[i, 1]);
                    B1r[i] = -A1i[i - 1] * A1i[i - 1] + A1r[i - 1] * A1r[i - 1] + 2 * (-B1i[i - 1] * yMP[i, 1] + B1r[i - 1] * xMP[i, 1]);
                    B2r[i] = -A2i[i - 1] * A2i[i - 1] + A2r[i - 1] * A2r[i - 1] + 2 * (-B2i[i - 1] * yMP[i, 1] + B2r[i - 1] * xMP[i, 1]);
                    B3r[i] = 2 * (-A1i[i - 1] * A2i[i - 1] + A1r[i - 1] * A2r[i - 1] - B3i[i - 1] * yMP[i, 1] + B3r[i - 1] * xMP[i, 1]);
                    C1r[i] = 2 * (-A1i[i - 1] * B1i[i - 1] + A1r[i - 1] * B1r[i - 1] - C1i[i - 1] * yMP[i, 1] + C1r[i - 1] * xMP[i, 1]);
                    C2r[i] = 2 * (-A2i[i - 1] * B2i[i - 1] + A2r[i - 1] * B2r[i - 1] - C2i[i - 1] * yMP[i, 1] + C2r[i - 1] * xMP[i, 1]);
                    C3r[i] = 2 * (-A1i[i - 1] * B3i[i - 1] + A1r[i - 1] * B3r[i - 1] - A2i[i - 1] * B1i[i - 1] + A2r[i - 1] * B1r[i - 1] - C3i[i - 1] * yMP[i, 1] + C3r[i - 1] * xMP[i, 1]);
                    C4r[i] = 2 * (-A1i[i - 1] * B2i[i - 1] + A1r[i - 1] * B2r[i - 1] - A2i[i - 1] * B3i[i - 1] + A2r[i - 1] * B3r[i - 1] - C4i[i - 1] * yMP[i, 1] + C4r[i - 1] * xMP[i, 1]);
                    A1i[i] = 2 * (A1i[i - 1] * xMP[i, 1] + A1r[i - 1] * yMP[i, 1]);
                    A2i[i] = 2 * (A2i[i - 1] * xMP[i, 1] + A2r[i - 1] * yMP[i, 1]) + cs2;
                    B1i[i] = 2 * (A1i[i - 1] * A1r[i - 1] + B1i[i - 1] * xMP[i, 1] + B1r[i - 1] * yMP[i, 1]);
                    B2i[i] = 2 * (A2i[i - 1] * A2r[i - 1] + B2i[i - 1] * xMP[i, 1] + B2r[i - 1] * yMP[i, 1]);
                    B3i[i] = 2 * (A1i[i - 1] * A2r[i - 1] + A1r[i - 1] * A2i[i - 1] + B3i[i - 1] * xMP[i, 1] + B3r[i - 1] * yMP[i, 1]);
                    C1i[i] = 2 * (A1i[i - 1] * B1r[i - 1] + A1r[i - 1] * B1i[i - 1] + C1i[i - 1] * xMP[i, 1] + C1r[i - 1] * yMP[i, 1]);
                    C2i[i] = 2 * (A2i[i - 1] * B2r[i - 1] + A2r[i - 1] * B2i[i - 1] + C2i[i - 1] * xMP[i, 1] + C2r[i - 1] * yMP[i, 1]);
                    C3i[i] = 2 * (A1i[i - 1] * B3r[i - 1] + A1r[i - 1] * B3i[i - 1] + A2i[i - 1] * B1r[i - 1] + A2r[i - 1] * B1i[i - 1] + C3i[i - 1] * xMP[i, 1] + C3r[i - 1] * yMP[i, 1]);
                    C4i[i] = 2 * (A1i[i - 1] * B2r[i - 1] + A1r[i - 1] * B2i[i - 1] + A2i[i - 1] * B3r[i - 1] + A2r[i - 1] * B3i[i - 1] + C4i[i - 1] * xMP[i, 1] + C4r[i - 1] * yMP[i, 1]);
                    // Check for overflow, unlike VBA an exception is not thrown when a double exceeds the maximum magnitude
                    if (Double.IsInfinity(C4i[i]) || Double.IsInfinity(C4r[i]) || Double.IsInfinity(C3i[i]) || Double.IsInfinity(C3r[i]) || Double.IsInfinity(C2i[i]) || Double.IsInfinity(C2r[i]) ||
                        Double.IsInfinity(C1i[i]) || Double.IsInfinity(C1r[i]) || Double.IsInfinity(B3i[i]) || Double.IsInfinity(B3r[i]) || Double.IsInfinity(B2i[i]) || Double.IsInfinity(B2r[i]) ||
                        Double.IsInfinity(B1i[i]) || Double.IsInfinity(B1r[i]) || Double.IsInfinity(A2i[i]) || Double.IsInfinity(A2r[i]) || Double.IsInfinity(A1i[i]) || Double.IsInfinity(A1r[i])) {
                        flagOverflow = true;
                        break;
                    }
                    calculatingO3Flag = false;
                    D1r[i] = -B1i[i - 1] * B1i[i - 1] + B1r[i - 1] * B1r[i - 1] + 2 * (-A1i[i - 1] * C1i[i - 1] + A1r[i - 1] * C1r[i - 1] - D1i[i - 1] * yMP[i, 1] + D1r[i - 1] * xMP[i, 1]);
                    D2r[i] = -B2i[i - 1] * B2i[i - 1] + B2r[i - 1] * B2r[i - 1] + 2 * (-A2i[i - 1] * C2i[i - 1] + A2r[i - 1] * C2r[i - 1] - D2i[i - 1] * yMP[i, 1] + D2r[i - 1] * xMP[i, 1]);
                    D3r[i] = 2 * (-A1i[i - 1] * C3i[i - 1] + A1r[i - 1] * C3r[i - 1] - A2i[i - 1] * C1i[i - 1] + A2r[i - 1] * C1r[i - 1] - B1i[i - 1] * B3i[i - 1] + B1r[i - 1] * B3r[i - 1] - D3i[i - 1] * yMP[i, 1] + D3r[i - 1] * xMP[i, 1]);
                    D4r[i] = 2 * (-A1i[i - 1] * C2i[i - 1] + A1r[i - 1] * C2r[i - 1] - A2i[i - 1] * C4i[i - 1] + A2r[i - 1] * C4r[i - 1] - B2i[i - 1] * B3i[i - 1] + B2r[i - 1] * B3r[i - 1] - D4i[i - 1] * yMP[i, 1] + D4r[i - 1] * xMP[i, 1]);
                    D5r[i] = -B3i[i - 1] * B3i[i - 1] + B3r[i - 1] * B3r[i - 1] + 2 * (-A1i[i - 1] * C4i[i - 1] + A1r[i - 1] * C4r[i - 1] - A2i[i - 1] * C3i[i - 1] + A2r[i - 1] * C3r[i - 1] - B1i[i - 1] * B2i[i - 1] + B1r[i - 1] * B2r[i - 1] - D5i[i - 1] * yMP[i, 1] + D5r[i - 1] * xMP[i, 1]);
                    E1r[i] = 2 * (-A1i[i - 1] * D1i[i - 1] + A1r[i - 1] * D1r[i - 1] - B1i[i - 1] * C1i[i - 1] + B1r[i - 1] * C1r[i - 1] - E1i[i - 1] * yMP[i, 1] + E1r[i - 1] * xMP[i, 1]);
                    E2r[i] = 2 * (-A2i[i - 1] * D2i[i - 1] + A2r[i - 1] * D2r[i - 1] - B2i[i - 1] * C2i[i - 1] + B2r[i - 1] * C2r[i - 1] - E2i[i - 1] * yMP[i, 1] + E2r[i - 1] * xMP[i, 1]);
                    E3r[i] = 2 * (-A1i[i - 1] * D3i[i - 1] + A1r[i - 1] * D3r[i - 1] - A2i[i - 1] * D1i[i - 1] + A2r[i - 1] * D1r[i - 1] - B1i[i - 1] * C3i[i - 1] + B1r[i - 1] * C3r[i - 1] - B3i[i - 1] * C1i[i - 1] + B3r[i - 1] * C1r[i - 1] - E3i[i - 1] * yMP[i, 1] + E3r[i - 1] * xMP[i, 1]);
                    E4r[i] = 2 * (-A1i[i - 1] * D2i[i - 1] + A1r[i - 1] * D2r[i - 1] - A2i[i - 1] * D4i[i - 1] + A2r[i - 1] * D4r[i - 1] - B2i[i - 1] * C4i[i - 1] + B2r[i - 1] * C4r[i - 1] - B3i[i - 1] * C2i[i - 1] + B3r[i - 1] * C2r[i - 1] - E4i[i - 1] * yMP[i, 1] + E4r[i - 1] * xMP[i, 1]);
                    E5r[i] = 2 * (-A1i[i - 1] * D5i[i - 1] + A1r[i - 1] * D5r[i - 1] - A2i[i - 1] * D3i[i - 1] + A2r[i - 1] * D3r[i - 1] - B1i[i - 1] * C4i[i - 1] + B1r[i - 1] * C4r[i - 1] - B2i[i - 1] * C1i[i - 1] + B2r[i - 1] * C1r[i - 1] - B3i[i - 1] * C3i[i - 1] + B3r[i - 1] * C3r[i - 1] - E5i[i - 1] * yMP[i, 1] + E5r[i - 1] * xMP[i, 1]);
                    E6r[i] = 2 * (-A1i[i - 1] * D4i[i - 1] + A1r[i - 1] * D4r[i - 1] - A2i[i - 1] * D5i[i - 1] + A2r[i - 1] * D5r[i - 1] - B1i[i - 1] * C2i[i - 1] + B1r[i - 1] * C2r[i - 1] - B2i[i - 1] * C3i[i - 1] + B2r[i - 1] * C3r[i - 1] - B3i[i - 1] * C4i[i - 1] + B3r[i - 1] * C4r[i - 1] - E6i[i - 1] * yMP[i, 1] + E6r[i - 1] * xMP[i, 1]);
                    D1i[i] = 2 * (A1i[i - 1] * C1r[i - 1] + A1r[i - 1] * C1i[i - 1] + B1i[i - 1] * B1r[i - 1] + D1i[i - 1] * xMP[i, 1] + D1r[i - 1] * yMP[i, 1]);
                    D2i[i] = 2 * (A2i[i - 1] * C2r[i - 1] + A2r[i - 1] * C2i[i - 1] + B2i[i - 1] * B2r[i - 1] + D2i[i - 1] * xMP[i, 1] + D2r[i - 1] * yMP[i, 1]);
                    D3i[i] = 2 * (A1i[i - 1] * C3r[i - 1] + A1r[i - 1] * C3i[i - 1] + A2i[i - 1] * C1r[i - 1] + A2r[i - 1] * C1i[i - 1] + B1i[i - 1] * B3r[i - 1] + B1r[i - 1] * B3i[i - 1] + D3i[i - 1] * xMP[i, 1] + D3r[i - 1] * yMP[i, 1]);
                    D4i[i] = 2 * (A1i[i - 1] * C2r[i - 1] + A1r[i - 1] * C2i[i - 1] + A2i[i - 1] * C4r[i - 1] + A2r[i - 1] * C4i[i - 1] + B2i[i - 1] * B3r[i - 1] + B2r[i - 1] * B3i[i - 1] + D4i[i - 1] * xMP[i, 1] + D4r[i - 1] * yMP[i, 1]);
                    D5i[i] = 2 * (A1i[i - 1] * C4r[i - 1] + A1r[i - 1] * C4i[i - 1] + A2i[i - 1] * C3r[i - 1] + A2r[i - 1] * C3i[i - 1] + B1i[i - 1] * B2r[i - 1] + B1r[i - 1] * B2i[i - 1] + B3i[i - 1] * B3r[i - 1] + D5i[i - 1] * xMP[i, 1] + D5r[i - 1] * yMP[i, 1]);
                    E1i[i] = 2 * (A1i[i - 1] * D1r[i - 1] + A1r[i - 1] * D1i[i - 1] + B1i[i - 1] * C1r[i - 1] + B1r[i - 1] * C1i[i - 1] + E1i[i - 1] * xMP[i, 1] + E1r[i - 1] * yMP[i, 1]);
                    E2i[i] = 2 * (A2i[i - 1] * D2r[i - 1] + A2r[i - 1] * D2i[i - 1] + B2i[i - 1] * C2r[i - 1] + B2r[i - 1] * C2i[i - 1] + E2i[i - 1] * xMP[i, 1] + E2r[i - 1] * yMP[i, 1]);
                    E3i[i] = 2 * (A1i[i - 1] * D3r[i - 1] + A1r[i - 1] * D3i[i - 1] + A2i[i - 1] * D1r[i - 1] + A2r[i - 1] * D1i[i - 1] + B1i[i - 1] * C3r[i - 1] + B1r[i - 1] * C3i[i - 1] + B3i[i - 1] * C1r[i - 1] + B3r[i - 1] * C1i[i - 1] + E3i[i - 1] * xMP[i, 1] + E3r[i - 1] * yMP[i, 1]);
                    E4i[i] = 2 * (A1i[i - 1] * D2r[i - 1] + A1r[i - 1] * D2i[i - 1] + A2i[i - 1] * D4r[i - 1] + A2r[i - 1] * D4i[i - 1] + B2i[i - 1] * C4r[i - 1] + B2r[i - 1] * C4i[i - 1] + B3i[i - 1] * C2r[i - 1] + B3r[i - 1] * C2i[i - 1] + E4i[i - 1] * xMP[i, 1] + E4r[i - 1] * yMP[i, 1]);
                    E5i[i] = 2 * (A1i[i - 1] * D5r[i - 1] + A1r[i - 1] * D5i[i - 1] + A2i[i - 1] * D3r[i - 1] + A2r[i - 1] * D3i[i - 1] + B1i[i - 1] * C4r[i - 1] + B1r[i - 1] * C4i[i - 1] + B2i[i - 1] * C1r[i - 1] + B2r[i - 1] * C1i[i - 1] + B3i[i - 1] * C3r[i - 1] + B3r[i - 1] * C3i[i - 1] + E5i[i - 1] * xMP[i, 1] + E5r[i - 1] * yMP[i, 1]);
                    E6i[i] = 2 * (A1i[i - 1] * D4r[i - 1] + A1r[i - 1] * D4i[i - 1] + A2i[i - 1] * D5r[i - 1] + A2r[i - 1] * D5i[i - 1] + B1i[i - 1] * C2r[i - 1] + B1r[i - 1] * C2i[i - 1] + B2i[i - 1] * C3r[i - 1] + B2r[i - 1] * C3i[i - 1] + B3i[i - 1] * C4r[i - 1] + B3r[i - 1] * C4i[i - 1] + E6i[i - 1] * xMP[i, 1] + E6r[i - 1] * yMP[i, 1]);
                    if (Double.IsInfinity(E6i[i]) || Double.IsInfinity(E6r[i]) || Double.IsInfinity(E5i[i]) || Double.IsInfinity(E5r[i]) || Double.IsInfinity(E4i[i]) || Double.IsInfinity(E4r[i]) ||
                        Double.IsInfinity(E3i[i]) || Double.IsInfinity(E3r[i]) || Double.IsInfinity(E2i[i]) || Double.IsInfinity(E2r[i]) || Double.IsInfinity(E1i[i]) || Double.IsInfinity(E1r[i]) ||
                        Double.IsInfinity(D5i[i]) || Double.IsInfinity(D5r[i]) || Double.IsInfinity(D4i[i]) || Double.IsInfinity(D4r[i]) || Double.IsInfinity(D3i[i]) || Double.IsInfinity(D3r[i]) ||
                        Double.IsInfinity(D2i[i]) || Double.IsInfinity(D2r[i]) || Double.IsInfinity(D1i[i]) || Double.IsInfinity(D1r[i])) {
                        flagOverflow = true;
                        break;
                    }
                }  //next i

                // Catch overflow condition with flag
                if (flagOverflow == true) {
                    if (calculatingO3Flag == false) {  //overflow occurred in order 4 or 5 term
                        maxO5Series = maxRefSeries - 1;
                        flagOverflow = false;  //reset flag
                        calcO5flag = false;
                        i3 = maxRefSeries;
                    }
                    else {  //overflow occurred in order 3 term
                        maxO3Series = maxRefSeries - 1;
                        maxRefSeries = maxRefSeries - 1;
                    }
                }

                // If overflow occurred in order 4 or 5 term, calculate additional order 3 terms if possible
                if (calcO5flag == false) {  //OverflowReturnIn
                    for (int i = i3 + 1; i <= iteration - 1; i++) {
                        maxRefSeries = i;
                        if ((i + 1) % 2 == 1) {
                            cs2 = -1;
                        }
                        else {
                            cs2 = 1;
                        }
                        A1r[i] = 2 * (-A1i[i - 1] * yMP[i, 1] + A1r[i - 1] * xMP[i, 1]) + cs2;
                        A2r[i] = 2 * (-A2i[i - 1] * yMP[i, 1] + A2r[i - 1] * xMP[i, 1]);
                        B1r[i] = -A1i[i - 1] * A1i[i - 1] + A1r[i - 1] * A1r[i - 1] + 2 * (-B1i[i - 1] * yMP[i, 1] + B1r[i - 1] * xMP[i, 1]);
                        B2r[i] = -A2i[i - 1] * A2i[i - 1] + A2r[i - 1] * A2r[i - 1] + 2 * (-B2i[i - 1] * yMP[i, 1] + B2r[i - 1] * xMP[i, 1]);
                        B3r[i] = 2 * (-A1i[i - 1] * A2i[i - 1] + A1r[i - 1] * A2r[i - 1] - B3i[i - 1] * yMP[i, 1] + B3r[i - 1] * xMP[i, 1]);
                        C1r[i] = 2 * (-A1i[i - 1] * B1i[i - 1] + A1r[i - 1] * B1r[i - 1] - C1i[i - 1] * yMP[i, 1] + C1r[i - 1] * xMP[i, 1]);
                        C2r[i] = 2 * (-A2i[i - 1] * B2i[i - 1] + A2r[i - 1] * B2r[i - 1] - C2i[i - 1] * yMP[i, 1] + C2r[i - 1] * xMP[i, 1]);
                        C3r[i] = 2 * (-A1i[i - 1] * B3i[i - 1] + A1r[i - 1] * B3r[i - 1] - A2i[i - 1] * B1i[i - 1] + A2r[i - 1] * B1r[i - 1] - C3i[i - 1] * yMP[i, 1] + C3r[i - 1] * xMP[i, 1]);
                        C4r[i] = 2 * (-A1i[i - 1] * B2i[i - 1] + A1r[i - 1] * B2r[i - 1] - A2i[i - 1] * B3i[i - 1] + A2r[i - 1] * B3r[i - 1] - C4i[i - 1] * yMP[i, 1] + C4r[i - 1] * xMP[i, 1]);
                        A1i[i] = 2 * (A1i[i - 1] * xMP[i, 1] + A1r[i - 1] * yMP[i, 1]);
                        A2i[i] = 2 * (A2i[i - 1] * xMP[i, 1] + A2r[i - 1] * yMP[i, 1]) + cs2;
                        B1i[i] = 2 * (A1i[i - 1] * A1r[i - 1] + B1i[i - 1] * xMP[i, 1] + B1r[i - 1] * yMP[i, 1]);
                        B2i[i] = 2 * (A2i[i - 1] * A2r[i - 1] + B2i[i - 1] * xMP[i, 1] + B2r[i - 1] * yMP[i, 1]);
                        B3i[i] = 2 * (A1i[i - 1] * A2r[i - 1] + A1r[i - 1] * A2i[i - 1] + B3i[i - 1] * xMP[i, 1] + B3r[i - 1] * yMP[i, 1]);
                        C1i[i] = 2 * (A1i[i - 1] * B1r[i - 1] + A1r[i - 1] * B1i[i - 1] + C1i[i - 1] * xMP[i, 1] + C1r[i - 1] * yMP[i, 1]);
                        C2i[i] = 2 * (A2i[i - 1] * B2r[i - 1] + A2r[i - 1] * B2i[i - 1] + C2i[i - 1] * xMP[i, 1] + C2r[i - 1] * yMP[i, 1]);
                        C3i[i] = 2 * (A1i[i - 1] * B3r[i - 1] + A1r[i - 1] * B3i[i - 1] + A2i[i - 1] * B1r[i - 1] + A2r[i - 1] * B1i[i - 1] + C3i[i - 1] * xMP[i, 1] + C3r[i - 1] * yMP[i, 1]);
                        C4i[i] = 2 * (A1i[i - 1] * B2r[i - 1] + A1r[i - 1] * B2i[i - 1] + A2i[i - 1] * B3r[i - 1] + A2r[i - 1] * B3i[i - 1] + C4i[i - 1] * xMP[i, 1] + C4r[i - 1] * yMP[i, 1]);
                        if (Double.IsInfinity(C4i[i]) || Double.IsInfinity(C4r[i]) || Double.IsInfinity(C3i[i]) || Double.IsInfinity(C3r[i]) || Double.IsInfinity(C2i[i]) || Double.IsInfinity(C2r[i]) ||
                            Double.IsInfinity(C1i[i]) || Double.IsInfinity(C1r[i]) || Double.IsInfinity(B3i[i]) || Double.IsInfinity(B3r[i]) || Double.IsInfinity(B2i[i]) || Double.IsInfinity(B2r[i]) ||
                            Double.IsInfinity(B1i[i]) || Double.IsInfinity(B1r[i]) || Double.IsInfinity(A2i[i]) || Double.IsInfinity(A2r[i]) || Double.IsInfinity(A1i[i]) || Double.IsInfinity(A1r[i])) {
                            flagOverflow = true;
                            break;
                        }
                    }  //next i
                    if (flagOverflow == true) {
                        maxO3Series = maxRefSeries - 1;
                        maxRefSeries = maxRefSeries - 1;
                    }
                }
            }
            if (maxO3Series == 0) maxO3Series = maxRefSeries;
            if (maxO5Series == 0) maxO5Series = maxRefSeries;
            if (maxRefSeries < 10) skipSeriesFlag = true;

            // Determine how many iterations to skip
            if (skipSeriesFlag == false) {
                for (int j = 1; j <= 4; j++) {  //4 corner pixels of range
                    int rP = 0, cP = 0;
                    if (j == 1) {  //top left, min Re & max Im
                        rP = rr1;
                        cP = cc1;
                    }
                    else if (j == 2) {  //top right, max Re & max Im
                        rP = rr1;
                        cP = cc2;
                    }
                    else if (j == 3) {  //bottom left, min Re & min Im
                        rP = rr2;
                        cP = cc1;
                    }
                    else if (j == 4) {  //bottom right, max Re & min Im
                        rP = rr2;
                        cP = cc2;
                    }
                    // Note rr1 and cc1 are negative while rr2 and cc2 are positive
                    double x0P = x0 + scaleTerm * cP;  //Re position
                    double y0P = y0 - scaleTerm * rP;  //Im position

                    // Iteration loop using perturbation method, save delta values
                    double[] DxP = new double[iteration];  //arrays created in C# are base 0, this is contrast to the arrays passed in from VBA or cloned from VBA arrays
                    double[] DyP = new double[iteration];  //upper bound of C# array dimension is exclusive unlike VBA
                    double x, y, Dxtemp;
                    int i2 = 1;
                    double xP = x0P;
                    double yP = y0P;
                    double Dx0 = cP * scaleTerm;
                    double Dy0 = -rP * scaleTerm;
                    double Dx = -Dx0;  //delta1 = Y1 - X1
                    double Dy = -Dy0;
                    bool flagBreak = false;
                    DxP[0] = Dx;
                    DyP[0] = Dy;
                    for (int i = 1; i <= maxRefSeries - 1; i++) {
                        i2 = i;
                        x = xMP[i, 1];
                        y = yMP[i, 1];
                        xP = Dx + x;
                        yP = Dy + y;
                        if (xP * xP + yP * yP < glitchDetect[i, 1]) {
                            flagBreak = true;
                            break;
                        }
                        if (xP * xP + yP * yP > bailout) {
                            flagBreak = true;
                            break;
                        }
                        if ((i + 1) % 2 == 1) {
                            Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy - Dx0;
                            Dy = 2 * (x * Dy + y * Dx + Dx * Dy) - Dy0;
                        }
                        else {
                            Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0;
                            Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dy0;
                        }
                        Dx = Dxtemp;
                        if (Dx * Dx + Dy * Dy > divThreshold) {
                            flagBreak = true;
                            break;
                        }
                        DxP[i] = Dx;
                        DyP[i] = Dy;
                    }  //next i
                    if (flagBreak == false) i2++;

                    // Compare delta values calculated using series approximation (DxS, DyS) to delta values calculated using
                    // perturbation theory (DxP, DyP) to determine how long the series approximation is valid
                    if (i2 < (10 + 1)) {
                        skipSeriesFlag = true;
                        break;
                    }

                    // Order 3 approximation
                    maxSeries = Math.Min(maxO3Series - 1, i2 - 1);
                    int k2 = 1;
                    double DxS, DyS;
                    flagBreak = false;
                    for (int k = 1; k <= maxSeries - 1; k++) {
                        k2 = k + 1;
                        DxS = A1r[k] * Dx0 + A2r[k] * Dy0 + B1r[k] * Dx0 * Dx0 + B2r[k] * Dy0 * Dy0 + B3r[k] * Dx0 * Dy0 + C1r[k] * Dx0 * Dx0 * Dx0 + C2r[k] * Dy0 * Dy0 * Dy0 + C3r[k] * Dx0 * Dx0 * Dy0 + C4r[k] * Dx0 * Dy0 * Dy0;
                        DyS = A1i[k] * Dx0 + A2i[k] * Dy0 + B1i[k] * Dx0 * Dx0 + B2i[k] * Dy0 * Dy0 + B3i[k] * Dx0 * Dy0 + C1i[k] * Dx0 * Dx0 * Dx0 + C2i[k] * Dy0 * Dy0 * Dy0 + C3i[k] * Dx0 * Dx0 * Dy0 + C4i[k] * Dx0 * Dy0 * Dy0;
                        if (DxP[k] == 0 || DyP[k] == 0) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DxP[k] - DxS) / DxP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DyP[k] - DyS) / DyP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                    }
                    if (flagBreak == true) k2 = k2 - 1;  //go back to when relative difference was below tolerance, this line is surprisingly important
                    if (j == 1) {
                        skipNumberO3 = k2;
                    }
                    else {
                        if (k2 < skipNumberO3) skipNumberO3 = k2;
                    }

                    // Order 5 approximation
                    maxSeries = Math.Min(maxO5Series - 1, i2 - 1);
                    int m = k2;  //purpose of 5th order approximation is to skip more iterations so we only need to check additional iterations
                    if (skipNumberO5 == -1) m = maxSeries + 1;  //avoid entering loop if k2 = m on a previous pass
                    flagBreak = false;
                    for (int k = m - 1; k <= maxSeries - 1; k++) {
                        k2 = k + 1;
                        DxS = A1r[k] * Dx0 + A2r[k] * Dy0 + B1r[k] * Dx0 * Dx0 + B2r[k] * Dy0 * Dy0 + B3r[k] * Dx0 * Dy0 + C1r[k] * Dx0 * Dx0 * Dx0 + C2r[k] * Dy0 * Dy0 * Dy0 + C3r[k] * Dx0 * Dx0 * Dy0 + C4r[k] * Dx0 * Dy0 * Dy0 + D1r[k] * Dx0 * Dx0 * Dx0 * Dx0 + D2r[k] * Dy0 * Dy0 * Dy0 * Dy0 + D3r[k] * Dx0 * Dx0 * Dx0 * Dy0 + D4r[k] * Dx0 * Dy0 * Dy0 * Dy0 + D5r[k] * Dx0 * Dx0 * Dy0 * Dy0 + E1r[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2r[k] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3r[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4r[k] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5r[k] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6r[k] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                        DyS = A1i[k] * Dx0 + A2i[k] * Dy0 + B1i[k] * Dx0 * Dx0 + B2i[k] * Dy0 * Dy0 + B3i[k] * Dx0 * Dy0 + C1i[k] * Dx0 * Dx0 * Dx0 + C2i[k] * Dy0 * Dy0 * Dy0 + C3i[k] * Dx0 * Dx0 * Dy0 + C4i[k] * Dx0 * Dy0 * Dy0 + D1i[k] * Dx0 * Dx0 * Dx0 * Dx0 + D2i[k] * Dy0 * Dy0 * Dy0 * Dy0 + D3i[k] * Dx0 * Dx0 * Dx0 * Dy0 + D4i[k] * Dx0 * Dy0 * Dy0 * Dy0 + D5i[k] * Dx0 * Dx0 * Dy0 * Dy0 + E1i[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2i[k] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3i[k] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4i[k] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5i[k] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6i[k] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                        if (DxP[k] == 0 || DyP[k] == 0) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DxP[k] - DxS) / DxP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                        if (Math.Abs((DyP[k] - DyS) / DyP[k]) > seriesTolerance) {
                            flagBreak = true;
                            break;
                        }
                    }
                    if (k2 == m) k2 = 0;
                    if (flagBreak == true) k2 = k2 - 1;  //go back to when relative difference was below tolerance, this line is surprisingly important
                    if (j == 1) {
                        skipNumberO5 = k2;
                    }
                    else {
                        if (k2 < skipNumberO5) skipNumberO5 = k2;
                    }

                    // Check if it is worth continuing
                    if (Math.Max(skipNumberO3, skipNumberO5) < 5) {
                        skipSeriesFlag = true;
                        skipNumber = 0;
                        break;
                    }

                }  //next j

                // Determine final number of iterations to skip and what order of approximation to use
                // Order 5 calculations are more expensive so they are penalized slightly in the comparison
                if (skipSeriesFlag == false) {
                    if (skipNumberO3 > (skipNumberO5 - 10)) {
                        skipNumber = skipNumberO3;
                        useOrder3 = true;
                    }
                    else {  //use order 5
                        skipNumber = skipNumberO5;
                        useOrder3 = false;
                    }
                }
            }
            if (numPointsMP == 1) numSkip = skipNumber;  //store number of iterations skipped in initial pass for calculation info output

            // Calculate points about reference point orbit in double precision

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int rP = res * ri + rr1;       //mapping the parallel loop index to the row number relative to the reference point
                    double y0P = ziP - pixelStepP * ri;
                    double x0P = zrP - pixelStepP;
                    double x, y;
                    double xP = 0, xtempP, yP = 0;
                    double Dx = 0, Dxtemp, Dy = 0, Dx0, Dy0;
                    int iterationP = 0;
                    bool skipCalc, goToGlitch;
                    double iterationPD = 0;
                    double log_zn, nu;

                    for (int cP = cc1; cP <= cc2; cP = cP + res) {
                        skipCalc = false;
                        x0P = x0P + pixelStepP;
                        if ((cP == 0 && rP == 0) == false) {
                            if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] == prevMaxIter) || (refreshType == 2 && numPointsMP > 1)) {

                                // Quick pan code
                                if (refreshType == -1 && numPointsMP == 1) {
                                    skipCalc = quickPan(dr, dc, rRef + rP, cRef + cP, navSize);
                                }

                                // If correcting a flagged glitch, only correct location that is flagged
                                if (numPointsMP > 1) {
                                    if (prevGlitchFlag[rRef + rP, cRef + cP] == 0) skipCalc = true;
                                }

                                // Series approximation to skip iterations
                                if (skipCalc == false) {
                                    int m = skipNumber;
                                    int k = 1;
                                    bool tryLessFlag = false;
                                    bool goToTryLess = true;
                                    Dx0 = cP * scaleTerm;
                                    Dy0 = -rP * scaleTerm;
                                    if (skipSeriesFlag == false) {
                                        while (goToTryLess == true) {
                                            goToTryLess = false;
                                            if (useOrder3 == true) {
                                                Dx = A1r[m - 1] * Dx0 + A2r[m - 1] * Dy0 + B1r[m - 1] * Dx0 * Dx0 + B2r[m - 1] * Dy0 * Dy0 + B3r[m - 1] * Dx0 * Dy0 + C1r[m - 1] * Dx0 * Dx0 * Dx0 + C2r[m - 1] * Dy0 * Dy0 * Dy0 + C3r[m - 1] * Dx0 * Dx0 * Dy0 + C4r[m - 1] * Dx0 * Dy0 * Dy0;
                                                Dy = A1i[m - 1] * Dx0 + A2i[m - 1] * Dy0 + B1i[m - 1] * Dx0 * Dx0 + B2i[m - 1] * Dy0 * Dy0 + B3i[m - 1] * Dx0 * Dy0 + C1i[m - 1] * Dx0 * Dx0 * Dx0 + C2i[m - 1] * Dy0 * Dy0 * Dy0 + C3i[m - 1] * Dx0 * Dx0 * Dy0 + C4i[m - 1] * Dx0 * Dy0 * Dy0;
                                            }
                                            else {
                                                Dx = A1r[m - 1] * Dx0 + A2r[m - 1] * Dy0 + B1r[m - 1] * Dx0 * Dx0 + B2r[m - 1] * Dy0 * Dy0 + B3r[m - 1] * Dx0 * Dy0 + C1r[m - 1] * Dx0 * Dx0 * Dx0 + C2r[m - 1] * Dy0 * Dy0 * Dy0 + C3r[m - 1] * Dx0 * Dx0 * Dy0 + C4r[m - 1] * Dx0 * Dy0 * Dy0 + D1r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 + D2r[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 + D3r[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 + D4r[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 + D5r[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 + E1r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2r[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3r[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4r[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5r[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6r[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                                                Dy = A1i[m - 1] * Dx0 + A2i[m - 1] * Dy0 + B1i[m - 1] * Dx0 * Dx0 + B2i[m - 1] * Dy0 * Dy0 + B3i[m - 1] * Dx0 * Dy0 + C1i[m - 1] * Dx0 * Dx0 * Dx0 + C2i[m - 1] * Dy0 * Dy0 * Dy0 + C3i[m - 1] * Dx0 * Dx0 * Dy0 + C4i[m - 1] * Dx0 * Dy0 * Dy0 + D1i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 + D2i[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 + D3i[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 + D4i[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 + D5i[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 + E1i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dx0 + E2i[m - 1] * Dy0 * Dy0 * Dy0 * Dy0 * Dy0 + E3i[m - 1] * Dx0 * Dx0 * Dx0 * Dx0 * Dy0 + E4i[m - 1] * Dx0 * Dy0 * Dy0 * Dy0 * Dy0 + E5i[m - 1] * Dx0 * Dx0 * Dx0 * Dy0 * Dy0 + E6i[m - 1] * Dx0 * Dx0 * Dy0 * Dy0 * Dy0;
                                            }
                                            if (Dx * Dx + Dy * Dy > divThreshold) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            xP = Dx + xMP[m, 1];
                                            yP = Dy + yMP[m, 1];
                                            if (xP * xP + yP * yP > bailout) {
                                                if (tryLessFlag == false) {
                                                    tryLessFlag = true;
                                                    goToTryLess = true;
                                                }
                                                else {
                                                    break;
                                                }
                                            }
                                            if (goToTryLess == false) {
                                                k = m;  //skip first m-1 iterations
                                            }
                                            else {
                                                m = m / 2;
                                            }
                                        }  //Wend
                                    }

                                    // Iteration loop
                                    if (k == 1) {  //no series approximation used
                                        Dx = -Dx0;  //delta1 = Y1 - X1
                                        Dy = -Dy0;
                                    }

                                    goToGlitch = false;
                                    for (int j = k; j <= iteration; j++) {
                                        iterationP = j;
                                        x = xMP[j, 1];
                                        y = yMP[j, 1];
                                        xP = Dx + x;
                                        yP = Dy + y;
                                        if (xP * xP + yP * yP < glitchDetect[j, 1]) {
                                            glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                            Interlocked.Increment(ref cntFlagGlitch);
                                            if (flagDebugGlitch == true) iterationPD = maxIter;
                                            goToGlitch = true;
                                            break;
                                        }
                                        if (xP * xP + yP * yP > bailout) break;
                                        if ((j + 1) % 2 == 1) {
                                            Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy - Dx0;
                                            Dy = 2 * (x * Dy + y * Dx + Dx * Dy) - Dy0;
                                        }
                                        else {
                                            Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0;
                                            Dy = 2 * (x * Dy + y * Dx + Dx * Dy) + Dy0;
                                        }
                                        Dx = Dxtemp;
                                        if (Dx * Dx + Dy * Dy > divThreshold) break;
                                    }  //next j

                                    if (goToGlitch == false) {

                                        // Discrete coloring
                                        if (flagSmooth == false) {
                                            while (xP * xP + yP * yP < 4 && iterationP < maxIter) {
                                                iterationP++;
                                                if (iterationP % 2 == 1) {
                                                    xtempP = xP * xP - yP * yP - x0P;
                                                    yP = 2 * xP * yP - y0P;
                                                }
                                                else {
                                                    xtempP = xP * xP - yP * yP + x0P;
                                                    yP = 2 * xP * yP + y0P;
                                                }
                                                xP = xtempP;
                                            }
                                            iterationPD = iterationP;
                                        }

                                        // Continuous coloring
                                        else {
                                            while (xP * xP + yP * yP < highBailout && iterationP < maxIter) {
                                                iterationP++;
                                                if (iterationP % 2 == 1) {
                                                    xtempP = xP * xP - yP * yP - x0P;
                                                    yP = 2 * xP * yP - y0P;
                                                }
                                                else {
                                                    xtempP = xP * xP - yP * yP + x0P;
                                                    yP = 2 * xP * yP + y0P;
                                                }
                                                xP = xtempP;
                                            }

                                            // Additional calculations for continuous coloring
                                            if (iterationP < maxIter) {
                                                log_zn = Math.Log(xP * xP + yP * yP) / 2;
                                                nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                                iterationPD = iterationP + 1 - nu;
                                                iterationPD = Math.Round(iterationPD, 3);
                                            }
                                            else {
                                                iterationPD = iterationP;
                                            }
                                        }

                                    }

                                    // Check if too many additional iterations past the reference were performed
                                    if (maxPastRef != 0 && iterationP > maxPastRef) {
                                        glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                        Interlocked.Increment(ref cntFlagGlitch);
                                        if (flagDebugGlitch == true) iterationPD = maxIter;
                                    }

                                    // Store result to array
                                    if (res == 1) imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] = iterationPD;
                                    else reduceResolution(ref imageDataTemp, rRef + rP, cRef + cP, iterationPD, res);
                                }
                            }
                        }

                    }  //next c
                });  //next r
                if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
                else excelApp.StatusBar = string.Format("Correcting initial results ({2})... {0:0%}      DLL in use on {1} logical processors.{3}", (i + 1) / (double)n, numProcessors, numPointsMP, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
            copyArray(ref glitchFlag, glitchFlagTemp);
            if (cntFlagGlitch > 0) flagGlitch = true;
        }

        public static void pDbl(ref double[,] imageData, bool flagSmooth, int flagDllOption, double zr, double zi, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int r1, int c1, int navSize, int rSize, int cSize)
        {  // Perpendicular Burning Ship fractal, standard double precision
            double pixelStep = scaleTerm * res;
            int rUpperIndex = ((rSize - 1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double[,] imageDataTemp = imageData.Clone() as double[,];  //base 1 array, imageData is not allowed in Parallel.For loop because it is passed by reference

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int r = res * ri + 2;          //mapping the parallel loop index to the row number
                    double y0 = zi - pixelStep * ri;
                    double x0 = zr - pixelStep;
                    double x, xtemp, y;
                    int iteration;
                    double iterationD;
                    bool skipCalc;
                    double log_zn, nu;

                    for (int c = 2; c <= cSize + 1; c = c + res) {
                        skipCalc = false;
                        x0 = x0 + pixelStep;
                        if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[r - 1, c - 1] == prevMaxIter)) {

                            // Quick pan code
                            if (refreshType == -1) {
                                skipCalc = quickPan(r1, c1, r, c, navSize);
                            }

                            // Iteration loop
                            if (skipCalc == false) {
                                iteration = 1;
                                x = x0;
                                y = y0;

                                // Discrete coloring
                                if (flagSmooth == false) {
                                    while (x * x + y * y < 4 && iteration < maxIter) {
                                        xtemp = x * x - y * y + x0;
                                        y = 2 * x * Math.Abs(y) + y0;
                                        x = xtemp;
                                        iteration++;
                                    }
                                    iterationD = iteration;
                                }

                                // Continuous coloring
                                else {
                                    while (x * x + y * y < highBailout && iteration < maxIter) {
                                        xtemp = x * x - y * y + x0;
                                        y = 2 * x * Math.Abs(y) + y0;
                                        x = xtemp;
                                        iteration++;
                                    }

                                    // Additional calculations for continuous coloring
                                    if (iteration < maxIter) {
                                        log_zn = Math.Log(x * x + y * y) / 2;
                                        nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                        iterationD = iteration + 1 - nu;
                                        iterationD = Math.Round(iterationD, 3);
                                    }
                                    else {
                                        iterationD = iteration;
                                    }
                                }

                                // Store result to array
                                if (res == 1) imageDataTemp[r - 1, c - 1] = iterationD;
                                else reduceResolution(ref imageDataTemp, r, c, iterationD, res);

                            }
                        }
                    }  //next c
                });  //next r
                excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
        }

        public static void pPert(ref double[,] imageData, ref int[,] glitchFlag, ref bool flagGlitch, bool flagSmooth, int flagDllOption, double x0, double y0, double scaleTerm, int res, int maxIter,
            int refreshType, int prevMaxIter, int dr, int dc, int navSize, int rr1, int rr2, int cc1, int cc2, int rRef, int cRef, int numPointsMP,
            double[,] xMP, double[,] yMP, double[,] glitchDetect, double divThreshold, bool flagDebugGlitch, int iteration, int maxPastRef, int[,] prevGlitchFlag)
        {  // Perpendicular Burning Ship fractal, perturbation methods
            double pixelStepP = scaleTerm * res;
            int rUpperIndex = ((rr2 - rr1) / res) + 1;  //upper index of the Parallel.For loop is exclusive so 1 is added
            double zrP = x0 + scaleTerm * cc1;  //min real position
            double ziP = y0 - scaleTerm * rr1;  //max imaginary position
            int bailout = 4;
            if (flagSmooth == true) bailout = highBailout;

            // The following are used to avoid passing variables by reference in the Parallel.For loop, array bases are unchanged when using the Clone method
            double[,] imageDataTemp = imageData.Clone() as double[,];
            int[,] glitchFlagTemp = glitchFlag.Clone() as int[,];
            flagGlitch = false;  //this is set in the Parallel.For loop using the Interlocked.Increment Method on cntFlagGlitch and then checking for non-zero values
            int cntFlagGlitch = 0;

            // Parallel Options
            int degreeOfParallelism = -1;  //default value, no limit
            int numProcessors = Environment.ProcessorCount;
            string dllOptionString = "";
            if (flagDllOption > -1 && (numProcessors - flagDllOption) > 0) {
                degreeOfParallelism = numProcessors - flagDllOption;
                dllOptionString = String.Format("      Degree of parallelism limited to {0}.", degreeOfParallelism);
            }
            var options = new ParallelOptions();
            options.MaxDegreeOfParallelism = degreeOfParallelism;

            // Initialize Excel status bar
            excelApp = (Excel.Application)Marshal.GetActiveObject("Excel.Application");
            if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... 0%      DLL in use on {0} logical processors.{1}", numProcessors, dllOptionString);
            else excelApp.StatusBar = string.Format("Correcting initial results ({1})... 0%      DLL in use on {0} logical processors.{2}", numProcessors, numPointsMP, dllOptionString);

            // The outer i loop is used to remove the excelApp.StatusBar update from the Parallel.For loop
            int n = getOuterLoopNum(rUpperIndex);  //number of times to run Parallel.For loop for status reporting
            int nd = rUpperIndex / n;              //division (integer)
            int nm = rUpperIndex % n;              //modulus
            int nr;                                //range
            int ns = 0;                            //sum
            for (int i = 0; i < n; i++) {
                if (i < nm) nr = nd + 1;
                else nr = nd;
                ns = ns + nr;

                Parallel.For(ns - nr, ns, options, ri => {  //parallel loop index (ri) is required to increment by 1
                    int rP = res * ri + rr1;       //mapping the parallel loop index to the row number relative to the reference point
                    double y0P = ziP - pixelStepP * ri;
                    double x0P = zrP - pixelStepP;
                    double x, y;
                    double xP = 0, xtempP, yP = 0;
                    double Dx, Dxtemp, Dy, Dx0, Dy0;
                    int iterationP = 0;
                    double iterationPD = 0;
                    bool skipCalc, goToGlitch;
                    double log_zn, nu;

                    for (int cP = cc1; cP <= cc2; cP = cP + res) {
                        skipCalc = false;
                        x0P = x0P + pixelStepP;
                        if ((cP == 0 && rP == 0) == false) {
                            if (refreshType < 1 || (refreshType == 2 && (int)imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] == prevMaxIter) || (refreshType == 2 && numPointsMP > 1)) {

                                // Quick pan code
                                if (refreshType == -1 && numPointsMP == 1) {
                                    skipCalc = quickPan(dr, dc, rRef + rP, cRef + cP, navSize);
                                }

                                // If correcting a flagged glitch, only correct location that is flagged
                                if (numPointsMP > 1) {
                                    if (prevGlitchFlag[rRef + rP, cRef + cP] == 0) skipCalc = true;
                                }

                                // Iteration loop
                                if (skipCalc == false) {
                                    Dx0 = cP * scaleTerm;
                                    Dy0 = -rP * scaleTerm;
                                    Dx = Dx0;
                                    Dy = Dy0;
                                    goToGlitch = false;

                                    for (int j = 1; j <= iteration; j++) {
                                        iterationP = j;
                                        x = xMP[j, 1];
                                        y = yMP[j, 1];
                                        xP = Dx + x;
                                        yP = Dy + y;
                                        if (xP * xP + yP * yP < glitchDetect[j, 1]) {
                                            glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                            Interlocked.Increment(ref cntFlagGlitch);
                                            if (flagDebugGlitch == true) iterationPD = maxIter;
                                            goToGlitch = true;
                                            break;
                                        }
                                        if (xP * xP + yP * yP > bailout) break;
                                        //Dy = 2*(x*(Abs(y + Dy) - Abs(y)) + Dx*Abs(y + Dy)) + Dy0
                                        //The size of the term x*y impacts the accuracy of the expression for Dy
                                        //This term is removed using the following relationships:
                                        //  Abs(A+B)-Abs(A) = Abs(B) except...
                                        //  Abs(A+B)-Abs(A) =-Abs(B) if Sgn(A) is opposite Sgn(B) AND Abs(A) > Abs(B)
                                        //  Abs(A+B)-Abs(A) = Abs(B)-2*Abs(A) if Sgn(A) is opposite Sgn(B) AND Abs(A) < Abs(B)
                                        Dxtemp = 2 * (x * Dx - y * Dy) + Dx * Dx - Dy * Dy + Dx0;
                                        if (Math.Sign(y) + Math.Sign(Dy) != 0) {
                                            Dy = 2 * (x * Math.Abs(Dy) + Dx * Math.Abs(y + Dy)) + Dy0;
                                        }
                                        else if (Math.Abs(y) > Math.Abs(Dy)) {
                                            Dy = 2 * (-x * Math.Abs(Dy) + Dx * Math.Abs(y + Dy)) + Dy0;
                                        }
                                        else {
                                            Dy = 2 * (x * (Math.Abs(y + Dy) - Math.Abs(y)) + Dx * Math.Abs(y + Dy)) + Dy0;
                                        }
                                        Dx = Dxtemp;
                                        if (Dx * Dx + Dy * Dy > divThreshold) break;
                                    }  //next j

                                    if (goToGlitch == false) {

                                        // Discrete coloring
                                        if (flagSmooth == false) {
                                            while (xP * xP + yP * yP < 4 && iterationP < maxIter) {
                                                xtempP = xP * xP - yP * yP + x0P;
                                                yP = 2 * xP * Math.Abs(yP) + y0P;
                                                xP = xtempP;
                                                iterationP++;
                                            }
                                            iterationPD = iterationP;
                                        }

                                        // Continuous coloring
                                        else {
                                            while (xP * xP + yP * yP < highBailout && iterationP < maxIter) {
                                                xtempP = xP * xP - yP * yP + x0P;
                                                yP = 2 * xP * Math.Abs(yP) + y0P;
                                                xP = xtempP;
                                                iterationP++;
                                            }

                                            // Additional calculations for continuous coloring
                                            if (iterationP < maxIter) {
                                                log_zn = Math.Log(xP * xP + yP * yP) / 2;
                                                nu = Math.Log(log_zn / Math.Log(2)) / Math.Log(2);
                                                iterationPD = iterationP + 1 - nu;
                                                iterationPD = Math.Round(iterationPD, 3);
                                            }
                                            else {
                                                iterationPD = iterationP;
                                            }
                                        }

                                    }

                                    // Check if too many additional iterations past the reference were performed
                                    if (maxPastRef != 0 && iterationP > maxPastRef) {
                                        glitchFlagTemp[rRef + rP, cRef + cP] = 1;
                                        Interlocked.Increment(ref cntFlagGlitch);
                                        if (flagDebugGlitch == true) iterationPD = maxIter;
                                    }

                                    // Store result to array
                                    if (res == 1) imageDataTemp[rRef - 1 + rP, cRef - 1 + cP] = iterationPD;
                                    else reduceResolution(ref imageDataTemp, rRef + rP, cRef + cP, iterationPD, res);
                                }
                            }
                        }

                    }  //next c
                });  //next r
                if (numPointsMP == 1) excelApp.StatusBar = string.Format("Calculating... {0:0%}      DLL in use on {1} logical processors.{2}", (i + 1) / (double)n, numProcessors, dllOptionString);
                else excelApp.StatusBar = string.Format("Correcting initial results ({2})... {0:0%}      DLL in use on {1} logical processors.{3}", (i + 1) / (double)n, numProcessors, numPointsMP, dllOptionString);
            }  //next i
            copyArray(ref imageData, imageDataTemp);  //assign imageDataTemp to imageData
            copyArray(ref glitchFlag, glitchFlagTemp);
            if (cntFlagGlitch > 0) flagGlitch = true;
        }
    }
}
